(* -----------------------------------------------------------------

  Braids.m

  Links--Gould Explorer uses (the closures of) braids as one of three 
  fundamental ways in which it describes links. (The other two forms are 
  Dowker--Thistlethwaite codes and planar diagram codes.)

  It uses several different braid presentations, and this file provides 
  conversion functions to translate between them. (Read the usage messages!) 
  There are no functions for the reduction or comparison of braids, and none \
\
for 
  the construction of braids from other sources (for example, from 
  Dowker--Thistlethwaite codes). There is a function to construct planar \
diagram 
  codes from braid words.

  Danger: For a given link name, it is possible for a user to define braid 
  presentations in each of various forms, which do not agree, and the \
resulting 
  evaluations of link invariants may be include unexpected inconsistencies. \
\
For 
  safety, the user should only enter one braid form for any given link name, \
\
and 
  let the automatic conversion functions do the conversions. This caveat also \
\

  applies to the other two forms of link definition.

  Warning: The Z-algorithms implemented by me use braid presentations of \
links 
  without associated braid indices; they internally assume and compute the \
index 
  of the braid group to be the minimum possible. In contrast, algorithms 
  borrowed from Dror Bar-Natan use braid presentations which carry explicit 
  braid indices. In time, the data-structrures of my code will be upgraded so \
\

  that all braid presentations carry braid group indices.

  Note that the BraidWord for the unknot (that is, PrimeKnot[0, "A", 1]) is 
  hard-coded as "" (below) rather than being loaded from the data table 
  "00A.brd". In general, algorithms are not set up to manipulate the identity \
\

  braid, but OpenTangleLinkInvariantOfALinkFromABraidWord is.
  

  Although it is more generally applicable, here we include 
  WriteAllLoadingMessagesFlag. In general, data-loading messages only appear 
  when the data takes more than a brief time to load up. This is arbitrarily 
  set for HTW knot data of more than 12 crossings. By setting: 
  \"WriteAllLoadingMessagesFlag = True\", all data-loading messages appear. 
  The defult is False.

  ------------------------------------------------------------------ *)

BeginPackage["LinksGouldExplorer`"];

WriteAllLoadingMessagesFlag = False;

BraidWord::usage = "
  BraidWord[L] is an alphabetic string which is a braid word presentation for \
\

  the named link L. That is, it is a presentation of a braid whose closure 
  (within the miminum-possible braid group) is L.

  It uses the standard convention that lower case alphabetic letters stand \
for 
  positive crossings, upper case for negative, for example \"AAbCaccc\" is a 
  braid word corresponding to the usual 
  sigma_1^{-2}sigma_2sigma_3^{-1}sigma_1sigma_3^{-3}. Of course, this means 
  that the local braid universe is limited to braids of index at most 27. In 
  contrast to various conventions used elsewhere, non-alphabetic symbols are 
  not permitted, so, for example, \"A^2bCac^3\" is not a valid braid word. 
  (For example, \textsc{K2K} permits that \"b-word\", although not 
  \"a^-2bc^-1ac^3\".) Braid words which are immediately reducible, such as 
  \"Aa\" and \"ACbBccaA\" are quite valid, as are braids with `gaps', such as \
\

  \"ac\" and \"BBBBB\". The unknot is described as the closure of the \
identity 
  braid \"\". Unfortunately, although it is a valid braid, the use of the 
  identity braid is not fully implemented in algorithms. Warning!

  Braid words are automatically loaded/computed for links L of the form 
  \"PrimeKnot[c, alt, i]\", \"Closed2Braid[k]\" and 
  \"TwoBridgeLinkSequence[{a1,a2}]\".
";

SoberBraidWordQ::usage = "
  SoberBraidWordQ[bw] is True if bw is a well-formed braid word. It tolerates \
\

  the unused \"\".
";

BraidSyllables::usage = "
  BraidSyllables[L] is a list of the syllables corresponding to 
  \"BraidWord[L]\". For instance, if \"BraidWord[L]\" is \"AAbCaccc\", it is 
  {\"AA\",\"b\",\"C\",\"a\",\"ccc\"}.
";

SoberBraidSyllablesQ::usage = "
  SoberBraidSyllablesQ[bs] is True if bs is a well-formed list of braid 
  syllables. It tolerates \"{}\".
";

SignedIntegerBraid::usage = "
  SignedIntegerBraid[L] is a list of signed integers corresponding to 
  \"BraidWord[L]\". For instance, if \"BraidWord[L]\" is \"AAbCaccc\", the 
  signed integer braid is \"{-1,-1,2,-3,1,3,3,3}\".

  In time, this function will be completely replaced by 
  \"SignedIntegerBraidRepresentative\".
";

SoberSignedIntegerBraidQ::usage = "
  SoberSignedIntegerBraidQ[sib] is True if \"sib\" is a valid signed integer 
  braid.
";

SignedIntegerBraidRepresentative::usage = "
  SignedIntegerBraidRepresentative[L] is SignedIntegerBraid[L] augmented with \
\

  the index n of the braid group that \"BraidWord[L]\" is regarded as being 
  in, and given a head \"SIBR\". If this index is automatically-computed, \
then 
  it is the minimal such n. For instance, if \"BraidWord[L]\" is  
  \"AAbCaccc\", the automatically-computed signed integer braid \
representative 
  is \"SIBR[4, {-1,-1,2,-3,1,3,3,3}]\". In contrast, a user may happily \
define 
  \"SIBR[5, {-1,-1,2,-3,1,3,3,3}]\", although not \"SIBR[3, 
  {-1,-1,2,-3,1,3,3,3}]\".
";

SoberSignedIntegerBraidRepresentativeQ::usage = "
  SoberSignedIntegerBraidRepresentativeQ[sibr] is True if \"sibr\" is a valid \
\

  signed integer braid.
";

EasyBraidWordQ::usage = "
  EasyBraidWordQ[bw] is True if the syllables of bw have no repeated letters 
  (positive or negative). Braid words with omitted letters are regarded as \
easy 
  as they can be regarded as having letters of zero power. By convention, the \
\

  trivial braid is not easy.
";

OppositeBraidWord::usage = "
  OppositeBraidWord[bw] is the same braid word, but with indices and \
crossings 
  numbered as viewed from behind!
";

StringIndexOfBraidWord::usage = "
  StringIndexOfBraidWord[bw] is the string index of the given braid word. \
This 
  is not the braid index of the braid (of the associated link) formed by its 
  closure.
";

SignedIntegerBraidWidth::usage = "
  SignedIntegerBraidWidth[bw].
";

StringIndex::usage = "
  StringIndex[link] is the string index of the representative braid for which \
\

  link is the closure.
";

KnotTheory`BraidIndex::usage = "
  KnotTheory`BraidIndex[link] is the braid index of link.
";

BraidWordLength::usage = "
  BraidWordLength[bw] is the length of the given braid word.
";

WritheOfBraidWord::usage = "
  WritheOfBraidWord[bw] is the writhe of the braid presentation of the named 
  braid word.
";

WritheOfSignedIntegerBraid::usage = "
  WritheOfSignedIntegerBraid[sib].
";

BraidSyllablesFromBraidIndicesAndPowers::usage = "
  BraidSyllablesFromBraidIndicesAndPowers[{indices, powers}].
";

BraidSyllablesFromSignedIntegerBraid::usage = "
  BraidSyllablesFromSignedIntegerBraid[sib].
";

BraidSyllablesFromBraidWord::usage = "
  BraidSyllablesFromBraidWord[sib].
";

BraidWordFromBraidSyllables::usage = "
  BraidWordFromBraidSyllables[bs].
";

BraidWordFromBraidIndicesAndPowers::usage = "
  BraidWordFromBraidIndicesAndPowers[{indices, powers}].
";

BraidWordFromSignedIntegerBraid::usage = "
  BraidWordFromSignedIntegerBraid[sib].
";

BraidIndicesAndPowersFromBraidSyllables::usage = "
  BraidIndicesAndPowersFromBraidSyllables[bs].
";

BraidIndicesAndPowersFromBraidWord::usage = "
  BraidIndicesAndPowersFromBraidWord[bw].
";

BraidIndicesAndPowersFromSignedIntegerBraid::usage = "
  BraidIndicesAndPowersFromSignedIntegerBraid[sib].
";

SignedIntegerBraidFromBraidSyllables::usage = "
  SignedIntegerBraidFromBraidSyllables[bs].
";

SignedIntegerBraidFromBraidWord::usage = "
  SignedIntegerBraidFromBraidWord[bw].
";

SignedIntegerBraidFromBraidIndicesAndPowers::usage = "
  SignedIntegerBraidFromBraidIndicesAndPowers[{indices, powers}].
";

SignedIntegerBraidRepresentativeFromBraidWord::usage = "
  SignedIntegerBraidRepresentativeFromBraidWord[bw].
";

ReflectionOfSignedIntegerBraidRepresentative::usage = "
  ReflectionOfSignedIntegerBraidRepresentative[sibr_SIBR] is the reflection \
\
of 
  the signed integer braid representative sibr.
";


AlreadyLoadedBraidWordsOfHTWPrimeKnots::usage = "
  AlreadyLoadedBraidWordsOfHTWPrimeKnots[c, alt] is a flag.
";

LoadBraidWordsOfHTWPrimeKnots::usage = "
  LoadBraidWordsOfHTWPrimeKnots[c, alt].
";

LoadStringIndicesOfHTWPrimeKnots::usage = "
  LoadStringIndicesOfHTWPrimeKnots[c, alt].
";

StoreStringIndicesOfHTWPrimeKnots::usage = "
  StoreStringIndicesOfHTWPrimeKnots[c, alt].
";

StoreStringIndicesOfHTWPrimeKnotsOfUpTocCrossings::usage = "
  StoreStringIndicesOfHTWPrimeKnotsOfUpTocCrossings[c].
";

(* ----------------------------------------------------------------- *)

(* Begin["`Private`"]; *)

SoberBraidWordQ[bw_] :=
  ((Head[bw]===String) && LetterQ[bw]);

SoberBraidSyllablesQ[bs_List?VectorQ] :=
  If[(Length[bs] >= 1), Apply[And, Map[SoberBraidWordQ, bs]], True];

SoberSignedIntegerBraidQ[sib_List?VectorQ] :=
  (Apply[And, Apply[((Head[#]===Integer) && (#=!=0))&, sib]] && \
(Max[Abs[sib]] <= 26));

SoberSignedIntegerBraidRepresentativeQ[sibr_SIBR] :=
  (
    SoberSignedIntegerBraidQ[sibr[[2]]]
    &&
    (Head[sibr[[1]]]===Integer)
    &&
    (Max[Abs[sibr[[2]]]] < sibr[[1]])
  );

(* ----------------------------------------------------------------- *)

EasyBraidWordQ[""] = False;

EasyBraidWordQ[bw_] :=
  Module[
    {bi, ebq},

    ebq =
      If[
        SoberBraidWordQ[bw],
        bi = Sort[BraidIndicesAndPowersFromBraidWord[bw][[1]]];
        (bi === Union[bi]),
        False
      ];

    ebq
  ];

(* ----------------------------------------------------------------- *)

OppositeBraidWord[bw_String?SoberBraidWordQ] :=
  Module[
    {sib, asib, ssib, n, rsib, obw},

    sib = SignedIntegerBraidFromBraidWord[bw];
    asib = Abs[sib];
    ssib = Sign[sib];
    n = Max[Union[Abs[sib]]] + 1;
    rsib = ssib * (n - asib);
    obw = BraidWordFromSignedIntegerBraid[rsib];

    obw
  ];

(* ----------------------------------------------------------------- *)

BraidWord[PrimeKnot[c_, alt_, i_]] :=
  (
    LoadBraidWordsOfHTWPrimeKnots[c, alt];
    BraidWord[PrimeKnot[c, alt, i]]
  ) /; SoberHTWPrimeKnotQ[PrimeKnot[c, alt, i]];

(* ----------------------------------------------------------------- *)

StringIndex[PrimeKnot[c_, alt_, i_]] :=
  (
    LoadStringIndicesOfHTWPrimeKnots[c, alt];
    StringIndex[PrimeKnot[c, alt, i]]
  ) /; SoberHTWPrimeKnotQ[PrimeKnot[c, alt, i]];

(* ----------------------------------------------------------------- *)

StringIndexOfBraidWord[bw_String?SoberBraidWordQ] :=
  SignedIntegerBraidWidth[SignedIntegerBraidFromBraidWord[bw]];

SignedIntegerBraidWidth[{}] = 1;

SignedIntegerBraidWidth[sib_List] :=
  (Max[Abs[Union[sib]]] + 1) /;
    SoberBraidWordQ[BraidWordFromSignedIntegerBraid[sib]];

StringIndex[L_] :=
  StringIndexOfBraidWord[BraidWord[L]] /;
    SoberBraidWordQ[BraidWord[L]];

KnotTheory`BraidIndex[PrimeKnot[c_, alt_, i_]] :=
  StringIndex[PrimeKnot[c, "A", i]] /;
    (SoberHTWPrimeKnotSetQ[c] && (c <= 12));

(* ----------------------------------------------------------------- *)

BraidWordLength[bw_String?SoberBraidWordQ] :=
  StringLength[bw];

(* ----------------------------------------------------------------- *)

WritheOfSignedIntegerBraid[sib_List] :=
  Apply[Plus, Sign[sib]] /; \
SoberBraidWordQ[BraidWordFromSignedIntegerBraid[sib]];

WritheOfBraidWord[bw_String?SoberBraidWordQ] :=
  WritheOfSignedIntegerBraid[SignedIntegerBraidFromBraidWord[bw]];

(* ----------------------------------------------------------------- *)

BraidWordFromSignedIntegerBraid[sib_List] :=
  StringJoin[
    sib /.
      Thread[Join[-Range[26], Range[26]] -> Join[CharacterRange["A", "Z"], \
CharacterRange["a", "z"]]]
  ];

BraidSyllablesFromBraidWord[bw_String?SoberBraidWordQ] :=
  BraidSyllablesFromBraidWord[bw] =
    Map[BraidWordFromSignedIntegerBraid, \
Split[SignedIntegerBraidFromBraidWord[bw]]];

BraidSyllablesFromSignedIntegerBraid[sib_List] :=
  BraidSyllablesFromBraidWord[BraidWordFromSignedIntegerBraid[sib]];

BraidSyllablesFromBraidIndicesAndPowers[{indices_List, powers_List}] :=
  BraidSyllablesFromSignedIntegerBraid[\
SignedIntegerBraidFromBraidIndicesAndPowers[{indices, powers}]];


SignedIntegerBraidFromBraidWord[bw_String?SoberBraidWordQ] :=
  Characters[bw] /.
    Thread[Join[CharacterRange["A", "Z"], CharacterRange["a", "z"]] -> \
Join[-Range[26], Range[26]]];

SignedIntegerBraidFromBraidIndicesAndPowers[{indices_List, powers_List}] :=
  Flatten[Inner[Table[#1,{#2}]&, Sign[powers] * indices, Abs[powers], List]];

SignedIntegerBraidFromBraidSyllables[bs_List] :=
  SignedIntegerBraidFromBraidIndicesAndPowers[\
BraidIndicesAndPowersFromBraidSyllables[bs]];

BraidWordFromBraidSyllables[bs_List] :=
  BraidWordFromBraidIndicesAndPowers[BraidIndicesAndPowersFromBraidSyllables[\
\
bs]];

BraidWordFromBraidIndicesAndPowers[{indices_List, powers_List}] :=
  BraidWordFromSignedIntegerBraid[SignedIntegerBraidFromBraidIndicesAndPowers[\
\
{indices, powers}]];


BraidIndicesAndPowersFromBraidWord[bw_String?SoberBraidWordQ] :=
  BraidIndicesAndPowersFromBraidSyllables[BraidSyllablesFromBraidWord[bw]];

BraidIndicesAndPowersFromSignedIntegerBraid[sib_List] :=
  BraidIndicesAndPowersFromBraidWord[BraidWordFromSignedIntegerBraid[sib]];

BraidIndicesAndPowersFromBraidSyllables[bs_List] :=
  Module[
    {indices, powers},

    indices = Flatten[Map[Union[SignedIntegerBraidFromBraidWord[#]]&, bs]];
    powers = Sign[indices] * Map[StringLength, bs];
    indices = Abs[indices];

    {indices, powers}
  ];

(* ----------------------------------------------------------------- *)

SignedIntegerBraid[Closed2Braid[k_Integer]] :=
  Table[Sign[k], {Abs[k]}];

SignedIntegerBraid[KnotTheory`TorusKnot[m_Integer, n_Integer?Positive]] :=
  Flatten[Table[Sign[m] * Range[n-1], {Abs[m]}]];

SignedIntegerBraid[L_] :=
  SignedIntegerBraidFromBraidWord[BraidWord[L]] /;
    SoberBraidWordQ[BraidWord[L]];

BraidWord[L_] :=
  BraidWordFromSignedIntegerBraid[SignedIntegerBraid[L]] /;
    MemberQ[{Closed2Braid, KnotTheory`TorusKnot}, Head[L]];

BraidSyllables[L_] :=
  BraidSyllablesFromBraidWord[BraidWord[L]] /;
    SoberBraidWordQ[BraidWord[L]];

(* ----------------------------------------------------------------- *)

BraidWord[LinkReflection[L_]] :=
  BraidWordFromSignedIntegerBraid[-SignedIntegerBraidFromBraidWord[BraidWord[\
\
L]]] /;
    SoberBraidWordQ[BraidWord[L]];

BraidWord[LinkReflection[L_]] :=
  ReflectionOfSignedIntegerBraidRepresentative[\
SignedIntegerBraidRepresentative[L]] /;
    (Head[SignedIntegerBraidRepresentative[L]]===SIBR);

BraidWord[LinkInversion[L_]] :=
  StringReverse[BraidWord[L]] /;
    SoberBraidWordQ[BraidWord[L]];

(* ----------------------------------------------------------------- *)

SignedIntegerBraidRepresentativeFromBraidWord[bw_String?SoberBraidWordQ] :=
  SIBR[StringIndexOfBraidWord[bw], SignedIntegerBraidFromBraidWord[bw]];

SignedIntegerBraidRepresentativeFromSignedIntegerBraid[sib_List] :=
  SIBR[SignedIntegerBraidWidth[sib], sib];

BraidWordFromSignedIntegerBraidRepresentative[sibr_SIBR] :=
  BraidWordFromSignedIntegerBraid[\
SignedIntegerBraidFromSignedIntegerBraidRepresentative[sibr]];

SignedIntegerBraidFromSignedIntegerBraidRepresentative[sibr_SIBR] :=
  sibr[[2]];

SignedIntegerBraidRepresentative[LinkReflection[L_]] :=
  ReflectionOfSignedIntegerBraidRepresentative[\
SignedIntegerBraidRepresentative[L]] /;
    SoberBraidWordQ[BraidWord[L]];

SignedIntegerBraidRepresentative[L_] :=
  SignedIntegerBraidRepresentativeFromSignedIntegerBraid[SignedIntegerBraid[L]\
\
] /;
    (Head[SignedIntegerBraid[L]]===List);

SignedIntegerBraidRepresentative[L_] :=
  SignedIntegerBraidRepresentativeFromBraidWord[BraidWord[L]] /;
    SoberBraidWordQ[BraidWord[L]];

(* ----------------------------------------------------------------- *)

ReflectionOfSignedIntegerBraidRepresentative[SIBR[n_Integer?Positive, \
l_List?VectorQ]] :=
  SIBR[n, -l];

SignedIntegerBraidRepresentative[LinkReflection[L_]] :=
  ReflectionOfSignedIntegerBraidRepresentative[\
SignedIntegerBraidRepresentative[L]] /;
    (Head[SignedIntegerBraidRepresentative[L]]===SIBR);

(* ----------------------------------------------------------------- *)

Apply[(AlreadyLoadedBraidWordsOfHTWPrimeKnots[##] = True)&, \
EmptyHTWPrimeKnotCrossingNumberAndAlternatingnessLabelSet, 1]; 
Apply[(AlreadyLoadedBraidWordsOfHTWPrimeKnots[##] = False)&, \
NonEmptyHTWPrimeKnotCrossingNumberAndAlternatingnessLabelSet, 1];

LoadBraidWordsOfHTWPrimeKnots[c_, alt_] :=
  Module[
    {fcode, ifile},

    If[
      Not[AlreadyLoadedBraidWordsOfHTWPrimeKnots[c, alt]],
      fcode = StringJoin[If[c < 10, "0", ""], ToString[c], alt];
      ifile = StringJoin[LGEDirectory, "/data/Prime Knots/Braid Words/", \
fcode, ".brd"];
      If[((c > 12) || WriteAllLoadingMessagesFlag), WriteString["stdout", \
"Loading: ", ifile, " ... "]];
      If[
        ((c == 0) && (alt === "A")),   (* kludge *)
        BraidWord[PrimeKnot[0, "A", 1]] = "",
        MapThread[
          (BraidWord[#1] = #2)&,
          {HTWPrimeKnots[c, alt], Import[ifile, "Words"]}
        ]
      ];
      If[((c > 12) || WriteAllLoadingMessagesFlag), WriteString["stdout", \
"done\n"]];
      AlreadyLoadedBraidWordsOfHTWPrimeKnots[c, alt] = True
    ];
  ] /; SoberHTWPrimeKnotSetQ[c, alt];

(* ----------------------------------------------------------------- *)

LoadStringIndicesOfHTWPrimeKnots[c_, alt_] :=
  Module[
    {ifile, fcode},

    fcode = StringJoin[If[c < 10, "0", ""], ToString[c], alt];
    ifile = StringJoin[LGEDirectory, "/data/Prime Knots/String Indices/", \
fcode, ".si"];
    If[((c > 12) || WriteAllLoadingMessagesFlag), WriteString["stdout", \
"Loading: ", ifile, " ... "]];
    MapThread[
      (StringIndex[#1] = ToExpression[#2])&,
      {HTWPrimeKnots[c, alt], Import[ifile, "Lines"]}
    ];
    If[((c > 12) || WriteAllLoadingMessagesFlag), WriteString["stdout", "done\
\
\n"]]
  ] /; SoberHTWPrimeKnotSetQ[c, alt];

(* ----------------------------------------------------------------- *)

StoreStringIndicesOfHTWPrimeKnots[c_, alt_] :=
  Module[
    {ofile, ostream, fcode, strinds},

(* This code calculates new string indices, rather than using old stored ones \
\
 *)
    strinds = Map[SignedIntegerBraidWidth, Map[SignedIntegerBraid, \
HTWPrimeKnots[c, alt]]];
    fcode = StringJoin[If[c < 10, "0", ""], ToString[c], alt];
    ofile = StringJoin[LGEDirectory, "/data/Prime Knots/String Indices/New ", \
\
fcode, ".si"];
    WriteString["stdout", "Storing: ", ofile, " ... "];
    ostream = OpenWrite[ofile];
    Map[WriteString[ostream, #, "\n"]&, strinds];
    Close[ostream];
    WriteString["stdout", "done\n"];
  ] /; SoberHTWPrimeKnotSetQ[c, alt];

StoreStringIndicesOfHTWPrimeKnots[c_] :=
  Scan[Apply[StoreStringIndicesOfHTWPrimeKnots, #]&, \
HTWPrimeKnotCrossingNumberAndAlternatingnessLabelSet[c]] /;
    SoberHTWPrimeKnotSetQ[c];

StoreStringIndicesOfHTWPrimeKnotsOfUpTocCrossings[c_] :=
  Scan[Apply[StoreStringIndicesOfHTWPrimeKnots, #]&, \
HTWPrimeKnotCrossingNumberAndAlternatingnessLabelSet[c]] /;
    SoberHTWPrimeKnotSetQ[c];

(* ----------------------------------------------------------------- *)

(* End[]; *)

EndPackage[];


