(* -----------------------------------------------------------------

  Laurent Polynomials.m

  Convert Laurent polynomials to and from arrays/lists of (integer) 
  coefficients.

  ------------------------------------------------------------------ *)

BeginPackage["LinksGouldExplorer`"];

IntegerArraysFromIntegerCoefficientLaurentPolynomial::usage = "
  IntegerArraysFromIntegerCoefficientLaurentPolynomial[poly] is a pair {l, M} \

  of arrays describing the n-variable Laurent polynomial poly. Here 
  l = {l^1, l^2, ..., l^n} is a set of lists l^j of integer exponents of 
  variables, and M is a rank n array describing the corresponding 
  coefficients. This works regardless of the number of variables in poly. 
  Where #(l^j) is the number of elements in set l^j, we have:

  poly =
    sum_{i_1 = 1, ..., #(l^1); i_2 = 1, ..., #(l^2); ...; i_n = 1, ..., \
#(l^n);}
      M_{i_1, i_2, ..., i_n}
      v_1^{l^1_{i_1}} v_2^{l^2_{i_2}} ... v_n^{l^n_{i_n}}.

  Important: The variables in poly are taken to be ordered alphabetically, \
so, 
  the coefficients of a Links--Gould polynomial in variables (q,p) will be 
  `reverse-ordered' by this function.
";

IntegerCoefficientLaurentPolynomialFromIntegerArrays::usage = "
  IntegerCoefficientLaurentPolynomialFromIntegerArrays[{ev, M}] reconstructs \
a 
  multivariable Laurent polynomial from the information in ev (a list of \
lists 
  of exponents appearing in the polynomial), and M (the corresponding 
  coeficients. This function is currently only implemented for the 2-variable \

  case.
";

(* ----------------------------------------------------------------- *)

(* Begin["`Private`"]; *)

LPCoeffs::usage = "
  LPCoeffs[poly, v, ev] is a ultility function called by
  IntegerArraysFromIntegerCoefficientLaurentPolynomial.
";

LPCoeffs[poly_, v_, ev_] :=
  Map[Coefficient[poly, v, #]&, ev];

IntegerArraysFromIntegerCoefficientLaurentPolynomial[poly_] :=
  Module[
    {ipoly, vars, numvars, ev, FF, j},

    ipoly = Expand[poly];
    vars = Sort[Variables[ipoly]];
    numvars = Length[vars];
    ev = Map[Exponent[ipoly, #, List]&, vars];
    FF = ipoly;
    Do[
      FF = Map[LPCoeffs[#, vars[[j]], ev[[j]]]&, FF],
      {j, numvars}
    ];

    {ev, FF}
  ];

(* This function is specialised to the 2-variable case *)

IntegerCoefficientLaurentPolynomialFromIntegerArrays[{ev_, FF_}] :=
  Module[
    {poly, ev1, ev2, v1, v2},

    {ev1, ev2} = ev;
    poly = Apply[Plus, Flatten[FF * Outer[(v1^#1 v2^#2)&, ev1, ev2]]];
    poly = Evaluate[poly /. v2 -> #2 /. v1 -> #1]&;

    poly
  ] /; ((Length[ev]==2) && (TensorRank[FF]==2));

(* ----------------------------------------------------------------- *)

(* End[]; *)

EndPackage[];

