(* -------------------------------------------------------------------

  Links--Gould Invariants.m

  A number of flags help define the specific behaviour of these functions. \
Their 
  values are set just below \"BeginPackage\".

  1. Precomputed evaluations for Links--Gould invariants are only \
automatically 
  loaded (when needed) if \"AutoLoadLGInvariantsFlag = True\" (the default).

  2. For LG^{m,n}, internal state model computations use state model \
parameters 
  defined in terms of either a unitary or nonunitary basis. In the latter \
case, 
  the state model parameters are purely polynomial, and the resulting 
  computations are more efficient. This default situation is ensured by \
setting: 
  \"UseNonunitaryBasisVariablesForLGStateModelParametersFlag = True\"

  3. In general, internal state model computations use the variables (Q, p), 
  where $Q equiv q^{1/2}. This makes the computations slightly more efficient \
\
\

  than if they use (q, p)$. For LG^{2,1} the eigenvalue variables (t_0, t_1) \
\
\
are 
  an alternative to these variables. Especially when combined  with \
nonunitary 
  basis variables, they lead to more efficient calculations. This default 
  situation is ensured by setting: \"UseEigenvalueVariablesForLG21Flag = True\
\
\
\".

  To use them, first set q \to t_0^(1/2) t_1^(1/2) and p \to t_0^(-1/4) 
  t_1^(1/4), that is, with decoding: t_0 -> p^(-2) q and t_1 \to p^(+2) q. \
These 
  are implemented, and useful for the formula-based computation of LG^{2,1} \
\
\
for 
  2-bridge links as 
  \"ConvertFromEigenvalueVariablesToDualitySymmetricVariables\" and 
  \"ConvertFromDualitySymmetricVariablesToEigenvalueVariables\".

  Aside: Writing q \to Q^2 to agree with the general LG^{m,n} case, these are \
\
\

  more symmetric: Q \to t_0^(+1/4) t_1^(+1/4) and p \to t_0^(-1/4) \
t_1^(+1/4), 
  that is: t_0 -> p^(-2) Q^2 and t_1 \to p^(+2) Q^2.

  This can be improved by instead using variables (T_0, T_1) defined by t_0 \t\
\
\
o 
  T_0^2 and t_1 \to T_1^2, which means the net transformation is: q \to T_0 \
\
\
T_1 
  and p \to T_0^(-1/2) T_1^(1/2), that is:

    q -> T0 * T1                      p -> T0^(-1/2) * T1^(+1/2).

  The decoding is then:

    T_0 -> q^(1/2) p^(-1)             T_1 -> q^(1/2) p^(+1).

  These conversions are implemented by the functions 
  \"ConvertToInternalComputationalVariablesForLGRules\" and 
  \"ConvertFromInternalComputationalVariablesForLGRules\".

  Note that the inversion of variables (q, p) corresponds to the inversion of \
\
\

  (T_0, T_1), equivalently, the inversion of (t_0, t_1).

  ----------------------------------------------------------------- *)

BeginPackage["LinksGouldExplorer`"];

AutoLoadLGInvariantsFlag = True;

UseNonunitaryBasisVariablesForLGStateModelParametersFlag = True;

UseEigenvalueVariablesForLG21Flag = True;

MaxIndexOfLGOfClosed2Braids = 5;

LinksGouldInvariant::usage = "
  LinksGouldInvariant[m, n, L][q,p] is the Links--Gould invariant LG^{m,n}_L 
  presented as a Laurent polynomial in the natural variables q and p. As we \
\
\
do 
  not yet have proof that LinksGouldInvariant[m, n, L] is always a \
polynomial, 
  we do not call the function LinksGouldPolynomial. If n>m, the function 
  evaluates and transforms LinksGouldInvariant[n,m,L]. Depending on L, the 
  function uses a formula for LG^{2,1}_L for 2-bridge links L, another if L \
\
\
is 
  an easy braid, and, more generally, if a braid word is available for L, the \
\
\

  state model method.
";

LinksGouldInvariantInEigenvalueVariables::usage = "
  LinksGouldInvariantInEigenvalueVariables[m, n, L][t1, t1] is 
  LinksGouldInvariant[m, n, L] expressed in terms of the eigenvalue variables \
\
\

  t_0 and t_1. The conversions are:

    LinksGouldInvariantExpressedInTermsOfInEigenvalueVariables[p^(-2) q, \
p^(+2) q] 
    = 
    LinksGouldInvariant[m, n, L][p, q]

    LinksGouldInvariant[m, n, L][t_0^(1/2) t_1^(1/2), t_0^(-1/4) t_1^(1/4)]
    = 
    LinksGouldInvariantExpressedInTermsOfInEigenvalueVariables[t0, t1]

  Thanks go to Atsushi Ishii for these variables.
";

ValidEvaluationOfLinksGouldInvariant::usage = "
  ValidEvaluationOfLinksGouldInvariant[poly] is True if the (pure function) 
  poly really is a two-variable polynomial and is invariant (or `-invariant') \
\
\

  under inversion of the second variable. This is a useful debugging tool.
";

LinksGouldInvariantAvailableStateModelIndices::usage = "
  LinksGouldInvariantAvailableStateModelIndices is the set of index pairs {m, \
\
\

  n} such that state model parameters for LG^{m,n} are available.
";

SoberLinksGouldInvariantAvailableStateModelIndicesQ::usage = "
  SoberLinksGouldInvariantAvailableStateModelIndicesQ[m, n] is True if {m,n} 
  is contained in LinksGouldInvariantAvailableStateModelIndices.
";

LinksGouldInvariantOfEasyBraidWordLink::usage = "
  LinksGouldInvariantOfEasyBraidWordLink[m, n, L] is the Links--Gould 
  invariant LG^{m,n}_L, where L is an easy braid.
";

LinksGouldInvariantOfTwoBridgeLinkGivenByConwayPresentation::usage = "
  LinksGouldInvariantOfTwoBridgeLinkGivenByConwayPresentation[2, 1, a] is 
  LG^{2,1} of the 2-bridge link C(a), but is only valid when C(a)=D(a/2) \
where 
  a/2 is a valid sequence of integers.
";

LinksGouldInvariantFromAStateModelOnABraidWord::usage = "
  LinksGouldInvariantFromAStateModelOnABraidWord[m, n, L] evaluates \
LG^{m,n}_L 
  via the state model method, where a braid word is available for L. This is 
  currently only available for n=1 and m=1,2,3.
";

qPolynomialCoefficientsOfPositivePowersOfpOfLinksGouldInvariant::usage = "
  qPolynomialCoefficientsOfPositivePowersOfpOfLinksGouldInvariant[m, n, L] is \
\
\

  LG^{m,n}_L; the evaluation presented as a list of q-polynomial coefficients \
\
\

  of positive powers of the appropriate duality-symmetric variable p. Danger: \
\
\

  This doesn't check whether the symmetry is p^i \to p^{-i} or p^i \to - 
  p^{-i}.
";

LinksGouldInvariantCoefficients::usage = "
  LinksGouldInvariantCoefficients[m, n, L] is a list of the coefficients of 
  LG^{m,n}_L(q,p); presented as a pair {{rp, rq}, mx} of two lists of \
integers rp 
  and rq describing powers of variables p and q and a matrix mx of 
  coefficients.
";

AlreadyLoadedStateModelParametersForLinksGouldInvariant::usage = "
  AlreadyLoadedStateModelParametersForLinksGouldInvariant[m, n] is a flag.
";

ConvertToInternalComputationalVariablesForLG::usage = "
  ConvertToInternalComputationalVariablesForLG[m, n, S].
";

ConvertFromInternalComputationalVariablesForLG::usage = "
  ConvertFromInternalComputationalVariablesForLG[m, n, S].
";

ConvertFromEigenvalueVariablesToDualitySymmetricVariables::usage = "
  ConvertFromEigenvalueVariablesToDualitySymmetricVariables[2, 1, S].
";

ConvertFromDualitySymmetricVariablesToEigenvalueVariables::usage = "
  ConvertFromDualitySymmetricVariablesToEigenvalueVariables[2, 1, S].
";


LoadStateModelParametersForLinksGouldInvariant::usage = "
  LoadStateModelParametersForLinksGouldInvariant[m, n] loads the precomputed 
  representations of the braid generator and its inverse, together with that 
  of the positive handle, for LG^{m,n}.
";

LinksGouldInvariantStateModelFeasibleHTWPrimeKnots::usage = "
  LinksGouldInvariantStateModelFeasibleHTWPrimeKnots[m, n, c, alt] is 
  hard-wired data.
";

LinksGouldInvariantStateModelFeasibleHTWPrimeKnotsOfUpTocCrossings::usage = "
  LinksGouldInvariantStateModelFeasibleHTWPrimeKnotsOfUpTocCrossings[m, n, \
c].
";

LinksGouldInvariantFeasibleHTWPrimeKnots::usage = "
  LinksGouldInvariantFeasibleHTWPrimeKnots[m, n, c, alt] is the list of 
  LG^{m,n}-feasible HTW prime knots of c crossings. For LG^{2,1}, for 
  alternating knots, this is an amalgam of two-bridge knots and state-model 
  feasible knots.
";

LinksGouldInvariantFeasibleHTWPrimeKnotsOfUpTocCrossings::usage = "
  LinksGouldInvariantFeasibleHTWPrimeKnotsOfUpTocCrossings[m, n, c].
";

NumberOfLinksGouldInvariantFeasibleHTWPrimeKnots::usage = "
  Length[LinksGouldInvariantFeasibleHTWPrimeKnots[m, n, c, alt]].
";

NumberOfLinksGouldInvariantFeasibleHTWPrimeKnotsOfUpTocCrossings::usage = "
  NumberOfLinksGouldInvariantFeasibleHTWPrimeKnotsOfUpTocCrossings[m, n, c].
";

OverlapLinksGouldInvariantStoredTwoBridgeHTWPrimeKnots::usage = "
  OverlapLinksGouldInvariantStoredTwoBridgeHTWPrimeKnots[2, 1, c] is the set \
\
\
of 
  LG^{2,1}-state-model-feasible prime knots of c crossings which are 2-bridge \
\
\

  knots (and so LG^{2,1}-evaluable via a formula.
";

NumberOfOverlapLinksGouldInvariantStoredTwoBridgeHTWPrimeKnots::usage = "
  NumberOfOverlapLinksGouldInvariantStoredTwoBridgeHTWPrimeKnots[2, 1, c].
";

LinksGouldInvariantStoredHTWPrimeKnots::usage = "
  LinksGouldInvariantStoredHTWPrimeKnots[m, n, c, alt] is the list of HTW 
  prime knots of c crossings for which evaluations of LG^{m,n} are stored in 
  the data tables. For LG^{2,1} and c <= 12 (respectively for LG^{3,1} and 
  c <= 10) these are HTWPrimeKnots[c, alt]. The routine to evaluate it is a 
  kludge!

  The values of LG^{2,1} are stored for the 
  \"LinksGouldInvariantFeasibleHTWPrimeKnots\" sets, and for various other 
  knots. Let w_1 be the Z_1-algorithm width of a given braid b; this is at 
  most the braid width of b. The Z_1 (or Z_2) algorithm often performs well 
  when w_1 is at most 4 or 5, even when the braid width is higher. Thus, 
  LG^{2,1} has been evaluated and stored for all prime knots of:

    w_1 <= 5 within mathbb{K}_{13}, mathbb{K}_{14}
    w_1 <= 4 within mathbb{K}_{15}, mathbb{K}_{16},
  and for a few sundries calculated for various idiosyncratic reasons.
";

LinksGouldInvariantStoredHTWPrimeKnotsOfUpTocCrossings::usage = "
  LinksGouldInvariantStoredHTWPrimeKnotsOfUpTocCrossings[m, n, c].
";

NumberOfLinksGouldInvariantStoredHTWPrimeKnots::usage = "
  NumberOfLinksGouldInvariantStoredHTWPrimeKnots[m, n, c, alt].
";

NumberOfLinksGouldInvariantStoredHTWPrimeKnotsOfUpTocCrossings::usage = "
  NumberOfLinksGouldInvariantStoredHTWPrimeKnotsOfUpTocCrossings[m, n, c].
";

LinksGouldInvariantEquivalentLinkTypesQ::usage = "
  LinksGouldInvariantEquivalentLinkTypesQ[m, n, L1, L2] is True if LG^{m,n} 
  cannot distinguish L1 from either L2 or from L2^*.
";

LinksGouldInvariantCliques::usage = "
  LinksGouldInvariantCliques[m, n, linklist] is the list of LG^{m,n}-cliques 
  within linklist.
";

LinksGouldInvariantCliquesOfLinksGouldInvariantFeasibleHTWPrimeKnots::usage = \
\
\
"
  LinksGouldInvariantCliquesOfLinksGouldInvariantFeasibleHTWPrimeKnots[m, n, \
\
\
c].
";

LinksGouldInvariantCliquesOfLinksGouldInvariantFeasibleHTWPrimeKnotsOfUpTocCro\
\
\
ssings::usage = "
  LinksGouldInvariantCliquesOfLinksGouldInvariantFeasibleHTWPrimeKnotsOfUpTocC\
\
\
rossings[m, n, c].
";


AlreadyLoadedLinksGouldInvariantsOfHTWPrimeKnots::usage = "
  AlreadyLoadedLinksGouldInvariantsOfHTWPrimeKnots[m, n, c, alt] is a flag.
";

LoadLinksGouldInvariantsOfHTWPrimeKnots::usage = "
  LoadLinksGouldInvariantsOfHTWPrimeKnots[m, n, c, alt].
";

StoreLinksGouldInvariantsOfHTWPrimeKnots::usage = "
  StoreLinksGouldInvariantsOfHTWPrimeKnots[m, n, c, alt].
";

StoreLinksGouldInvariantsOfHTWPrimeKnotsOfUpTocCrossings::usage = "
  StoreLinksGouldInvariantsOfHTWPrimeKnotsOfUpTocCrossings[m, n, c].
";

AlreadyLoadedLinksGouldInvariantsOfClosed2Braids::usage = "
  AlreadyLoadedLinksGouldInvariantsOfClosed2Braids is a flag.
";

LoadLinksGouldInvariantsOfClosed2Braids::usage = "
  LoadLinksGouldInvariantsOfClosed2Braids[m, n] load the precomputed values \
\
\
of 
  LG^{m,n} for various closed 2-braids.
";

KauffmanLinksGouldPolynomial::usage = "
  KauffmanLinksGouldPolynomial[L][q] is the KLG polynomial expressed in 
  variable q. This polynomial is the intersection between the Kauffman 
  polynomial and the Links--Gould invariant.
";

(* ----------------------------------------------------------------- *)

(* Begin["`Private`"]; *)

LinksGouldInvariant[m_Integer?Positive, n_Integer?Positive, L_] :=
  (LinksGouldInvariant[n, m, L] /. #1 -> 1/#1) /; (m < n);

LinksGouldInvariant[m_Integer?Positive, n_Integer?Positive, \
LinkReflection[L_]] :=
  LinksGouldInvariant[m, n, L] /. #1 -> 1/#1;

LinksGouldInvariant[m_Integer?Positive, n_Integer?Positive, \
LinkInversion[L_]] :=
  LinksGouldInvariant[m, n, L];

(* ------------------------------------------------------------------ *)

(*
  LG^{m,n} of links with "easy" braids. The following utility functions may \
\
\
in 
  time need to be moved to the "Representation Theory.m", and/or made public.
*)

zetaforAYD::usage = "
  zetaforAYD[Uqgl[m, n], ayd] is an auxiliary function used by LG^{m,n} for 
  evaluation on easy braids.
";

zetaforAYD[Uqgl[m_, n_], ayd_List] :=
  (zetaforAYD[Uqgl[m, n], ayd] =
    Expand[- alpha^2 n +  Apply[Plus, (ayd (ayd + 1 - 2 alpha - 2 * \
Range[Length[ayd]]))]]
  ) /; SoberAlgebraQ[Uqgl[m, n]];

chiforAYD::usage = "
  chiforAYD[Uqgl[m, n], ayd] is an auxiliary function used by 
  LinksGouldInvariantOfEasyBraidWordLink.
";

chiforAYD[Uqgl[m_, n_], ayd_List] :=
  (chiforAYD[Uqgl[m, n], ayd] =
  Module[
    {oprs, rho, rhoplusalphadelta, rhopluswtplus2alphadelta, qBs, chi, wayd},

    oprs = RootSystem[Uqgl[m, n], {1}, {+1}];
    rho = HalfSumofPositiveRoots[Uqgl[m, n]];
    rhoplusalphadelta = rho + Join[Table[0, {m}], Table[alpha, {n}]];
    wayd = WeightFromAllowableYoungDiagram[gl[m,n], ayd];
    rhopluswtplus2alphadelta = rho + wayd + Join[Table[0, {m}], Table[2 * \
alpha, {n}]];
    qBs =
      Map[
        qBracket[GradedInnerProductOfAlgebraWeights[gl[m, n], \
rhoplusalphadelta, #]] / qBracket[GradedInnerProductOfAlgebraWeights[gl[m, \
n], rhopluswtplus2alphadelta, #]]&,
        oprs
      ];
    chi = Apply[Times, qBs];
    chi = ConvertToqBrackets[SymmetriseqExpressions[chi]];

    chi
  ]) /; SoberAlgebraQ[Uqgl[m, n]];

qDimensionofV0::usage = "
  qDimensionofV0[Uqgl[m, n], RepName] is the q-dimension of the \
U_q[gl(m|n)]_0 
  module V^0(RepName).
";

qDimensionofV0[Uqgl[m_, n_], RepName_] :=
  (qDimensionofV0[Uqgl[m, n], RepName] =
  Module[
    {eprs, rho, qBs, qdim},

    eprs = RootSystem[gl[m, n], {0}, {+1}];
    rho = HalfSumofPositiveRoots[gl[m, n]];
    qBs =
      Map[qBracket[GradedInnerProductOfAlgebraWeights[gl[m, n], RepName + \
rho, #]]/qBracket[GradedInnerProductOfAlgebraWeights[gl[m, n], rho, #]]&,
        eprs];
    qdim = Apply[Times, qBs];
    qdim = Simplify[ExpandqBrackets[qdim]];

    qdim
  ]) /; SoberAlgebraQ[Uqgl[m, n]];

xiforUqglmn::usage = "
  xiforUqglmn[Uqgl[m, n], RepName, k] is an auxiliary function used by 
  LinksGouldInvariantOfEasyBraidWordLink.
";

xiforUqglmn[Uqgl[m_, n_], RepName_, k_Integer] :=
  (xiforUqglmn[Uqgl[m, n], RepName, k] =
  Module[
    {ayds, sm, xi},

    ayds = AllowableYoungDiagramSet[Uqgl[m, n], RepName];
    sm =
      Map[
        (-1)^((k-1)* Apply[Plus, #]) * q^Expand[- k * zetaforAYD[Uqgl[m,n], \
\
\
#]] * chiforAYD[Uqgl[m,n], #] *
        qDimensionofV0[Uqgl[m,n], WeightFromAllowableYoungDiagram[gl[m,n], \
#]]&,
      ayds
      ];
    sm = Cancel[ExpandqBrackets[sm]];
    xi = Apply[Plus, sm];

    xi
  ]) /; SoberAlgebraQ[Uqgl[m, n]];

LinksGouldInvariantOfEasyBraidWordLink[m_Integer?Positive, \
n_Integer?Positive, L_] :=
  (LinksGouldInvariantOfEasyBraidWordLink[m, n, L] =
  Module[
    {RepName, bw, writhe, bi, bp, powerlist, lg},

    RepName = LGIHighestWeight[Uqgl[m, n]];
    bw = BraidWord[L];
    writhe = WritheOfBraidWord[bw];
    {bi, bp} = BraidIndicesAndPowersFromBraidWord[bw]; 
    powerlist = Join[Table[0, {Max[bi] - Length[bi]}], bp];  (* inserts zeros \
\
\
*)
    lg = q^(-n alpha * (alpha + m) * writhe) * Apply[Times, \
Map[xiforUqglmn[Uqgl[m,n], RepName, #]&, powerlist]];
    lg = Expand[PowerClean[Factor[lg]]];
    lg = ConvertFromRepresentationToInternalVariables[Uqgl[m, n], RepName, \
lg];
    lg = Evaluate[lg /. {q -> #1, p -> #2}]&;

    lg
  ]) /; (EasyBraidWordQ[BraidWord[L]] && SoberAlgebraQ[Uqgl[m, n]] && (m >= \
\
\
n));

(* ------------------------------------------------------------------- *)

(* Closed 2-braids are Easy, so this is the right place for these functions. \
\
\
*)

Outer[
  (AlreadyLoadedLinksGouldInvariantsOfClosed2Braids[##] = False)&,
  Range[MaxIndexOfLGOfClosed2Braids], Range[MaxIndexOfLGOfClosed2Braids], 1, \
\
\
1
];

LoadLinksGouldInvariantsOfClosed2Braids[m_Integer?Positive, \
n_Integer?Positive] :=
  Module[
    {ifile},

    If[
      Not[AlreadyLoadedLinksGouldInvariantsOfClosed2Braids[m, n]],
      ifile = StringJoin[LGEDirectory, "/data/Closed 2-Braids/Links--Gould \
Invariants/LG",  ToString[m], ToString[n], "/Closed2Braids.lg"];
      Get[ifile];
      AlreadyLoadedLinksGouldInvariantsOfClosed2Braids[m, n] = True
    ];
  ] /; ((m <= MaxIndexOfLGOfClosed2Braids) && (n <= \
MaxIndexOfLGOfClosed2Braids));

LinksGouldInvariant[m_Integer?Positive, n_Integer?Positive, \
Closed2Braid[k_Integer?Positive]] :=
  (
    LoadLinksGouldInvariantsOfClosed2Braids[m, n];
    LinksGouldInvariant[m, n, Closed2Braid[k]]
  ) /;
  (
    (n <= m <= MaxIndexOfLGOfClosed2Braids)
    &&
    (((m <= 4) && (k <= 5)) || (k <= 3))
  );

LinksGouldInvariant[m_Integer?Positive, n_Integer?Positive, \
Closed2Braid[k_Integer?Negative]] :=
  LinksGouldInvariant[m, n, Closed2Braid[-k]];

(* ------------------------------------------------------------------ *)

(* LG^{2,1} for 2-bridge links *)

b1::usage = "
  b1[n] is a utility function for evaluating 
  LinksGouldInvariantOfTwoBridgeLinkGivenByConwayPresentation.
";

b2::usage = "
  b2[n] is a utility function for evaluating 
  LinksGouldInvariantOfTwoBridgeLinkGivenByConwayPresentation.
";

b3::usage = "
  b3[n] is a utility function for evaluating 
  LinksGouldInvariantOfTwoBridgeLinkGivenByConwayPresentation.
";

A::usage = "
  A[n] is a utility function for evaluating 
  LinksGouldInvariantOfTwoBridgeLinkGivenByConwayPresentation.
";

b1[n_Integer] := b1[n] =
  Expand[Factor[(t0^n t1^n - 1)/(t0 t1 - 1)]];

b2[n_Integer] := b2[n] =
  Expand[Factor[
    2n (t0 - 1)(t1 - 1)/(t0 t1 - 1) - b1[n] * (((t0 t1 + 1)(t0 - 1)(t1 - \
1)/(t0 t1 - 1)) + 1)
  ]];

b3[n_Integer] := b3[n] =
  Expand[Factor[(t0 - 1)(t1 - 1) * b1[n] + 1]];

A[n_Integer] := A[n] =
  {
    {b1[n+1], b2[n+1], b3[n+1]},
    {  b1[n],   b2[n],   b3[n]},
    {      0,       1,       0}
  };

ConvertFromEigenvalueVariablesToDualitySymmetricVariables[2, 1, S_] :=
  PowerExpand[S /. {t0 -> p^(-2) q, t1 -> p^2 q}];

ConvertFromDualitySymmetricVariablesToEigenvalueVariables[2, 1, S_] :=
  PowerExpand[S /. {q -> t0^(1/2) t1^(1/2), p -> t0^(-1/4) t1^(1/4)}];

(* Accepts C(a) as input, but works internally on D(b=a/2). *)

LinksGouldInvariantOfTwoBridgeLinkGivenByConwayPresentation[2, 1, \
a_List?VectorQ] :=
  Module[
    {b, lg},

    b = a/2;
    lg = Dot[{0, 1, 0}, Apply[Dot, Map[A, Reverse[b * \
(-1)^Range[Length[b]]]]], {1, 1, 0}];
    lg = ConvertFromEigenvalueVariablesToDualitySymmetricVariables[2, 1, \
Expand[lg]];
    lg = Evaluate[lg /. {q -> #1, p -> #2}]&;

    lg
  ] /; Apply[And, Map[EvenQ, a]];

(* ------------------------------------------------------------------ *)

(* Default: Use the state model method. *)

LinksGouldInvariantAvailableStateModelIndices =
  Thread[{Range[3], 1}];

SoberLinksGouldInvariantAvailableStateModelIndicesQ[m_Integer?Positive, \
n_Integer?Positive] :=
  MemberQ[LinksGouldInvariantAvailableStateModelIndices, {m, n}];

Apply[(AlreadyLoadedStateModelParametersForLinksGouldInvariant[##] = False)&, \
\
\
LinksGouldInvariantAvailableStateModelIndices, 1];

LoadStateModelParametersForLinksGouldInvariant[m_, n_] :=
  Module[
    {fcode, ifile},

    If[
      Not[AlreadyLoadedStateModelParametersForLinksGouldInvariant[m, n]],
      fcode = StringJoin[ToString[m], ToString[n]];
      ifile =
        StringJoin[
          LGEDirectory, "/data/State Model Parameters/Links--Gould \
Invariants/",
          If[
            UseNonunitaryBasisVariablesForLGStateModelParametersFlag,
            "Nonunitary Basis",
            "Unitary Basis"
          ],
          "/LG", fcode, ".smp"
        ];
      Get[ifile];
      AlreadyLoadedStateModelParametersForLinksGouldInvariant[m, n] = True
    ];
  ] /; SoberLinksGouldInvariantAvailableStateModelIndicesQ[m, n];

ConvertToInternalComputationalVariablesForLGRules::usage = "
  ConvertToInternalComputationalVariablesForLGRules[m, n].
";

ConvertFromInternalComputationalVariablesForLGRules::usage = "
  ConvertFromInternalComputationalVariablesForLGRules[m, n].
";

ConvertToInternalComputationalVariablesForLGRules[m_, n_] :=
  If[
    (({m,n}==={2,1}) && UseEigenvalueVariablesForLG21Flag),
    {q -> T0 * T1, p -> T0^(-1/2) * T1^(1/2)},
    {q -> Q^2}
  ] /; SoberLinksGouldInvariantAvailableStateModelIndicesQ[m, n];

ConvertFromInternalComputationalVariablesForLGRules[m_, n_] :=
  If[
    (({m,n}==={2,1}) && UseEigenvalueVariablesForLG21Flag),
    {T0 -> q^(1/2) * p^(-1), T1 -> q^(1/2) * p},
    {Q -> q^(1/2)}
  ] /; SoberLinksGouldInvariantAvailableStateModelIndicesQ[m, n];

ConvertToInternalComputationalVariablesForLG[m_, n_, S_] :=
  PowerExpand[S /. ConvertToInternalComputationalVariablesForLGRules[m, n]] \
\
\
/;
    SoberLinksGouldInvariantAvailableStateModelIndicesQ[m, n];

ConvertFromInternalComputationalVariablesForLG[m_, n_, S_] :=
  PowerExpand[S /. ConvertFromInternalComputationalVariablesForLGRules[m, n]] \
\
\
/;
    SoberLinksGouldInvariantAvailableStateModelIndicesQ[m, n];

LinksGouldInvariantFromAStateModelOnABraidWord[m_, n_, L_] :=
  Module[
    {otli, lg},

    LoadStateModelParametersForLinksGouldInvariant[m, n];
    otli = OpenTangleLinkInvariantOfALinkFromABraidWord[Uqgl[m, n], \
LGIHighestWeight[Uqgl[m,n]], L];
    lg = ConvertFromInternalComputationalVariablesForLG[m, n, otli];
    lg = Evaluate[lg /. {q -> #1, p -> #2}]&;

    lg
  ] /; (SoberLinksGouldInvariantAvailableStateModelIndicesQ[m, n] && \
SoberBraidWordQ[BraidWord[L]]);

LinksGouldInvariant[m_Integer?Positive, n_Integer?Positive, PrimeKnot[c_, \
alt_, i_]] :=
  (
    LoadLinksGouldInvariantsOfHTWPrimeKnots[m, n, c, alt];
    LinksGouldInvariant[m, n, PrimeKnot[c, alt, i]]
  ) /;
  (
    AutoLoadLGInvariantsFlag
    &&
    SoberHTWPrimeKnotQ[PrimeKnot[c, alt, i]]
    &&
    MemberQ[LinksGouldInvariantStoredHTWPrimeKnots[m, n, c, alt], \
PrimeKnot[c, alt, i]]
  );

LinksGouldInvariant[m_Integer?Positive, n_Integer?Positive, L_] :=
  (LinksGouldInvariant[m, n, L] =
  Module[
    {bw, lg, lgqp, lgqpi, lgmqpi, s},

    bw = BraidWord[L];
    lg =
      Which[
        (({m,n}=={2,1}) && TwoBridgeHTWPrimeKnotQ[L]),
        LinksGouldInvariantOfTwoBridgeLinkGivenByConwayPresentation[2, 1, \
ConwayPresentation[L]],

        (EasyBraidWordQ[bw] && (m >= n)),
        LinksGouldInvariantOfEasyBraidWordLink[m, n, L],

        (SoberLinksGouldInvariantAvailableStateModelIndicesQ[m, n] && \
SoberBraidWordQ[bw]),
        LinksGouldInvariantFromAStateModelOnABraidWord[m, n, L]
      ];

(* These next few lines rearrange the terms nicely *)
    If[
      Not[ValidEvaluationOfLinksGouldInvariant[m, n, lg]],
      lgqp = lg[q, p];
      lgqpi = lg[q, 1/p];
      lgmqpi = lg[q, -1/p];
      s = Which[(lgqp===lgqpi), 1, (lgqp===lgmqpi), -1];
      lg = Collect[lgqp, p] /. p^(i_Integer?Negative) :> 0 /. \
{p^(i_Integer:1) :> (p^i + s*1/p^i)};
      lg = Evaluate[lg /. {q -> #1, p -> #2}]&
    ];

    lg
  ]) /;
  (
    (({m,n}=={2,1}) && TwoBridgeHTWPrimeKnotQ[L])
    ||
    (EasyBraidWordQ[BraidWord[L]] && (m >= n))
    ||
    (SoberBraidWordQ[BraidWord[L]] && \
SoberLinksGouldInvariantAvailableStateModelIndicesQ[m, n])
  );

ValidEvaluationOfLinksGouldInvariant[m_Integer?Positive, n_Integer?Positive, \
\
\
poly_] :=
  ((poly[q, p] == poly[q, 1/p]) || (poly[q, p] == poly[q, -1/p])) /; (m >= \
n);

LinksGouldInvariantInEigenvalueVariables[2, 1, L_] :=
  Module[
    {lg},

    lg = Expand[LinksGouldInvariant[2, 1, L][q, p]];
    lg = ConvertFromDualitySymmetricVariablesToEigenvalueVariables[2, 1, lg];
    lg = Collect[lg, t1];
    lg = Evaluate[lg /. {t0 -> #1, t1 -> #2}]&;

    lg
  ] /;
  (
    TwoBridgeHTWPrimeKnotQ[L]
    ||
    EasyBraidWordQ[BraidWord[L]]
    ||
    SoberBraidWordQ[BraidWord[L]]
  );

(* ------------------------------------------------------------------- *)

LinksGouldInvariantStateModelFeasibleHTWPrimeKnots[2, 1, c_, alt_] :=
  HTWPrimeKnots[c, alt] /;
    (SoberHTWPrimeKnotSetQ[c, alt] && (c <= 12));

LinksGouldInvariantStateModelFeasibleHTWPrimeKnots[2, 1, c_, alt_] :=
  (LinksGouldInvariantStateModelFeasibleHTWPrimeKnots[2, 1, c, alt] =
  Module[
    {smax, knotlist},

    smax = {5, 5, 5, 5}[[c-12]];
    knotlist = Flatten[Map[HTWPrimeKnotsOnsStringBraids[c, alt, #]&, Range[1, \
\
\
smax]]];

    knotlist
  ]) /; (SoberHTWPrimeKnotSetQ[c, alt] && (c > 12));

LinksGouldInvariantStateModelFeasibleHTWPrimeKnots[3, 1, c_, alt_] :=
  HTWPrimeKnots[c, alt] /;
    (SoberHTWPrimeKnotSetQ[c, alt] && (c <= 10));

LinksGouldInvariantStateModelFeasibleHTWPrimeKnots[3, 1, c_, alt_] :=
  (LinksGouldInvariantStateModelFeasibleHTWPrimeKnots[3, 1, c, alt] =
  Module[
    {smax, knotlist},

    smax = {3, 3, 2, 2, 2, 2}[[c-10]];
    knotlist = Flatten[Map[HTWPrimeKnotsOnsStringBraids[c, alt, #]&, \
Range[smax]]];

    knotlist
  ]) /; (SoberHTWPrimeKnotSetQ[c, alt] && (c > 10));

LinksGouldInvariantStateModelFeasibleHTWPrimeKnots[m_Integer?Positive, \
n_Integer?Positive, c_] :=
  (LinksGouldInvariantStateModelFeasibleHTWPrimeKnots[m, n, c] =
     Flatten[Map[LinksGouldInvariantStateModelFeasibleHTWPrimeKnots[m, n, c, \
\
\
#]&, {"A", "N"}]]
  ) /; SoberHTWPrimeKnotSetQ[c];

LinksGouldInvariantStateModelFeasibleHTWPrimeKnotsOfUpTocCrossings[m_Integer?\
\
\
Positive, n_Integer?Positive, c_] :=
  (LinksGouldInvariantStateModelFeasibleHTWPrimeKnotsOfUpTocCrossings[m, n, \
\
\
c] =
    Flatten[Map[LinksGouldInvariantStateModelFeasibleHTWPrimeKnots[m, n, #]&, \
\
\
Range[0, c]]]
  ) /; SoberHTWPrimeKnotSetQ[c];

NumberOfLinksGouldInvariantStateModelFeasibleHTWPrimeKnots[m_Integer?Positive,\
\
\
 n_Integer?Positive, c_, alt_] :=
  Length[LinksGouldInvariantStateModelFeasibleHTWPrimeKnots[m, n, c, alt]] /;
    (SoberHTWPrimeKnotSetQ[c, alt]);

NumberOfLinksGouldInvariantStateModelFeasibleHTWPrimeKnots[m_Integer?Positive,\
\
\
 n_Integer?Positive, c_] :=
  Length[LinksGouldInvariantStateModelFeasibleHTWPrimeKnots[m, n, c]] /;
    SoberHTWPrimeKnotSetQ[c];

NumberOfLinksGouldInvariantStateModelFeasibleHTWPrimeKnotsOfUpTocCrossings[m_\
\
\
Integer?Positive, n_Integer?Positive, c_] :=
  Length[LinksGouldInvariantStateModelFeasibleHTWPrimeKnotsOfUpTocCrossings[m,\
\
\
 n, c]] /;
    SoberHTWPrimeKnotSetQ[c];

(* ----------------------------------------------------------------- *)

LinksGouldInvariantFeasibleHTWPrimeKnots[2, 1, c_, alt_] :=
  LinksGouldInvariantStateModelFeasibleHTWPrimeKnots[2, 1, c, alt] /;
    (SoberHTWPrimeKnotSetQ[c, alt] && (c <= 12));

LinksGouldInvariantFeasibleHTWPrimeKnots[2, 1, c_, alt_] :=
  (LinksGouldInvariantFeasibleHTWPrimeKnots[2, 1, c, alt] =
  Union[
    If[(alt=="A"), TwoBridgeHTWPrimeKnots[c], {}],
    LinksGouldInvariantStateModelFeasibleHTWPrimeKnots[2, 1, c, alt]
  ]) /; (SoberHTWPrimeKnotSetQ[c, alt] && (c > 12));

LinksGouldInvariantFeasibleHTWPrimeKnots[m_Integer?Positive, \
n_Integer?Positive, c_, alt_] :=
  LinksGouldInvariantStateModelFeasibleHTWPrimeKnots[m, n, c, alt] /;
    SoberHTWPrimeKnotSetQ[c, alt];

LinksGouldInvariantFeasibleHTWPrimeKnots[m_Integer?Positive, \
n_Integer?Positive, c_] :=
  (LinksGouldInvariantFeasibleHTWPrimeKnots[m, n, c] =
     Flatten[Map[LinksGouldInvariantFeasibleHTWPrimeKnots[m, n, c, #]&, {"A", \
\
\
"N"}]]
  ) /; SoberHTWPrimeKnotSetQ[c];

LinksGouldInvariantFeasibleHTWPrimeKnotsOfUpTocCrossings[m_Integer?Positive, \
\
\
n_Integer?Positive, c_] :=
  (LinksGouldInvariantFeasibleHTWPrimeKnotsOfUpTocCrossings[m, n, c] =
    Flatten[Map[LinksGouldInvariantFeasibleHTWPrimeKnots[m, n, #]&, Range[0, \
\
\
c]]]
  ) /; SoberHTWPrimeKnotSetQ[c];

NumberOfLinksGouldInvariantFeasibleHTWPrimeKnots[m_Integer?Positive, \
n_Integer?Positive, c_, alt_] :=
  Length[LinksGouldInvariantFeasibleHTWPrimeKnots[m, n, c, alt]] /;
    (SoberHTWPrimeKnotSetQ[c, alt]);

NumberOfLinksGouldInvariantFeasibleHTWPrimeKnots[m_Integer?Positive, \
n_Integer?Positive, c_] :=
  Length[LinksGouldInvariantFeasibleHTWPrimeKnots[m, n, c]] /;
    SoberHTWPrimeKnotSetQ[c];

NumberOfLinksGouldInvariantFeasibleHTWPrimeKnotsOfUpTocCrossings[m_Integer?\
Positive, n_Integer?Positive, c_] :=
  Length[LinksGouldInvariantFeasibleHTWPrimeKnotsOfUpTocCrossings[m, n, c]] \
\
\
/;
    SoberHTWPrimeKnotSetQ[c];

(* ----------------------------------------------------------------- *)

LinksGouldInvariantStoredHTWPrimeKnots[m_Integer?Positive, 1, c_, alt_] :=
  HTWPrimeKnots[c, alt] /;
    (SoberHTWPrimeKnotSetQ[c, alt] && (m <= 2) && (c <= 12));

LinksGouldInvariantStoredHTWPrimeKnots[3, 1, c_, alt_] :=
  HTWPrimeKnots[c, alt] /;
    (SoberHTWPrimeKnotSetQ[c, alt] && (c <= 10));

(* This is kludgy code *)

LinksGouldInvariantStoredHTWPrimeKnots[m_Integer?Positive, \
n_Integer?Positive, c_, alt_] :=
  (LinksGouldInvariantStoredHTWPrimeKnots[m, n, c, alt] =
  Module[
    {fcode, ifile, istream, sn, item},

    fcode = StringJoin[If[c < 10, "0", ""], ToString[c], alt];
    ifile = StringJoin[LGEDirectory, "/data/Prime Knots/Links--Gould \
Invariants/LG", ToString[m], ToString[n], "/", fcode, ".lg"];
    istream = OpenRead[ifile];
    sn = {};
    While[
      (item =!= EndOfFile),
      item = Read[istream, Record];
      Skip[istream, Expression];
      sn = {sn, item}
    ];
    Close[istream];
    sn = Complement[Flatten[sn], {EndOfFile}];
    
    If[
      ($VersionNumber >= 5.1),
      sn = Flatten[StringCases[sn, RegularExpression["(P.*\\])\\]"] :> \
ToExpression["$1"]]],
      sn = Map[StringDrop[StringDrop[#, 26], -3] &, sn]; (* an awful kludge \
\
\
*)
      sn = Sort[Map[ToExpression, sn]]
    ];

    sn
  ]) /; (SoberHTWPrimeKnotSetQ[c, alt] && (n==1) && (((m <= 2)&&(c > 12)) || \
\
\
((m==3)&&(c > 10))));

LinksGouldInvariantStoredHTWPrimeKnots[m_Integer?Positive, \
n_Integer?Positive, c_] :=
  Flatten[Thread[LinksGouldInvariantStoredHTWPrimeKnots[m, n, c, \
HTWPrimeKnotAlternatingnessLabelSet]]] /;
    SoberHTWPrimeKnotSetQ[c];

LinksGouldInvariantStoredHTWPrimeKnotsOfUpTocCrossings[m_Integer?Positive, \
n_Integer?Positive, c_] :=
  Flatten[Map[LinksGouldInvariantStoredHTWPrimeKnots[m, n, #]&, Range[0, c]]] \
\
\
/;
    SoberHTWPrimeKnotSetQ[c];


NumberOfLinksGouldInvariantStoredHTWPrimeKnots[m_Integer?Positive, \
n_Integer?Positive, c_, alt_] :=
  Length[LinksGouldInvariantStoredHTWPrimeKnots[m, n, c, alt]] /;
    (SoberHTWPrimeKnotSetQ[c, alt]);

NumberOfLinksGouldInvariantStoredHTWPrimeKnots[m_Integer?Positive, \
n_Integer?Positive, c_] :=
  Length[LinksGouldInvariantStoredHTWPrimeKnots[m, n, c]] /;
    SoberHTWPrimeKnotSetQ[c];

NumberOfLinksGouldInvariantStoredHTWPrimeKnotsOfUpTocCrossings[m_Integer?\
Positive, n_Integer?Positive, c_] :=
  Length[LinksGouldInvariantStoredHTWPrimeKnotsOfUpTocCrossings[m, n, c]] /;
    SoberHTWPrimeKnotSetQ[c];

Outer[
  (AlreadyLoadedListOfLinksGouldInvariantStoredHTWPrimeKnots[Apply[Sequence, \
\
\
#1], Apply[Sequence, #2]] = False)&,
  LinksGouldInvariantAvailableStateModelIndices, \
HTWPrimeKnotCrossingNumberAndAlternatingnessLabelSet, 1, 1
];

(* ----------------------------------------------------------------- *)

OverlapLinksGouldInvariantStoredTwoBridgeHTWPrimeKnots[2, 1, c_] :=
  (Intersection[LinksGouldInvariantStateModelFeasibleHTWPrimeKnots[2, 1, c, \
\
\
"A"], TwoBridgeHTWPrimeKnots[c]]) /;
    SoberHTWPrimeKnotSetQ[c];

NumberOfOverlapLinksGouldInvariantStoredTwoBridgeHTWPrimeKnots[2, 1, c_] :=
  Length[OverlapLinksGouldInvariantStoredTwoBridgeHTWPrimeKnots[2, 1, c]] /;
    SoberHTWPrimeKnotSetQ[c];

(* ----------------------------------------------------------------- *)

LinksGouldInvariantEquivalentLinkTypesQ[m_Integer?Positive, \
n_Integer?Positive, L1_, L2_] :=
  (LinksGouldInvariantEquivalentLinkTypesQ[m, n, L1, L2] =
  Module[
    {p1, p2, p2r},

    p1 = LinksGouldInvariant[m, n, L1];
    p2 = LinksGouldInvariant[m, n, L2];
    p2r = LinksGouldInvariant[m, n, LinkReflection[L2]];

    ((p1 === p2) || (p1 === p2r))
  ])/; ((Head[LinksGouldInvariant[m,n,L1]]=!=LinksGouldInvariant)&&(Head[\
LinksGouldInvariant[m,n,L2]]=!=LinksGouldInvariant));

LinksGouldInvariantCliques[m_Integer?Positive, n_Integer?Positive, \
linklist_List] :=
  Module[
    {rlinklist, linksandreflinks, polys, stpolyslinksandreflinks, cliques},

    rlinklist = Thread[LinkReflection[linklist]];
    linksandreflinks = Join[linklist, rlinklist];
    polys = Map[LinksGouldInvariant[m, n, #]&, linksandreflinks];
    stpolyslinksandreflinks = Sort[Transpose[{polys, linksandreflinks}]];
    cliques = Split[stpolyslinksandreflinks, (#1[[1]] == #2[[1]]) &];
    cliques = cliques[[All, All, 2]];
    cliques = cliques /. LinkReflection[W_] -> W;
    cliques = Map[Union[Sort[#]]&, cliques];
    cliques = Select[cliques, (Length[#] > 1) &];
    cliques = Union[Sort[cliques]];

    cliques
  ] /; Apply[And, Map[(Head[BraidWord[#]]===String)&, linklist]];

LinksGouldInvariantCliquesOfLinksGouldInvariantStoredHTWPrimeKnots[m_Integer?\
\
\
Positive, n_Integer?Positive, c_, alt_] :=
  (LinksGouldInvariantCliquesOfLinksGouldInvariantStoredHTWPrimeKnots[m, n, \
\
\
c, alt] =
    LinksGouldInvariantCliques[m, n, \
LinksGouldInvariantStoredHTWPrimeKnots[m, n, c, alt]]
  ) /; SoberHTWPrimeKnotSetQ[c, alt];

LinksGouldInvariantCliquesOfLinksGouldInvariantStoredHTWPrimeKnots[m_Integer?\
\
\
Positive, n_Integer?Positive, c_] :=
  (LinksGouldInvariantCliquesOfLinksGouldInvariantStoredHTWPrimeKnots[m, n, \
\
\
c] =
    LinksGouldInvariantCliques[m, n, \
LinksGouldInvariantStoredHTWPrimeKnots[m, n, c]]
  ) /; SoberHTWPrimeKnotSetQ[c];

LinksGouldInvariantCliquesOfLinksGouldInvariantStoredHTWPrimeKnotsOfUpTocCross\
\
\
ings[m_Integer?Positive, n_Integer?Positive, c_] :=
  (LinksGouldInvariantCliquesOfLinksGouldInvariantStoredHTWPrimeKnotsOfUpTocCr\
\
\
ossings[m, n, c] =
    LinksGouldInvariantCliques[m, n, \
LinksGouldInvariantStoredHTWPrimeKnotsOfUpTocCrossings[m, n, c]]
  ) /; SoberHTWPrimeKnotSetQ[c];

(* ------------------------------------------------------------------- *)

qPolynomialCoefficientsOfPositivePowersOfpOfLinksGouldInvariant[m_Integer?\
Positive, n_Integer?Positive, L_] :=
  With[
    {poly = LinksGouldInvariant[m, n, L][q, p]},
    Coefficient[poly, p, Range[0, Exponent[poly, p]]]
  ] /; (Head[LinksGouldInvariant[m,n,L]]=!=LinksGouldInvariant);

LinksGouldInvariantCoefficients[m_Integer?Positive, n_Integer?Positive, L_] \
\
\
:=
  (LinksGouldInvariantCoefficients[m, n, L] =
    IntegerArraysFromIntegerCoefficientLaurentPolynomial[LinksGouldInvariant[\
\
\
m, n, L][q, p]]
  ) /; (Head[LinksGouldInvariant[m,n,L]]=!=LinksGouldInvariant);

(* ----------------------------------------------------------------- *)

Outer[
  (AlreadyLoadedLinksGouldInvariantsOfHTWPrimeKnots[Apply[Sequence, #1], \
Apply[Sequence, #2]] = False)&,
  LinksGouldInvariantAvailableStateModelIndices, \
HTWPrimeKnotCrossingNumberAndAlternatingnessLabelSet, 1, 1
];

LoadLinksGouldInvariantsOfHTWPrimeKnots[m_Integer?Positive, \
n_Integer?Positive, c_, alt_] :=
  Module[
    {fcode, ifile},

    If[
      Not[AlreadyLoadedLinksGouldInvariantsOfHTWPrimeKnots[m, n, c, alt]],
      fcode = StringJoin[If[c < 10, "0", ""], ToString[c], alt];
      ifile = StringJoin[LGEDirectory, "/data/Prime Knots/Links--Gould \
Invariants/LG", ToString[m], ToString[n], "/", fcode, ".lg"];
      If[((c > 12) || WriteAllLoadingMessagesFlag), WriteString["stdout", \
"Loading: ", ifile, " ... "]];
      Get[ifile];
      If[((c > 12) || WriteAllLoadingMessagesFlag), WriteString["stdout", \
"done\n"]];
      AlreadyLoadedLinksGouldInvariantsOfHTWPrimeKnots[m, n, c, alt] = True
    ];
  ] /; SoberHTWPrimeKnotSetQ[c, alt];

StoreLinksGouldInvariantsOfHTWPrimeKnots[m_Integer?Positive, \
n_Integer?Positive, c_, alt_] :=
  Module[
    {fcode, ofile, ostream},

    fcode = StringJoin[If[c < 10, "0", ""], ToString[c], alt];
    ofile = StringJoin[LGEDirectory, "/data/Prime Knots/Links--Gould \
Invariants/LG", ToString[m], ToString[n], "/New ", fcode, ".lg"];
    ostream = OpenWrite[ofile, PageWidth -> 80];
    Map[LinksGouldInvariant[m, n, #]&, HTWPrimeKnots[c, alt]];
    Map[
      (
        WriteString[ostream, "LinksGouldInvariant[", m, ", ", n, ", ", \
InputForm[#], "] = \n  "];
        Write[ostream, LinksGouldInvariant[m, n, #]];
        WriteString[ostream, "\n"]
      )&,
      HTWPrimeKnots[c, alt]
    ];
    Close[ostream];
    WriteString["stdout", "done\n"];
  ] /; SoberHTWPrimeKnotSetQ[c, alt];

StoreLinksGouldInvariantsOfHTWPrimeKnotsOfUpTocCrossings[m_Integer?Positive, \
\
\
n_Integer?Positive, c_] :=
  Scan[StoreLinksGouldInvariantsOfHTWPrimeKnots[m, n, #]&, Range[0, c]] /;
    SoberHTWPrimeKnotSetQ[c];

(* ----------------------------------------------------------------- *)

KauffmanLinksGouldPolynomial[L_] :=
  Evaluate[LinksGouldInvariant[2, 1, L][q, 1] /. q -> #]& /;
    (Head[LinksGouldInvariant[2, 1, L]]=!=LinksGouldInvariant);

(* ----------------------------------------------------------------- *)

(* End[]; *)

EndPackage[];



