(* -----------------------------------------------------------------

  Two-Bridge Knots.m

  Any two-bridge link may be expressed (nonuniquely) as a Conway presentation \
\

  C(a), where a is a list of even nonzero integers. If a is even-length 
  (respectively, odd-length), then C(a) is a true knot (respectively, a 
  2-component link). Here, we are only consider true knots. Moreover, C(-a) \
\
is 
  the reflection of C(a) and C(-a^r) is the inversion of a (where a^r is the 
  reverse of a). We only consider knot types, so take only one of the 4 cases \
\

  C(a), C(-a), C(a^r), and C(-a^r). See "The 2-bridge knots of up to 16 
  crossings" -- David De Wit.

  Functions pertaining to the origin of the data are not included.

  ------------------------------------------------------------------ *)

BeginPackage["LinksGouldExplorer`"];

MinimumCrossingNumberOfTwoBridgeHTWPrimeKnots::usage = "
  MinimumCrossingNumberOfTwoBridgeHTWPrimeKnots.
";

MaximumCrossingNumberOfTwoBridgeHTWPrimeKnots::usage = "
  MaximumCrossingNumberOfTwoBridgeHTWPrimeKnots.
";

NonEmptyTwoBridgeHTWPrimeKnotCrossingNumberAndAlternatingnessLabelSet::usage \
\
= "
  NonEmptyTwoBridgeHTWPrimeKnotCrossingNumberAndAlternatingnessLabelSet is \
the 
  set of labels {c, alt}; such that the sets TwoBridgeHTWPrimeKnots[c, alt] 
  each contain some nonzero number of 2-bridge knots.
";

EmptyTwoBridgeHTWPrimeKnotCrossingNumberAndAlternatingnessLabelSet::usage = "
  EmptyTwoBridgeHTWPrimeKnotCrossingNumberAndAlternatingnessLabelSet is the 
  set of labels {c,alt}; such that the sets TwoBridgeHTWPrimeKnots[c, alt] 
  each contain no 2-bridge knots.
";

TwoBridgeHTWPrimeKnotQ::usage = "
  TwoBridgeHTWPrimeKnotQ[PrimeKnot[c, alt, i]] is True if PrimeKnot[c, alt, \
\
i] 
  is a 2-bridge knot.
";

TwoBridgeHTWPrimeKnots::usage = "
  TwoBridgeHTWPrimeKnots[c] is the set of 2-bridge prime knots of c \
crossings.
";

TwoBridgeHTWPrimeKnotsOfUpTocCrossings::usage = "
  TwoBridgeHTWPrimeKnotsOfUpTocCrossings[c] is the set of 2-bridge prime \
knots 
  of up to c crossings.
";


TwoBridgeHTWPrimeKnotsOnsStringBraids::usage = "
  TwoBridgeHTWPrimeKnotsOnsStringBraids[c, s] is the set of 2-bridge prime 
  knots of c crossings with K2K'-braids on s strings.
";

AlreadyLoadedListsOfTwoBridgeHTWPrimeKnotsOnsStringBraids::usage = "
  AlreadyLoadedListsOfTwoBridgeHTWPrimeKnotsOnsStringBraids[c] is a flag.
";

LoadListsOfTwoBridgeHTWPrimeKnotsOnsStringBraids::usage = "
  LoadListsOfTwoBridgeHTWPrimeKnotsOnsStringBraids[c] loads the list of 
  correspondences between 2-bridge knots and HTW prime knots of c crossings.
";

NumberOfTwoBridgeHTWPrimeKnots::usage = "
  NumberOfTwoBridgeHTWPrimeKnots[c, alt] is the number of 2-bridge knots in 
  TwoBridgeHTWPrimeKnots[c, alt].
";

NumberOfTwoBridgeHTWPrimeKnotsOfUpTocCrossings::usage = "
  NumberOfTwoBridgeHTWPrimeKnotsOfUpTocCrossings[c] is the total number of 
  2-bridge HTW prime knots of crossing number at most c.
";


TwoBridgeHTWPrimeKnotsOnsStringBraids::usage = "
  TwoBridgeHTWPrimeKnotsOnsStringBraids[c, alt, s].
";

TwoBridgeHTWPrimeKnotsOfUpTocCrossingsOnsStringBraids::usage = "
  TwoBridgeHTWPrimeKnotsOfUpTocCrossingsOnsStringBraids[c, s].
";

NumberOfTwoBridgeHTWPrimeKnotsOnsStringBraids::usage = "
  NumberOfTwoBridgeHTWPrimeKnotsOnsStringBraids[c, alt, s].
";

NumberOfTwoBridgeHTWPrimeKnotsOfUpTocCrossingsOnsStringBraids::usage = "
  NumberOfTwoBridgeHTWPrimeKnotsOfUpTocCrossingsOnsStringBraids[c, s].
";


ConwayPresentation::usage = "
  ConwayPresentation[PrimeKnot[c, \"A\", i]] is a list a such that the 
  2-bridge knot with Conway presentation C(a) is of knot type 
  PrimeKnot[c, \"A\", i]. This is only meaningful if PrimeKnot[c, \"A\", i] 
  actually is a 2-bridge knot!
";

AlreadyLoadedConwayPresentationsOfTwoBridgeHTWPrimeKnots::usage = "
  AlreadyLoadedConwayPresentationsOfTwoBridgeHTWPrimeKnots[c] is a flag.
";

LoadConwayPresentationsOfTwoBridgeHTWPrimeKnots::usage = "
  LoadConwayPresentationsOfTwoBridgeHTWPrimeKnots[c].
";


SlopeOfTwoBridgeLink::usage = "
  SlopeOfTwoBridgeLink[L] is the rational number that is the slope of the 
  2-bridge link L for which a Conway presentation is known.
";

DTCodeOfTwoBridgeLink::usage = "
  DTCodeOfTwoBridgeLink[L] is a DT code corresponding to the 2-bridge link L 
  for which a Conway presentation is known.
";

BraidWordOfTwoBridgeLink::usage = "
  BraidWordOfTwoBridgeLink[L] is a braid word corresponding to the 2-bridge 
  link L for which a Conway presentation is known.
";

(* ----------------------------------------------------------------- *)

(* Begin["`Private`"]; *)

MinimumCrossingNumberOfTwoBridgeHTWPrimeKnots = 3;

MaximumCrossingNumberOfTwoBridgeHTWPrimeKnots = \
MaximumCrossingNumberOfHTWPrimeKnots;

NonEmptyTwoBridgeHTWPrimeKnotCrossingNumberAndAlternatingnessLabelSet =
  Thread[{Range[MinimumCrossingNumberOfTwoBridgeHTWPrimeKnots, \
MaximumCrossingNumberOfTwoBridgeHTWPrimeKnots], "A"}];

EmptyTwoBridgeHTWPrimeKnotCrossingNumberAndAlternatingnessLabelSet =
  Complement[
    HTWPrimeKnotCrossingNumberAndAlternatingnessLabelSet,
    NonEmptyTwoBridgeHTWPrimeKnotCrossingNumberAndAlternatingnessLabelSet
  ];

Apply[(AlreadyLoadedListsOfTwoBridgeHTWPrimeKnots[##] = True)&, \
EmptyTwoBridgeHTWPrimeKnotCrossingNumberAndAlternatingnessLabelSet, 1];
Apply[(AlreadyLoadedListsOfTwoBridgeHTWPrimeKnots[##] = False)&, \
NonEmptyTwoBridgeHTWPrimeKnotCrossingNumberAndAlternatingnessLabelSet, 1];
Apply[(TwoBridgeHTWPrimeKnots[##] = {})&, \
EmptyTwoBridgeHTWPrimeKnotCrossingNumberAndAlternatingnessLabelSet, 1];

LoadListsOfTwoBridgeHTWPrimeKnots[c_, alt_] :=
  Module[
    {ifile, fcode},

    If[
      Not[AlreadyLoadedListsOfTwoBridgeHTWPrimeKnots[c, alt]],
      fcode = StringJoin[If[c < 10, "0", ""], ToString[c], alt];
      ifile = StringJoin[LGEDirectory, "/data/Prime Knots/Two-Bridge Knots/", \
\
fcode, ".tb"];
      If[((c > 12) || WriteAllLoadingMessagesFlag), WriteString["stdout", \
"Loading: ", ifile, " ... "]];
      Get[ifile];
      If[((c > 12) || WriteAllLoadingMessagesFlag), WriteString["stdout", \
"done\n"]];
      AlreadyLoadedListsOfTwoBridgeHTWPrimeKnots[c, alt] = True
    ];
  ] /; SoberHTWPrimeKnotSetQ[c, alt];

TwoBridgeHTWPrimeKnots[c_, alt_] :=
  (
    LoadListsOfTwoBridgeHTWPrimeKnots[c, alt];
    TwoBridgeHTWPrimeKnots[c, alt]
  ) /; SoberHTWPrimeKnotSetQ[c, alt];

TwoBridgeHTWPrimeKnots[c_] :=
  Flatten[Thread[TwoBridgeHTWPrimeKnots[c, \
HTWPrimeKnotAlternatingnessLabelSet]]] /;
    SoberHTWPrimeKnotSetQ[c];

TwoBridgeHTWPrimeKnotsOfUpTocCrossings[c_] :=
  Flatten[Map[TwoBridgeHTWPrimeKnots, Range[0, c]]] /;
    SoberHTWPrimeKnotSetQ[c];

NumberOfTwoBridgeHTWPrimeKnots[c_, alt_] :=
  Length[TwoBridgeHTWPrimeKnots[c, alt]] /;
    SoberHTWPrimeKnotSetQ[c, alt];

NumberOfTwoBridgeHTWPrimeKnots[c_] :=
  Length[TwoBridgeHTWPrimeKnots[c]] /;
    SoberHTWPrimeKnotSetQ[c];

NumberOfTwoBridgeHTWPrimeKnotsOfUpTocCrossings[c_] :=
  Length[TwoBridgeHTWPrimeKnotsOfUpTocCrossings[c]] /;
    SoberHTWPrimeKnotSetQ[c];

(* ----------------------------------------------------------------- *)

TwoBridgeHTWPrimeKnotQ[PrimeKnot[c_, alt_, i_]] :=
  (TwoBridgeHTWPrimeKnotQ[PrimeKnot[c, alt, i]] =
    MemberQ[TwoBridgeHTWPrimeKnots[c, alt], PrimeKnot[c, alt, i]]
  ) /; SoberHTWPrimeKnotQ[PrimeKnot[c, alt, i]];

TwoBridgeHTWPrimeKnotQ[L_] :=
  False /; (Head[L]=!=PrimeKnot);

(* ----------------------------------------------------------------- *)

TwoBridgeHTWPrimeKnotsOnsStringBraids[c_, alt_, s_Integer?Positive] :=
  Intersection[TwoBridgeHTWPrimeKnots[c, alt], \
HTWPrimeKnotsOnsStringBraids[c, alt, s]] /;
    (SoberHTWPrimeKnotSetQ[c, alt] && (s < c));

(* ----------------------------------------------------------------- *)

Apply[(AlreadyLoadedConwayPresentationsOfTwoBridgeHTWPrimeKnots[##] = True)&, \
\
EmptyTwoBridgeHTWPrimeKnotCrossingNumberAndAlternatingnessLabelSet, 1];
Apply[(AlreadyLoadedConwayPresentationsOfTwoBridgeHTWPrimeKnots[##] = \
False)&, NonEmptyTwoBridgeHTWPrimeKnotCrossingNumberAndAlternatingnessLabelSet\
\
, 1];

LoadConwayPresentationsOfTwoBridgeHTWPrimeKnots[c_, alt_] :=
  Module[
    {ifile, fcode},

    If[
      Not[AlreadyLoadedConwayPresentationsOfTwoBridgeHTWPrimeKnots[c, alt]],
      fcode = StringJoin[If[c < 10, "0", ""], ToString[c], alt];
      ifile = StringJoin[LGEDirectory, "/data/Prime Knots/Conway \
Presentations/", fcode, ".cp"];
      If[((c > 12) || WriteAllLoadingMessagesFlag), WriteString["stdout", \
"Loading: ", ifile, " ... "]];
      Get[ifile];
      If[((c > 12) || WriteAllLoadingMessagesFlag), WriteString["stdout", \
"done\n"]];
      AlreadyLoadedConwayPresentationsOfTwoBridgeHTWPrimeKnots[c, alt] = True
    ];
  ] /; SoberHTWPrimeKnotSetQ[c, alt];

(* ----------------------------------------------------------------- *)

ConwayPresentation[PrimeKnot[c_, alt_, i_]] :=
  (
    LoadConwayPresentationsOfTwoBridgeHTWPrimeKnots[c, alt];
    ConwayPresentation[PrimeKnot[c, alt, i]]
  ) /; TwoBridgeHTWPrimeKnotQ[PrimeKnot[c, alt, i]];

(* ----------------------------------------------------------------- *)

TwoBridgeHTWPrimeKnotsOnsStringBraids[c_, alt_, s_Integer?Positive] :=
  (TwoBridgeHTWPrimeKnotsOnsStringBraids[c, alt, s] =
    Select[TwoBridgeHTWPrimeKnots[c, alt], (StringIndex[#]==s)&]
  ) /;
  (
    SoberHTWPrimeKnotSetQ[c, alt]
    &&
    (s <= MaximumCrossingNumberOfHTWPrimeKnots)
  );

TwoBridgeHTWPrimeKnotsOnsStringBraids[c_, s_Integer?Positive] :=
  (TwoBridgeHTWPrimeKnotsOnsStringBraids[c, s] =
    Select[TwoBridgeHTWPrimeKnots[c], (StringIndex[#]==s)&]
  ) /;
  (
    SoberHTWPrimeKnotSetQ[c]
    &&
    (s <= MaximumCrossingNumberOfHTWPrimeKnots)
  );

TwoBridgeHTWPrimeKnotsOfUpTocCrossingsOnsStringBraids[c_, s_Integer?Positive] \
\
:=
  (TwoBridgeHTWPrimeKnotsOfUpTocCrossingsOnsStringBraids[c, s] =
    Select[TwoBridgeHTWPrimeKnotsOfUpTocCrossings[c], (StringIndex[#]==s)&]
  ) /;
  (
    SoberHTWPrimeKnotSetQ[c]
    &&
    (s <= MaximumCrossingNumberOfHTWPrimeKnots)
  );

NumberOfTwoBridgeHTWPrimeKnotsOnsStringBraids[c_, alt_, s_Integer?Positive] \
\
:=
  Length[TwoBridgeHTWPrimeKnotsOnsStringBraids[c, alt, s]] /;
  (
    SoberHTWPrimeKnotSetQ[c, alt]
    &&
    (s <= MaximumCrossingNumberOfHTWPrimeKnots)
  );

NumberOfTwoBridgeHTWPrimeKnotsOnsStringBraids[c_, s_Integer?Positive] :=
  Length[TwoBridgeHTWPrimeKnotsOnsStringBraids[c, s]] /;
  (
    SoberHTWPrimeKnotSetQ[c]
    &&
    (s <= MaximumCrossingNumberOfHTWPrimeKnots)
  );

NumberOfTwoBridgeHTWPrimeKnotsOfUpTocCrossingsOnsStringBraids[c_, \
s_Integer?Positive] :=
  Length[TwoBridgeHTWPrimeKnotsOfUpTocCrossingsOnsStringBraids[c, s]] /;
  (
    SoberHTWPrimeKnotSetQ[c]
    &&
    (s <= MaximumCrossingNumberOfHTWPrimeKnots)
  );

(* ----------------------------------------------------------------- *)

SlopeOfConwayPresentation::usage = "
  SlopeOfConwayPresentation[a] is the slope of a 2-bridge link with Conway 
  presentation C(a).
";

SlopeOfConwayPresentation[a_List?VectorQ] :=
  1/(a[[1]] + SlopeOfConwayPresentation[Drop[a, 1]]) /; 
    (Length[a] > 1);

SlopeOfConwayPresentation[a_List?VectorQ] :=
  1/a[[1]] /; (Length[a]==1);
  
SlopeOfTwoBridgeLink[L_] :=
  SlopeOfConwayPresentation[ConwayPresentation[L]] /;
    (Head[ConwayPresentation[L]]===List);

(* ----------------------------------------------------------------- *)

DTCodeOfConwayPresentation::usage = "
  DTCodeOfConwayPresentation[a] is a DT code for a 2-bridge link with Conway 
  presentation C(a). It is only valid where a is an even-length sequence
  of even integers.
";

(*
  This appears to be valid even when some (but not all) of the a_i are 0,
  but we disallow these as C(a_1,a_2,a_3,0) is not actually C(a_1,a_2,a_3).
*)

DTCodeOfConwayPresentation[a_List?VectorQ] :=
  Module[
    {n, A, s, r, r2, ro, re, FF, Oo, Oe, O, k, m, o, tt, dtc},

    n = Length[a]; A = Abs[a]; s = Sign[a];
    r = Range[n]; r2 = Range[n/2];
    ro = Range[1, n - 1, 2]; re = Range[2, n, 2];
    FF =  Apply[Plus, A[[ro]]] + Map[Apply[Plus, Drop[A, # - 1]]&, r];
    Oo = Map[Apply[Plus, Take[A[[ro]], # - 1]]&, r2];
    Oe = FF[[1]] + Map[Apply[Plus, Take[A[[re]], # - 1]]&, r2];
    O = Flatten[Transpose[{Oo, Oe}]];
    k = Map[Range, A];
    m = (FF + 1 - k) * s^k;
    o = (O + k) * s^(k + 1);
    tt = Flatten[Transpose[{Flatten[m], Flatten[o]}]];
    dtc = Transpose[Sort[Transpose[{Select[tt, OddQ], Select[tt, \
EvenQ]}]]][[2]];

    dtc
  ] /; (EvenQ[Length[a]] && Apply[And, Map[EvenQ, a]] && Apply[And, Map[(# \
=!= 0)&, a]]);

DTCodeOfTwoBridgeLink[L_PrimeKnot] := 
  DTCodeOfConwayPresentation[ConwayPresentation[L]] /;
    (TwoBridgeHTWPrimeKnotQ[L] && Head[ConwayPresentation[L]]===List);

(* ------------------------------------------------------------------- *)

BraidWordOfConwayPresentation::usage = "
  BraidWordOfConwayPresentation[a] is a braid word for a 2-bridge link with 
  Conway presentation C(a). It is only valid where a is a length-2 sequence \
\
of 
  even integers, but could in principle be expanded to longer sequences.
";

BraidWordOfConwayPresentation[a_List?VectorQ] :=
  Module[
    {n, A, s, O, S, alpha, beta, b1, b2, b1, b2, gamma, bw},

    n = Length[a];
    A = Abs[a/2];
    s = Sign[a];
    O = Map[Apply[Plus, Take[A, #]]&, Range[n] - 1];
    S = (-1)^Range[n] * s;
    alpha = (O +  Map[Reverse[Range[#]]&, A]) * S;
    beta = -(O + Map[Reverse[Range[#]]&, A - 1]) * S;
    {b1, b2} = alpha; {b1, b2} = beta;
    gamma = Join[b1, b2, Reverse[b1], b1, Reverse[b2], b2];
    bw = BraidWordFromSignedIntegerBraid[gamma];

    bw
  ] /; ((Length[a]==2) && Apply[And, Map[EvenQ, a]] && Apply[And, Map[(# =!= \
\
0)&, a]]);

BraidWordOfTwoBridgeLink[L_PrimeKnot] := 
  BraidWordOfConwayPresentation[ConwayPresentation[L]] /;
    (TwoBridgeHTWPrimeKnotQ[L] && Head[ConwayPresentation[L]]===List);

(* ----------------------------------------------------------------- *)

(* End[]; *)

EndPackage[];


