(*
  Uqglmn.m

  U_q[gl(m|n)] regarded as an associative Hopf superalgebra.

   ----------------------------------------------------------------- *)

BeginPackage["LinksGouldExplorer`"];

OperatorQ::usage = "
  OperatorQ[W] is True if W is some `operator', in particular an algebra 
  generator which some functions treat differently from a scalar.
";

AltGenerator::usage = "
  AltGenerator[Uqgl[m, n], a, b, N] is the U_q[gl(m|n)] alternative 
  `generator' mathbf{E}^a_b or $K_a^N.
";

qSub::usage = "
  qSub[Uqgl[m, n], a] is the scalar q_a = q^{(-)^{[a]}}.
";

DeltaSub::usage = "
  DeltaSub[Uqgl[m, n], a] is the scalar (-)^{[a]}(q-1/q), which is 
  equivalently (q_a)^{+1} - (q_a)^{-1}.
";

SimpleAlgebraGeneratorQ::usage = "
  SimpleAlgebraGeneratorQ[G] if G is a simple generator of a valid algebra.
";

NonSimpleAlgebraGeneratorQ::usage = "
  NonSimpleAlgebraGeneratorQ[G] if G is a nonsimple generator of a valid 
  algebra.
";

ExpandNonSimpleGeneratorOnce::usage = "
  ExpandNonSimpleGeneratorOnce[Generator[Uqgl[m, n], a, b, 1]] expands the 
  nonsimple generator E^a_b once through an index c roughly in the middle of \
a 
  and b. That is, it uses a bisection strategy. The resulting product is not 
  generally normally ordered.
";

ExpandAllNonSimpleGenerators::usage = "
  ExpandAllNonSimpleGenerators[S] recursively applies 
  ExpandNonSimpleGeneratorOnce until the expression S contains only simple 
  generators. The resulting product is not generally normally ordered.
";

ApplyZhangsLemmaOnce::usage = "
  ApplyZhangsLemmaOnce[G] expands an alternative generator G once using 
  Zhang's Lemma.
";

ExpandAllAltGeneratorsToStandardGeneratorsUsingZhangsLemma::usage = "
  ExpandAllAltGeneratorsToStandardGeneratorsUsingZhangsLemma[S] recursively 
  applies ApplyZhangsLemmaOnce until the expression S contains only standard 
  (but not necessarily simple) generators. The resulting product is not 
  generally normally ordered.
";

(* ------------------------------------------------------------------ *)

(* Begin["`Private`"]; *)

(*
  Determine if an expression contains things that shouldn't commute, and then
  describe some properties of the noncommutative product.
*)

OperatorQ[x_NonCommutativeMultiply] := Map[OperatorQ, Apply[Or, x]];
OperatorQ[x_Times] := Map[OperatorQ, Apply[Or, x]];
OperatorQ[x_Plus] := Map[OperatorQ, Apply[Or, x]];

OperatorQ[x_Generator] := OperatorQ[x] = True;
OperatorQ[x_AltGenerator] := OperatorQ[x] = True;

OperatorQ[ket[_]] := True;
OperatorQ[bra[_]] := True;

OperatorQ[_] := False;

Attributes[NonCommutativeMultiply] = {Flat, OneIdentity};

Unprotect[NonCommutativeMultiply];
  W_ ** Y_ := W * Y /; (!OperatorQ[W] || !OperatorQ[Y]);
  (c_ W_) ** Y_ := c * (W ** Y) /; !OperatorQ[c];
  W_ ** (c_ Y_) := c * (W ** Y) /; !OperatorQ[c];
Protect[NonCommutativeMultiply];

(* ----------------------------------------------------------------- *)

Rank[Uqgl[m_, n_]] :=
  Rank[gl[m, n]] /;
    SoberAlgebraQ[Uqgl[m,n]];

SoberAlgebraQ[Uqgl[m_, n_]] :=
  SoberAlgebraQ[gl[m, n]];

SoberAlgebraIndexQ[Uqgl[m_, n_], a_] :=
  SoberAlgebraIndexQ[gl[m, n], a];

GradingOfAlgebraIndex[Uqgl[m_, n_], a_] :=
  GradingOfAlgebraIndex[gl[m, n], a] /;
    SoberAlgebraIndexQ[Uqgl[m, n], a];

SoberAlgebraGeneratorQ[Generator[Uqgl[m_, n_], a_, b_, N_]] :=
  (SoberAlgebraGeneratorQ[Generator[gl[m,n], a, b]] && NumberQ[N] && \
((N==1)||(a==b)));

SoberAlgebraGeneratorQ[AltGenerator[Uqgl[m_, n_], a_, b_, N_]] :=
  SoberAlgebraGeneratorQ[Generator[Uqgl[m,n], a, b, N]];

GradingOfAlgebraGenerator[Generator[Uqgl[m_, n_], a_, b_, N_]] :=
  GradingOfAlgebraGenerator[Generator[gl[m,n], a, b]] /;
    SoberAlgebraGeneratorQ[Generator[Uqgl[m,n], a, b, N]];

LiftingOfAlgebraGenerator[Generator[Uqgl[m_, n_], a_, b_, N_]] :=
  LiftingOfAlgebraGenerator[Generator[gl[m,n], a, b]] /;
    SoberAlgebraGeneratorQ[Generator[Uqgl[m,n], a, b, N]];

(* ----------------------------------------------------------------- *)

GeneratorSet[Uqgl[m_, n_], gradings_List?VectorQ, liftings_List?VectorQ] :=
  Module[
    {gs},

    gs = Flatten[Table[Generator[Uqgl[m,n], a, b, 1], {a, m+n}, {b, m+n}]];
    gs = Select[gs, MemberQ[gradings, GradingOfAlgebraGenerator[#]]&];
    gs = Select[gs, MemberQ[liftings, LiftingOfAlgebraGenerator[#]]&];
    gs = Sort[gs, OrderedGeneratorsQ[#1, #2]&];

    gs
  ] /; SoberAlgebraQ[Uqgl[m,n]];

(* ----------------------------------------------------------------- *)

HermitianConjugate[Uqgl[m_, n_], Generator[Uqgl[m_, n_], a_Integer?Positive, \
a_, N_]] :=
  Generator[Uqgl[m,n], a, a, N] /;
    SoberAlgebraGeneratorQ[Generator[Uqgl[m,n], a, a, N]];

HermitianConjugate[Uqgl[m_, n_], Generator[Uqgl[m_, n_], a_, b_, N_]] :=
  AltGenerator[Uqgl[m,n], b, a, N] /;
    SoberAlgebraGeneratorQ[Generator[Uqgl[m,n], a, b, N]];

HermitianConjugate[Uqgl[m_, n_], AltGenerator[Uqgl[m_, n_], a_, b_, N_]] :=
  Generator[Uqgl[m,n], b, a, N] /;
    SoberAlgebraGeneratorQ[Generator[Uqgl[m,n], a, b, N]];

(* ----------------------------------------------------------------- *)

qSub[Uqgl[m_, n_], a_] :=
  q^((-1)^GradingOfAlgebraIndex[Uqgl[m,n], a]) /;
    SoberAlgebraIndexQ[Uqgl[m, n], a];

DeltaSub[Uqgl[m_, n_], a_] :=
  (-1)^GradingOfAlgebraIndex[Uqgl[m,n], a] * (q - 1/q) /;
    SoberAlgebraIndexQ[Uqgl[m, n], a];

(* ----------------------------------------------------------------- *)

SimpleAlgebraGeneratorQ[Generator[Uqgl[m_, n_], a_, b_, N_]] :=
  (Abs[a-b] < 2) /; SoberAlgebraGeneratorQ[Generator[Uqgl[m, n], a, b, N]];

SimpleAlgebraGeneratorQ[AltGenerator[Uqgl[m_, n_], a_, b_, N_]] :=
  SimpleAlgebraGeneratorQ[Generator[Uqgl[m, n], a, b, N]] /;
    SoberAlgebraGeneratorQ[Generator[Uqgl[m, n], a, b, N]];

NonSimpleAlgebraGeneratorQ[Generator[Uqgl[m_, n_], a_, b_, N_]] :=
  (SoberAlgebraGeneratorQ[Generator[Uqgl[m, n], a, b, N]] && \
!SimpleAlgebraGeneratorQ[Generator[Uqgl[m,n], a, b, 1]]);

NonSimpleAlgebraGeneratorQ[AltGenerator[Uqgl[m_, n_], a_, b_, N_]] :=
  (SoberAlgebraGeneratorQ[Generator[Uqgl[m, n], a, b, N]] && \
!SimpleAlgebraGeneratorQ[AltGenerator[Uqgl[m,n], a, b, 1]]);

(* ----------------------------------------------------------------- *)

ExpandNonSimpleGeneratorOnce[Generator[Uqgl[m_, n_], a_, b_, 1]] :=
  Module[
    {c, Gac, Gcb, eG},

    c = Floor[(a+b)/2];
    Gac = Generator[Uqgl[m, n], a, c, 1];
    Gcb = Generator[Uqgl[m, n], c, b, 1];
    eG = Gac ** Gcb - qSub[Uqgl[m,n], c]^Sign[a-b] * Gcb ** Gac;

    eG
  ] /; NonSimpleAlgebraGeneratorQ[Generator[Uqgl[m,n], a, b, 1]];

ExpandNonSimpleGeneratorOnce[AltGenerator[Uqgl[m_, n_], a_, b_, 1]] :=
  Generator[Uqgl[m,n], a, b, 1] /; \
SimpleAlgebraGeneratorQ[Generator[Uqgl[m,n], a, b, 1]];

ExpandNonSimpleGeneratorOnce[AltGenerator[Uqgl[m_, n_], a_, b_, 1]] :=
  Module[
    {c, Gac, Gcb, eG},

    c = Floor[(a+b)/2];
    Gac = AltGenerator[Uqgl[m, n], a, c, 1];
    Gcb = AltGenerator[Uqgl[m, n], c, b, 1];
    eG = Gac ** Gcb - qSub[Uqgl[m, n], c]^(-Sign[a-b]) * Gcb ** Gac;

    eG
  ] /; NonSimpleAlgebraGeneratorQ[AltGenerator[Uqgl[m, n],a, b, 1]];

ExpandAllNonSimpleGenerators[S_] :=
  Module[
    {Sold, Snew},

    Sold = S;
    Snew =
      Sold //.
      {
        Generator[Algebra_, a_, b_, 1] :> \
ExpandNonSimpleGeneratorOnce[Generator[Algebra, a, b, 1]] /;
          NonSimpleAlgebraGeneratorQ[Generator[Algebra, a, b, 1]],
        AltGenerator[Algebra_, a_, b_, 1] :> \
ExpandNonSimpleAltGeneratorOnce[AltGenerator[Algebra, a, b,1]] /;
          NonSimpleAlgebraGeneratorQ[AltGenerator[Algebra, a, b, 1]]
      };
    While[
      (Snew =!= Sold),
      Sold = Snew;
      Snew = MapAll[Distribute[#, Plus, NonCommutativeMultiply]&, Snew];
      Snew = Collect[Snew, NonCommutativeMultiply[_]]
    ];

    Snew
  ] /; SoberAlgebraQ[Algebra];

(* ----------------------------------------------------------------- *)

ApplyZhangsLemmaOnce[AltGenerator[Uqgl[m_, n_], a_, b_, 1]] :=
  (
    Generator[Uqgl[m, n], a, b, 1]
      +
      Sign[a - b] *
      Sum[
        DeltaSub[Uqgl[m, n], c] * Generator[Uqgl[m, n], c, b, 1] ** \
AltGenerator[Uqgl[m, n], a, c, 1],
        {c, Min[a, b] + 1, Max[a, b] - 1}
      ]
  ) /; NonSimpleAlgebraGeneratorQ[AltGenerator[Uqgl[m, n], a, b, 1]];

ApplyZhangsLemmaOnce[AltGenerator[Uqgl[m_, n_], a_, b_, 1]] :=
  Generator[Uqgl[m, n], a, b, 1] /;
    SimpleAlgebraGeneratorQ[AltGenerator[Uqgl[m, n], a, b, 1]];

ExpandAllAltGeneratorsToStandardGeneratorsUsingZhangsLemma[S_] :=
  Module[
    {Sold, Snew},

    Sold = S;
    Snew =
      Sold //.
      {
        AltGenerator[Uqgl[m_, n_], a_, b_, 1] :> \
ApplyZhangsLemmaOnce[AltGenerator[Uqgl[m, n], a, b, 1]]
      };
    While[
      (Snew =!= Sold),
      Sold = Snew;
      Snew = MapAll[Distribute[#, Plus, NonCommutativeMultiply]&, Snew];
      Snew = Collect[Snew, NonCommutativeMultiply[_]]
    ];

    Snew
  ] /; SoberAlgebraQ[Algebra];

(* ----------------------------------------------------------------- *)

RootSystem[Uqgl[m_, n_], gradings_List?VectorQ, liftings_List?VectorQ] :=
  RootSystem[gl[m, n], gradings, liftings]  /;
    SoberAlgebraQ[Uqgl[m,n]];

(* ----------------------------------------------------------------- *)

AdjointWeightOfAlgebraGenerator[Generator[Uqgl[m_, n_], a_, b_, N_]] :=
  AdjointWeightOfAlgebraGenerator[Generator[gl[m, n], a, b]] /;
    SoberAlgebraGeneratorQ[Generator[Uqgl[m, n], a, b, N]];

(* ----------------------------------------------------------------- *)

CoProduct[Generator[Uqgl[m_, n_], a_, a_, N_]] :=
  (CoProduct[Generator[Uqgl[m, n], a, a, N]] =
    TensorProduct[Generator[Uqgl[m, n], a, a, N], Generator[Uqgl[m, n], a, a, \
N]]
  ) /; SoberAlgebraGeneratorQ[Generator[Uqgl[m, n], a, a, N]];

CoProduct[Generator[Uqgl[m_, n_], a_, b_, 1]] :=
  (CoProduct[Generator[Uqgl[m,n], a, b, 1]] =
  Module[
    {Eab, Sab, KahalfP, KahalfM, KbhalfP, KbhalfM, c, cp},

    Eab = Generator[Uqgl[m,n], a, b, 1];
    Sab = Sign[a-b];
    KahalfP = Generator[Uqgl[m,n], a, a, + Sab/2];
    KahalfM = Generator[Uqgl[m,n], a, a, - Sab/2];
    KbhalfP = Generator[Uqgl[m,n], b, b, + Sab/2];
    KbhalfM = Generator[Uqgl[m,n], b, b, - Sab/2];
    cp =
      TensorProduct[Eab, KahalfP ** KbhalfM] + TensorProduct[KahalfM ** \
KbhalfP, Eab] -
      Sab *
      Sum[
        DeltaSub[Uqgl[m,n], c] *
        TensorProduct[
          Generator[Uqgl[m,n], c, c, - Sab/2] ** KbhalfP ** \
Generator[Uqgl[m,n], a, c, 1],
          Generator[Uqgl[m,n], c, b, 1] ** KahalfP ** Generator[Uqgl[m,n], c, \
c, - Sab/2]
        ],
        {c, Min[a,b]+1, Max[a,b]-1}
      ];

    cp
  ]) /; SoberAlgebraGeneratorQ[Generator[Uqgl[m, n], a, b, 1]];

(* ----------------------------------------------------------------- *)

Antipode[Generator[Uqgl[m_, n_], a_, a_, N_]] :=
  (Antipode[Generator[Uqgl[m, n], a, a, N]] =
    Generator[Uqgl[m, n], a, a, -N]
  ) /; SoberAlgebraGeneratorQ[Generator[Uqgl[m, n], a, a, N]];

Antipode[Generator[Uqgl[m_, n_], a_, b_, 1]] :=
  (Antipode[Generator[Uqgl[m,n], a, b, 1]] =
    - PowerExpand[(qSub[Uqgl[m,n], a] * qSub[Uqgl[m,n], b])^(1/2 Sign[a-b])] \
* Generator[Uqgl[m,n], a, b, 1]
  ) /; SoberAlgebraGeneratorQ[Generator[Uqgl[m, n], a, b, 1]];

(* ----------------------------------------------------------------- *)

CoUnit[Generator[Uqgl[m_, n_], a_, b_, N_]] :=
  (CoUnit[Generator[Uqgl[m, n], a, b, N]] =
    If[(a==b), 1, 0]
  ) /; SoberAlgebraGeneratorQ[Generator[Uqgl[m, n], a, b, N]];

(* ----------------------------------------------------------------- *)

(* End[]; *)

EndPackage[];

