(* -----------------------------------------------------------------

  Z Algorithms.m

  Algorithms (Z_0, Z_1, Z_2) for the evaluation of link invariants using a 
  state model for links supplied as the vertical closures of braids. 
  Computations use sparse tensor multiplication.

  ------------------------------------------------------------------ *)

BeginPackage["LinksGouldExplorer`"];

TensorOfUnclosedBraid::usage = "
  TensorOfUnclosedBraid[Algebra, RepName, bw, n] forms a tensor corresponding \
\
\

  to the representation of the unclosed braid bw, regarded as being within 
  braid group B_n. It works recursively, storing smaller versions of itself.
";

BraidRelationChecker::usage = "
  BraidRelationChecker[Algebra, RepName] checks that the appropriate 
  representation of the braid generator satisfies the braid relation.
";

MatveevDeltaUnknottingOperationTest::usage = "
  MatveevDeltaUnknottingOperationTest[Algebra, RepName] checks to see whether \
\
\

  application of the Matveev Delta--Nabla unknotting test to the appropriate 
  representation of the braid generator indicates that the resulting \
invariant 
  is necessarily trivial.
";

Z0Algorithm::usage = "
  Z0Algorithm[Algebra, RepName, bw] is the rank (1,1) (width 1) sparse tensor \
\
\

  corresponding to the contraction of the link represented as the closure of 
  the given braid into a (1,1)-tangle, evaluated via the Z_0 algorithm.
";

Z1Algorithm::usage = "
  Z1Algorithm[Algebra, RepName, bw] is the rank (1,1) (width 1) sparse tensor \
\
\

  corresponding to the contraction of the link represented as the closure of 
  the given braid into a (1,1)-tangle, evaluated via the Z_1 algorithm. This 
  function may (untested) also work with some `split braid' links like that 
  with the braid word \"ac\".
";

Z2Algorithm::usage = "
  Z2Algorithm[Algebra, RepName, bw] is the rank (1,1) (width 1) sparse tensor \
\
\

  corresponding to the contraction of the link represented as the closure of 
  the given braid word into a (1,1)-tangle, evaluated via the Z_2 algorithm. 
  Data is spewed out so as to give a hint of how the calculation is 
  progressing. This algorithm currently fails for some braid words such as 
  \"ab\", for which a contraction yields a zero tensor. [So fix it!] This set \
\
\

  includes many `split braid' links such as \"ac\". Also note that its full 
  functionality is currently disabled, and it revert to the Z1 algorithm.
";

Z1Width::usage = "
  Z1Width[bw] is the maximum number of strings used by the Z_1 algorithm when \
\
\

  processing bw. It works by synthesising the Z_1 algorithm.
";

Z1BraidPreConditioner::usage = "
  Z1BraidPreConditioner[bw] returns rbw, a rotated version of bw such that \
the 
  Z1Width of rbw is minimised over all possible rotations (and flips, 
  horizontal and vertical) of bw. As currrently implemented, sometimes it 
  returns a braid which is actually less efficient than the original, as the 
  performance of the Z1 algorithm is so not easily predetermined.
";

OpenTangleLinkInvariantOfALinkFromABraidWord::usage = "
  OpenTangleLinkInvariantOfALinkFromABraidWord[Algebra, RepName, L] evaluates \
\
\

  the appropriate open (i.e. the (1,1)-) tangle link invariant for the given 
  link L, by use of a Z-algorithm. This presumes the knowldege of a braid \
word 
  for L. Currently, {Algebra, RepName} may be {U_q[gl(m|n)], (0|alpha)} or 
  {U_q[sl(2)], {D}.
";

(* ------------------------------------------------------------------ *)

(* Begin["`Private`"]; *)

TensorOfUnclosedBraid[Algebra_, RepName_, bw_String?SoberBraidWordQ, \
n_Integer?Positive] :=
  (TensorOfUnclosedBraid[Algebra, RepName, bw, n] =
  Module[
    {index, power, D, Z},

    {index, power} = Flatten[BraidIndicesAndPowersFromBraidWord[bw]];
    Z = RepresentationOfTheBraidGeneratorAsASparseArray[Algebra, RepName, \
power];
    D = DimensionOfModule[Algebra, RepName];
    If[(index>1), Z = \
SparseTensorProductOfSparseTensors[IdentitySparseTensor[D,index-1],Z]];
    If[(n>index+1), Z = \
SparseTensorProductOfSparseTensors[Z,IdentitySparseTensor[D,n-index-1]]];

    Z
  ]) /; (Length[BraidSyllablesFromBraidWord[bw]]==1);

TensorOfUnclosedBraid[Algebra_, RepName_, bw_String?SoberBraidWordQ, \
n_Integer?Positive] :=
  (TensorOfUnclosedBraid[Algebra, RepName, bw, n] =
  Module[
    {bs, bslast, bwstub, index, power, Z, W},

    bs = BraidSyllablesFromBraidWord[bw];
    bslast = Take[bs,-1];
    bwstub = BraidWordFromBraidSyllables[Drop[bs,-1]];
    {index, power} = \
Flatten[BraidIndicesAndPowersFromBraidSyllables[bslast]];
    Z = TensorOfUnclosedBraid[Algebra, RepName, bwstub, n];
    W = RepresentationOfTheBraidGeneratorAsASparseArray[Algebra, RepName, \
power];
    Z = AccreteSparseTensorXToSparseTensorZatStrings[Z, W, index];
    Z[[All,2]] = Expand[Z[[All,2]]];
    Z = TidySparseTensorComponents[Z];
    Z
  ]) /; (Length[BraidSyllablesFromBraidWord[bw]]>1);


(* ----------------------------------------------------------------- *)

BraidRelationChecker[Algebra_, RepName_] :=
  BraidRelationChecker[Algebra, RepName] =
  Module[
    {T1, T2, brtest},

    T1 = TensorOfUnclosedBraid[Algebra, RepName, "aba", 3];
    T2 = TensorOfUnclosedBraid[Algebra, RepName, "bab", 3];
    brtest = SameQ[T1,T2];

    brtest
  ];

MatveevDeltaUnknottingOperationTest[Algebra_, RepName_] :=
  MatveevDeltaUnknottingOperationTest[Algebra, RepName] =
  Module[
    {T1, T2, mtest},

    T1 = TensorOfUnclosedBraid[Algebra, RepName, "aBa", 3];
    T2 = TensorOfUnclosedBraid[Algebra, RepName, "bAb", 3];
    mtest = If[SameQ[T1,T2], "Trivial", "Not necessarily trivial"];

    mtest
  ];

(* ----------------------------------------------------------------- *)

Z0Algorithm[Algebra_, RepName_, bw_String?SoberBraidWordQ] :=
  Module[
    {Z, H, w},

    w = StringIndexOfBraidWord[bw];
    Z = TensorOfUnclosedBraid[Algebra, RepName, bw, w];
    H = RepresentationOfTheHandleAsASparseArray[Algebra, RepName, 1];
    Do[
      Z = ContractSparseTensorZOverSparseTensorHOnStrings[Z, H, 1],
      {StringIndexOfBraidWord[bw]-1}
    ];

    Z
  ];

(* ----------------------------------------------------------------- *)


Z1Algorithm[Algebra_, RepName_, bw_String?SoberBraidWordQ] :=
  Module[
    {
      D,        (* dimension of the underlying module                   *)
      bs,       (* braid syllables of bw *)
      minibraid, minisib,
      miniindices, minipowers, addpoint,
      indices, powers, (* description of bs *)
      ci, cp,   (* current index and power                             *)
      Z,        (* tensor corresponding to the processed part of bw  *)
      W,        (* currently-being-accumulated tensor                   *)
      H,        (* Handle *)
      wZ,       (* number of strings in the current Z                   *)
      endstr,   (* `last' string in the current beta                    *)
      sc, nmsc, (* seen, no more seen crosings *)
      ripset, ripstr (* strings to be ripped *)
    },

    bs = BraidSyllablesFromBraidWord[bw];
    indices = Flatten[Map[Union[SignedIntegerBraidFromBraidWord[#]]&, bs]];
    powers = Sign[indices] * Map[StringLength, bs]; indices = Abs[indices];

    ci = indices[[1]]; cp = powers[[1]];
    D = DimensionOfModule[Algebra, RepName];
    Z = RepresentationOfTheBraidGeneratorAsASparseArray[Algebra, RepName, \
cp];
    If[ci > 1, Z = SparseTensorProductOfSparseTensors[IdentitySparseTensor[D, \
\
\
ci-1], Z]];
    indices = Drop[indices, 1]; powers = Drop[powers, 1]; bs = Drop[bs, 1];
   
    While[
      (indices =!= {}),
      ci = indices[[1]]; cp = powers[[1]];
      wZ = UpperRankOfSparseTensor[Z];
      If[
        (ci >= wZ),
       Z = SparseTensorProductOfSparseTensors[Z, IdentitySparseTensor[D, \
ci-wZ+1]];
       
      ];
      W = RepresentationOfTheBraidGeneratorAsASparseArray[Algebra, RepName, \
\
\
cp];
      
      Z = AccreteSparseTensorXToSparseTensorZatStrings[Z, W, ci];
      indices = Drop[indices, 1]; powers = Drop[powers, 1]; bs = Drop[bs, 1];
      

      wZ = UpperRankOfSparseTensor[Z];
      If[
        indices =!= {},
        sc = Union[Select[indices, (# < wZ)&]];
        nmsc = Complement[Range[wZ-1], sc];
        ripset = Union[Intersection[{1}, nmsc], Intersection[{wZ}, nmsc+1], \
\
\
Intersection[nmsc, nmsc+1]];
        While[
          (ripset =!= {}),
          ripstr = ripset[[-1]];
         
          endstr = indices[[-1]] + 1;
          H = RepresentationOfTheHandleAsASparseArray[Algebra, RepName, \
Sign[endstr-ripstr]];
(* We know H is diagonal, so precontract Z before passing it *)
          Z = Select[Z, (#[[1,1,ripstr]]==#[[1,2,ripstr]])&];
          Z = ContractSparseTensorZOverSparseTensorHOnStrings[Z, H, ripstr];
          indices = Map[If[#>ripstr, #-1, #]&, indices];
          ripset = Drop[ripset,-1];
          
        ]
      ]
    ];

   
    H = RepresentationOfTheHandleAsASparseArray[Algebra, RepName, 1];
    Z = ContractSparseTensorZOverSparseTensorHOnStrings[Z, H, 1];


    Z
  ];

(* ----------------------------------------------------------------- *)

(* This needs modifying to also return the Z1Length *)

Z1Width[bw_String?SoberBraidWordQ] :=
  Module[
    {sib, cw, mw, s, sc, nmsc, ripset, ripstr, j},

    sib = Abs[SignedIntegerBraidFromBraidWord[bw]];
    sib = Flatten[Map[Union, Split[sib]]];
    cw = 0; mw = 0;
    While[
      (Length[sib] > 0),
      s = sib[[1]] + 1;
      If[(s > cw), cw = s; If[(cw > mw), mw = cw]];
      sib = Drop[sib, 1];
      sc = Union[Select[sib, (# < cw)&]];
      nmsc = Complement[Range[cw-1], sc];
      ripset = Union[Intersection[{1,cw}, nmsc], Intersection[nmsc, nmsc + \
1]];
      While[
        ripset =!= {},
        ripstr = ripset[[-1]];
        Do[If[sib[[j]] >= ripstr, sib[[j]]--], {j, Length[sib]}];
        ripset = Drop[ripset,-1];
        cw--
      ]
    ];

    mw
  ];

(* ----------------------------------------------------------------- *)

Z1WidthAndLength::usage = "
  Z1WidthAndLength is a modification of Z1Width to also sort braid words by 
  minimal Z_1-length (for a given Z_1-width). The current implementation 
  doesn't work correctly, but curiously sometimes works. It is still under 
  development.
";

Z1WidthAndLength[bw_String?SoberBraidWordQ] :=
  Z1WidthAndLength[bw] =
  Module[
    {sib, cw, mw, s, cl, cml, ml, sc, nmsc, ripset, ripstr, j},

    sib = Abs[SignedIntegerBraidFromBraidWord[bw]];
    sib = Flatten[Map[Union, Split[sib]]];
    cw = 0; mw = 0;
    cl = 0; cml = 0; ml = 0;
    While[
      (Length[sib] > 0),
      s = sib[[1]] + 1;
      If[(s > cw), cw = s; If[(cw > mw), mw = cw; cl = 1]];
      cl++;
      If[cw == mw, cml = cl; If[(cml > ml), ml = cml]];
      sib = Drop[sib, 1];
      sc = Union[Select[sib, (# < cw)&]];
      nmsc = Complement[Range[cw-1], sc];
      ripset = Union[Intersection[{1,cw}, nmsc], Intersection[nmsc, nmsc+1]];
      While[
        ripset =!= {},
        ripstr = ripset[[-1]];
        Do[If[sib[[j]] >= ripstr, sib[[j]]--], {j, Length[sib]}];
        ripset = Drop[ripset,-1];
        cw--
      ]
    ];

    {mw, ml}
  ];

(* ----------------------------------------------------------------- *)

(*
  Z1BraidPreConditioner currently calls Z1WidthAndLength, but that function \
\
\
is wrong,
  (although sometimes helpful!). However, if the call changes to just \
Z1Width,
  the power is weakened, although that function is correct.
*)

Z1BraidPreConditioner[bw_String?SoberBraidWordQ] :=
  Z1BraidPreConditioner[bw] =
  Module[
    {l, bwt, obw, obwt, tt, ztt, bwout},

    l = BraidWordLength[bw];
    bwt = Map[StringJoin[RotateLeft[Characters[bw], #]]&, Range[l] - 1];
    obw = OppositeBraidWord[bw];
    obwt =  Map[StringJoin[RotateLeft[Characters[obw], #]]&, Range[l] - 1];
    tt = Join[bwt, obwt];
    tt = Join[tt, Map[StringReverse, tt]];
    ztt = Map[Z1WidthAndLength, tt];
    bwout = Sort[Transpose[{ztt, tt}]][[1, 2]];

    bwout
  ];
(* ----------------------------------------------------------------- *)

(*
  The Z algorithms assume that an input braid word of width n is intended to 
  correspond to a braid within B_n (that is, not within some larger group, 
  B_{n+1} for instance). This should be upgraded so that they accept as input \
\
\

  a signed integer braid representative, or a braid word with an attached 
  braid group index.
  
  They work by loading state model parameters for particular invariants via 
  their labelling through an Algebra and RepName. This should be upgraded so 
  that they accept state model parameters as their inputs, and their calling 
  should separate data from functions. These comments apply to other \
functions
  in this file.

  The idea of block-accretion in the Z2 algorithm works, but is inefficient 
  because the accretion is, for some unknown reason, expensive. So the \
present 
  implementation disables it by forcing the block-length to be 1, which 
  reduces the Z2 algorithm to the Z1. This implementation of the algorithm 
  works for cases where the tangle is reduced to a 0-string braid at some 
  point, for example for the input braid word "ab", but is inefficent in that \
\
\

  it doesn't bail out at that point.
*)

Z2Algorithm[Algebra_, RepName_, ""] :=
  IdentitySparseTensor[DimensionOfModule[Algebra, RepName], 1];

Z2Algorithm[Algebra_, RepName_, bw_String?SoberBraidWordQ] :=
  Module[
    {
      D, Z, W, H, bs, cbs, indices, powers, blockbw,
      blocklen, blockbs, blockindices, blockpowers, addpoint,
      wZ, seencrossings, nomoremseencrossings, ripset, ripstr, endstr, i
    },

   

    bs = BraidSyllablesFromBraidWord[bw];
    cbs = bs;
    D = DimensionOfModule[Algebra, RepName];
    Z = IdentitySparseTensor[D, 2];

    While[
      (cbs =!= {}),
      {indices, powers} = BraidIndicesAndPowersFromBraidSyllables[cbs];
      blocklen  = Min[Flatten[Map[Position[indices, #]&, \
Union[indices]][[All, -1]]]];
      blocklen = 1; (* This reverts to the Z1 algorithm *)

      blockbs = Take[cbs, blocklen];
      {blockindices, blockpowers} = \
BraidIndicesAndPowersFromBraidSyllables[blockbs];
      If[
        (Max[blockindices] >= UpperRankOfSparseTensor[Z]),
       
        Z = SparseTensorProductOfSparseTensors[Z, IdentitySparseTensor[D, \
Max[blockindices]-UpperRankOfSparseTensor[Z]+1]];
     
      ];
      addpoint = Min[blockindices];
      blockbw = BraidWordFromBraidIndicesAndPowers[{blockindices-addpoint+1, \
\
\
blockpowers}];
      W = TensorOfUnclosedBraid[Algebra, RepName, blockbw, \
StringIndexOfBraidWord[blockbw]];
      
      Z = AccreteSparseTensorXToSparseTensorZatStrings[Z, W, addpoint];
      endstr = Last[indices] + 1;
      indices = Drop[indices, blocklen]; powers = Drop[powers, blocklen]; bs \
\
\
= Drop[bs, blocklen];
     

      wZ = UpperRankOfSparseTensor[Z];
      seencrossings = Union[Select[indices, (# < wZ)&]];
      nomoremseencrossings = Complement[Range[wZ-1], seencrossings];
      ripset =
        Complement[
          Union[
            Intersection[{1},nomoremseencrossings],
            Intersection[{wZ},nomoremseencrossings+1],
            Intersection[nomoremseencrossings,nomoremseencrossings+1]
          ],
          {endstr}
        ];
      If[(Length[ripset]==UpperRankOfSparseTensor[Z]), ripset = \
Drop[ripset,-1]];

      If[
        (ripset =!= {}),
       
        Do[
          ripstr = ripset[[i]];
          Z = Select[Z, (#[[1,1,ripstr]]==#[[1,2,ripstr]])&],
          {i, Length[ripset]}
        ];
       
      ];

      While[
        (ripset =!= {}),
        ripstr = Last[ripset];
        H = RepresentationOfTheHandleAsASparseArray[Algebra, RepName, \
Sign[endstr-ripstr]];
       
        Z = ContractSparseTensorZOverSparseTensorHOnStrings[Z, H, ripstr];
        indices = Map[If[#>ripstr, #-1, #]&, indices];
        If[endstr>ripstr, endstr--];
        ripset = Drop[ripset,-1];
      
      ];
      cbs = BraidSyllablesFromBraidIndicesAndPowers[{indices,powers}]
    ];

(* This test fails if Z = {{e[{1},{1}],0}}, that is ZeroSparseTensor[{1,1}] \
\
\
*)
       
    Z
  ] /; (StringLength[bw] > 0)

(* ----------------------------------------------------------------- *)
  OpenTangleLinkInvariantOfALinkFromABraidWord[Algebra_, RepName_, L_] :=
  Module[
    {bw, otli},

    bw = BraidWord[L];
   

(*    bw = Z1BraidPreConditioner[bw]; *)
    otli =
      If[
        (bw===""),
        1,
        Z2Algorithm[Algebra, RepName, bw][[1,2]]
      ];

    otli
  ] /; SoberBraidWordQ[BraidWord[L]]; 

(* ----------------------------------------------------------------- *)

(* End[]; *)

EndPackage[];



