(*
  glmn.m

  Gradings, generators and roots for gl(m|n).

   ----------------------------------------------------------------- *)

BeginPackage["LinksGouldExplorer`"];

SoberAlgebraQ::usage = "
  SoberAlgebraQ[Algebra] is True if Algebra is known to Links--Gould \
Explorer. 
  Specifically, this is so for Algebra being gl[m,n] or Uqgl[m,n], for 
  positive integers m and n.
";

SoberAlgebraIndexQ::usage = "
  SoberAlgebraIndexQ[Algebra, a] is True if a is a valid index for a valid 
  algebra Algebra, which may be gl[m,n] or Uqgl[m,n].
";

GradingOfAlgebraIndex::usage = "
  GradingOfAlgebraIndex[Algebra, a] is the Z_2-grading of the index a of the 
  superalgebra Algebra, which may be gl[m,n] or Uqgl[m,n]. For either, if 
  a <= m, the grading is 0 (even), else if m+1 <= a <= m+n, the grading is 
  1 (odd). 
";

Generator::usage = "
  Generator[Algebra, indices] is the Algebra generator with given indices. 
  Specifically, Generator[gl[m, n], a, b] is the gl[m,n] generator E^a_b, and \

  Generator[Uqgl[m, n], a, b, N] is the U_q[gl(m|n)] `generator' E^a_b or 
  K_a^N.
";

SoberAlgebraGeneratorQ::usage = "
  SoberAlgebraGeneratorQ[G] is True if G is a valid generator for a valid 
  algebra Algebra, which may be gl[m,n] or Uqgl[m,n].
";

GradingOfAlgebraGenerator::usage = "
  GradingOfAlgebraGenerator[G] is the Z_2-grading of generator G, of some 
  valid algebra.
";

LiftingOfAlgebraGenerator::usage = "
  LiftingOfAlgebraGenerator[G] is the lifting (lowering: -1, Cartan: 0 or 
  raising: 1) of generator G, of some valid algebra.
";

GeneratorSet::usage = "
  GeneratorSet[Algebra, gradings, liftings] is the set of generators for 
  Algebra of the given gradings (a subset of {0,1}) and liftings (a subset of \

  {-1,0,+1}). For example, the set of odd raising generators is 
  GeneratorSet[Algebra, {1}, {+1}] and that of even noncartan generators is 
  GeneratorSet[Algebra, {0}, {-1,+1}], Some combinations of grading and 
  lifting are redundant. Algebra may be gl[m,n] or Uqgl[m,n], and if the 
  latter, the output is ordered by decreasing weight order (using 
  OrderedGeneratorsQ).
";


HermitianConjugate::usage = "
  HermitianConjugate[Algebra, W] is the Hermitian conjugate of Algebra object \

  W, which is in general a weighted combination of products of generators. 
  Algebra may be gl[m,n] or Uqgl[m,n].
";


Rank::usage = "
  Rank[Algebra] is the rank of Algebra, which may be gl[m,n] or Uqgl[m,n].
";

SoberAlgebraWeightQ::usage = "
  SoberAlgebraWeightQ[Algebra, w] is True if w is a valid weight for Algebra.
  Algebra may be gl[m,n] or Uqgl[m,n].
";

FundamentalWeight::usage = "
  FundamentalWeight[gl[m,n], a] is the gl(m|n) fundamental weight number a.
";

RootSystem::usage = "
  RootSystem[gl[m,n], gradings, liftings] is the set of roots for gl(m|n) of 
  the given gradings and liftings.
";

SignOfAlgebraWeight::usage = "
  SignOfAlgebraWeight[Algebra, gamma] is the sign (-1, 0 or 1) of an Algebra 
  weight gamma.
";

LiftingOfAlgebraWeight::usage = "
  LiftingOfAlgebraWeight[Algebra, gamma] is the lifting (-1, 0 or 1) of the 
  Algebra weight gamma.
";

GradedInnerProductOfAlgebraWeights::usage = "
  GradedInnerProductOfAlgebraWeights[gl[m, n], w1, w2] is the graded inner 
  product of two graded weights w1 and w2 in the gl(m|n) weight space.
";

AdjointWeightOfAlgebraGenerator::usage = "
  AdjointWeightOfAlgebraGenerator[G] is the adjoint action weight of the 
  generator G of some valid algebra, that is, it is the weight of the root 
  associated with G.
";

AdjointWeightOfAlgebraGeneratorProduct::usage = "
  AdjointWeightOfAlgebraGeneratorProduct[S] is the adjoint action weight of 
  the generator product S of some valid algebra, that is, it is the weight of \

  the root associated with S.
";

HalfSumofPositiveRoots::usage = "
  HalfSumofPositiveRoots[gl[m,n], i] is the half sum of positive roots of 
  grading i in {0,1}, for gl(m|n). If i is omitted, it is the graded half sum \

  of positive roots rho_0 - rho_1.
";

(* ----------------------------------------------------------------- *)

(* Begin["`Private`"]; *)

SoberAlgebraQ[gl[m_Integer?Positive, n_Integer?Positive]] := True;

SoberAlgebraIndexQ[gl[m_, n_], a_Integer?Positive] :=
  SoberAlgebraIndexQ[gl[m, n], a] =
    (SoberAlgebraQ[gl[m,n]] && (a <= m + n));

GradingOfAlgebraIndex[gl[m_, n_], a_] :=
  (If[(a <= m), 0, 1]) /; SoberAlgebraIndexQ[gl[m, n], a];

SoberAlgebraGeneratorQ[Generator[gl[m_, n_], a_, b_]] :=
  SoberAlgebraGeneratorQ[Generator[gl[m,n], a, b]] =
    (SoberAlgebraIndexQ[gl[m, n], a] && SoberAlgebraIndexQ[gl[m, n], b]);

GradingOfAlgebraGenerator[Generator[gl[m_, n_], a_, b_]] :=
  Mod[GradingOfAlgebraIndex[gl[m,n], a] + GradingOfAlgebraIndex[gl[m,n], b], \
2] /;
    SoberAlgebraGeneratorQ[Generator[gl[m,n], a, b]];

LiftingOfAlgebraGenerator[Generator[gl[m_, n_], a_, b_]] :=
  Sign[b-a] /; SoberAlgebraGeneratorQ[Generator[gl[m,n], a, b]];

(* ----------------------------------------------------------------- *)

GeneratorSet[gl[m_, n_], gradings_List?VectorQ, liftings_List?VectorQ] :=
  Module[
    {gs},

    gs = Flatten[Table[Generator[gl[m,n], a, b], {a, m+n}, {b, m+n}]];
    gs = Select[gs, MemberQ[gradings, GradingOfAlgebraGenerator[#]]&];
    gs = Select[gs, MemberQ[liftings, LiftingOfAlgebraGenerator[#]]&];

    gs
  ] /; SoberAlgebraQ[gl[m,n]];

(* ----------------------------------------------------------------- *)

HermitianConjugate[Algebra_, W_Plus] :=
  Map[HermitianConjugate[Algebra, #]&, W] /;
    SoberAlgebraQ[Algebra];

HermitianConjugate[Algebra_, W_Times] :=
  Map[HermitianConjugate[Algebra, #]&, W] /;
    SoberAlgebraQ[Algebra];

HermitianConjugate[Algebra_, W_NonCommutativeMultiply] :=
  Apply[NonCommutativeMultiply, Reverse[Map[HermitianConjugate[Algebra, #]&, \
Apply[List, W]]]] /;
    SoberAlgebraQ[Algebra];

HermitianConjugate[Algebra_, W_] := W /;
  (SoberAlgebraQ[Algebra] && !OperatorQ[W]);    (* as q is ahem, real *)


HermitianConjugate[gl[m_, n_], Generator[gl[m_, n_], a_, b_]] :=
  Generator[gl[m,n], b, a, N] /;
    SoberAlgebraGeneratorQ[Generator[gl[m,n], a, b]];

(* ----------------------------------------------------------------- *)

FundamentalWeight[gl[m_, n_], a_] :=
  (FundamentalWeight[gl[m,n], a] =
    ReplacePart[Table[0, {m+n}], 1, a]
  ) /; SoberAlgebraIndexQ[gl[m, n], a];

(* ----------------------------------------------------------------- *)

Rank[gl[m_, n_]] := m + n - 1 /;
  SoberAlgebraQ[gl[m,n]];

SoberAlgebraWeightQ[Algebra_, gamma_List?VectorQ] :=
  (SoberAlgebraQ[Algebra] && (Length[gamma] == (Rank[Algebra] + 1)));

SignOfAlgebraWeight[Algebra_, gamma_] :=
  Module[
    {temp, sgnwt},

    temp = Select[Sign[gamma], (# =!= 0 &)];
    sgnwt = If[(Length[temp] == 0), 0, temp[[1]]];

    sgnwt
  ] /; SoberAlgebraWeightQ[Algebra, gamma];

(* ----------------------------------------------------------------- *)

RootSystem[gl[m_, n_], gradings_List?VectorQ, liftings_List?VectorQ] :=
  Map[AdjointWeightOfAlgebraGenerator, GeneratorSet[gl[m,n], gradings, \
liftings]] /;
    SoberAlgebraQ[gl[m,n]];

(* ----------------------------------------------------------------- *)

GradedInnerProductOfAlgebraWeights[gl[m_, n_], gamma1_, gamma2_] :=
  (Take[gamma1, m] . Take[gamma2, m] - Drop[gamma1, m] . Drop[gamma2, m]) /;
    (SoberAlgebraWeightQ[gl[m,n], gamma1] && SoberAlgebraWeightQ[gl[m,n], \
gamma2]);

(* ----------------------------------------------------------------- *)

AdjointWeightOfAlgebraGeneratorProduct[W_NonCommutativeMultiply] :=
  Apply[Plus, Map[AdjointWeightOfAlgebraGenerator, Apply[List, W]]] /;
    SoberAlgebraQ[W[[1,1]]];

AdjointWeightOfAlgebraGenerator[Generator[gl[m_, n_], a_, b_]] :=
   FundamentalWeight[gl[m,n], a] - FundamentalWeight[gl[m,n], b] /;
     SoberAlgebraGeneratorQ[Generator[gl[m,n], a, b]];

(* ----------------------------------------------------------------- *)

HalfSumofPositiveRoots[Algebra_, grading_Integer] :=
  Apply[Plus, RootSystem[Algebra, {grading}, {1}]]/2 /;
    SoberAlgebraQ[Algebra];

HalfSumofPositiveRoots[Algebra_] :=
  (HalfSumofPositiveRoots[Algebra, 0] - HalfSumofPositiveRoots[Algebra, 1]) \
/;
    SoberAlgebraQ[Algebra];

(* ----------------------------------------------------------------- *)

(* End[]; *)

EndPackage[];

