(* -----------------------------------------------------------------

  q-Expressions.m

  Manipulate q-expressions.

  ------------------------------------------------------------------ *)

BeginPackage["LinksGouldExplorer`"];

qBracket::usage = "
  qBracket[WW] is the q-bracket of an Algebra element WW (typically a scalar 
  multiple of Id), that is [WW]_q =(q^(+WW) - q^(-WW))/(q^(+1) - q^(-1)). To 
  expand qBracket[WW] into a rational expression, use ExpandqBrackets.
";

ExpandqBrackets::usage = "
  ExpandqBrackets[S] expands any q-brackets appearing in the expression S.
";

ConvertToqBrackets::usage = "
  ConvertToqBrackets[S] installs q-brackets in an expression S by rewriting 
  subexpressions of the form q^(+WW) - q^(-WW) as q-brackets (q^(+1)-q^(-1)) 
  [WW]_q. The function assumes that the expression S is already symmetrised, 
  and that the power WW is not a number.
";

SymmetriseqBinomials1::usage = "
  SymmetriseqBinomials1[S] applies transformatons to an expression S to 
  symmetrise two-term expressions involving powers of q. It is tricky to only \
\

  symmetrise q-polynomials containing exactly 2 terms.
";

SymmetriseqBinomials2::usage = "
  SymmetriseqBinomials2[S] applies some transformations to an expression S to \
\

  symmetrise two-term expressions involving powers of q.
";

SymmetriseqExpressions::usage = "
  SymmetriseqExpressions[S] applies transformatons to an expression S to 
  symmetrise two-term expressions involving powers of q.
";

PowerClean::usage = "
  PowerClean[S] rewrites an expression S, tidying up sums within exponents.
";

FactorWithinPowers::usage = "
  FactorWithinPowers[S] rewrites an expression S, factoring expressions \
within 
  powers.
";

CleanqExpressions::usage = "
  CleanqExpressions[S] rewrites an expression S by replacing q-binonials in S \
\

  with q-brackets.
";

(* ------------------------------------------------------------------ *)

(* Begin["`Private`"]; *)

ExpandqBrackets[S_] := S //. qBracket[WW_] :> (q^(+WW) - q^(-WW))/(q^(+1) - \
\
q^(-1));

ConvertToqBrackets[S_] :=
  S //.
  {
    (  q^(x1_) - q^(x2_)) :> qBracket[x1] (q^(+1) - q^(-1)) /; ((x1+x2==0) && \
\
!NumberQ[x1]),
    (- q^(x1_) + q^(x2_)) :> qBracket[x2] (q^(+1) - q^(-1)) /; ((x1+x2==0) && \
\
!NumberQ[x1]),

    qBracket[ca_ * alpha] :> - qBracket[- ca alpha] /; Negative[ca],
    qBracket[ca_ alpha + cb_] :> - qBracket[- ca alpha - cb] /; Negative[ca]
  };

(* ----------------------------------------------------------------- *)

SymmetriseqBinomials1Rules :=
  {
    ((cx_Integer) + (cyy_Integer:1) q) :> q^(1/2) * (cx q^(-1/2) + cyy \
q^(1/2)) \
/; ((Abs[cx]==1) && (Abs[cyy]==1)),

    (  1 + q^(yy_)) :> q^Expand[yy/2] * (  q^Expand[-yy/2] + q^Expand[yy/2]),
    (- 1 + q^(yy_)) :> q^Expand[yy/2] * (- q^Expand[-yy/2] + q^Expand[yy/2]),
    (  1 - q^(yy_)) :> q^Expand[yy/2] * (  q^Expand[-yy/2] - q^Expand[yy/2]),

    (  q^(x_:1) + q^(yy_:1)) :> q^Expand[(x+yy)/2] * (  q^Expand[(x-yy)/2] + \
\
q^Expand[(-x+yy)/2]) /; (x+yy=!=0),
    (  q^(x_:1) - q^(yy_:1)) :> q^Expand[(x+yy)/2] * (  q^Expand[(x-yy)/2] - \
\
q^Expand[(-x+yy)/2]) /; (x+yy=!=0),
    (- q^(x_:1) + q^(yy_:1)) :> q^Expand[(x+yy)/2] * (- q^Expand[(x-yy)/2] + \
\
q^Expand[(-x+yy)/2]) /; (x+yy=!=0)
  };

Attributes[SymmetriseqBinomials1] = {Listable};

SymmetriseqBinomials1[S_Times] :=
  Map[SymmetriseqBinomials1, S];

SymmetriseqBinomials1[S_Power] :=
  Power[SymmetriseqBinomials1[S[[1]]], S[[2]]];

SymmetriseqBinomials1[S_Plus] :=
  If[
    (Length[S]>2),
    Map[SymmetriseqBinomials1, S],
    If[
      PolynomialQ[S, HoldPattern[q^a]],
      S //. SymmetriseqBinomials1Rules,
      Map[SymmetriseqBinomials1, S]
    ]
  ];

SymmetriseqBinomials1[S_] := S;

(* ----------------------------------------------------------------- *)


SymmetriseqBinomials2[S_] :=
  S //.
  {
    (  1 - q^(yy_)) * (1 + q^(yy_)) :> q^yy * (  q^(-yy) - q^(+yy)),
    (- 1 + q^(yy_)) * (1 + q^(yy_)) :> q^yy * (- q^(-yy) + q^(+yy)),

  (* Experimental. Will suffice as a kludge for now ... *)
    (- q^(-1) + q^(+1)) * (q^(-1) + q^(+1)) -> (- q^(-2) + q^(+2)),

    Power[  1 - q^(yy_), (z_)] * Power[1 + q^(yy_), (z_)] :> q^(yy z) * \
Power[  \
q^(-yy) - q^(+yy), z],
    Power[- 1 + q^(yy_), (z_)] * Power[1 + q^(yy_), (z_)] :> q^(yy z) * \
Power[- \
q^(-yy) + q^(+yy), z],

    (  q^(x_) - q^(yy_)) * (q^(x_) + q^(yy_)) :> q^Expand[x+yy] * (  \
q^Expand[ \
x-yy] - q^Expand[-x+yy]),
    (- q^(x_) + q^(yy_)) * (q^(x_) + q^(yy_)) :> q^Expand[x+yy] * (- \
q^Expand[ \
x-yy] + q^Expand[-x+yy]),

    Power[  q^(x_) - q^(yy_), z_] * Power[ q^(x_) + q^(yy_), z_] :> \
q^Expand[(x+yy)z] * Power[  q^Expand[ x-yy] - q^Expand[-x+yy], z],
    Power[- q^(x_) + q^(yy_), z_] * Power[ q^(x_) + q^(yy_), z_] :> \
q^Expand[(x+yy)z] * Power[- q^Expand[ x-yy] + q^Expand[-x+yy], z]
  };

SymmetriseqExpressions[S_] :=
  SymmetriseqBinomials2[
  PowerExpand[
  SymmetriseqBinomials1[
    S
  ]]];

(* ----------------------------------------------------------------- *)

PowerClean[S_] :=
  S //. {Power[A_, B_] :> Power[A, Expand[B]]};

FactorWithinPowers[S_] :=
  S //. {Power[x_, z_] :> Power[Factor[x], z]};

(* ----------------------------------------------------------------- *)

CleanqExpressions[0] := 0;

CleanqExpressions[S_] :=
  PowerClean[
  PowerExpand[
  ConvertToqBrackets[
  SymmetriseqBinomials2[
  PowerExpand[
  SymmetriseqBinomials1[
  Factor[
  PowerClean[
  PowerExpand[
  FactorWithinPowers[
  ExpandqBrackets[
    S
  ]]]]]]]]]]];

(* ----------------------------------------------------------------- *)

(* End[]; *)

EndPackage[];


