(* -----------------------------------------------------------------

  Representations.m

  Construct an orthonormal basis for the U_q[gl(m|n)] module V(0|alpha), and 
  then matrix elements for the associated representation pi_{(0|alpha)}. In 
  time, this material to be generalised to other representations.

  ------------------------------------------------------------------ *)

BeginPackage["LinksGouldExplorer`"];

LGIHighestWeight::usage = "
  LGIHighestWeight[Uqgl[m, n]] is the particular highest weight (0|alpha) of 
  the U_q[gl(m|n)] representation used in the construction of the \
Links--Gould 
  invariant LG^{m,n}.
";

DimensionOfModule::usage = "
  DimensionOfModule[Algebra, RepName] is the dimension of the representation 
  of Algebra, of highest-weight RepName. If the Algebra is U_q[gl(m|n)], then \
\

  we calculate the dimension of the gl(m|n) (and hence the U_q[gl(m|n)]) 
  representation of highest weight RepName, where RepName is expected to be 
  typical (not checked), using the Kac--Weyl formula from 
  cite[p1243]{Zhang:1993}.
";

ConvertFromRepresentationToInternalVariables::usage = "
  ConvertFromRepresentationToInternalVariables[Uqgl[m,n], RepName, S] \
converts 
  the variables used in the underlying U_q[gl(m|n)] representation (0|alpha) 
  (that is q and alpha) into those used internally in the evaluation of the 
  link invariant (that is q and p).

  The theory comes from the gl(m|n) situation:  The U_q[gl(m|n)] 
  representation of highest weight LGIHighestWeight=(0|alpha) has lowest 
  weight LGIHighestWeightBAR=(-n|alpha+m). The highest weight of the 
  representation dual to LGIHighestWeight is then 
  LGIHighestWeight*=-LGIHighestWeightBAR=(n|-alpha-m), and this \
representation 
  is equivalent to the representation of (0|-alpha-m+n). Thus, LG will be 
  invariant under the transformation alpha -> -alpha-m+n. Under this 
  transformation, the variable p=q^(alpha+(m-n)/2) becomes p^(-1).

  To illustrate, for (m|n) being:
    (1|1), the dual of (0|alpha) is (0|-alpha),   so p = q^alpha
    (2|1), the dual of (0|alpha) is (0|-alpha-1), so p = q^(alpha+1/2)
    (3|1), the dual of (0|alpha) is (0|-alpha-2), so p = q^(alpha+1)
    (2|2), the dual of (0|alpha) is (0|-alpha),   so p = q^alpha.
";

ConvertFromInternalToRepresentationVariables::usage = "
  ConvertFromInternalToRepresentationVariables[Uqgl[m,n], RepName, S] \
converts 
  the variables used internally in the evaluation of the link invariant (that \
\

  is q and p) into in those used in the underlying U_q[gl(m|n)] \
representation 
  RepName=(0|alpha) (that is q and alpha).
";

GeneratorProductsForBasisVectorsOfModule::usage = "
  GeneratorProductsForBasisVectorsOfModule[Algebra, RepName] is a list of the \
\

  products of lowering generators used in the definition of the basis (ket) 
  vectors in the underlying Algebra representation RepName, where Algebra is 
  some Uqgl[m,n].  The ket are ordered by nonincreasing weight within 
  increasing mathbb{Z}-graded levels, that is, products of increasing length. \
\

  The terms within each product are in reverse weight order, in accordance 
  with the ordered cumulative construction of lower-weight kets from higher 
  ones. The correct way to think about the reverse weight order is that the 
  product is read from right to left. The function exploits the fact that the \
\

  output from \"GeneratorSet\" is already ordered by decreasing adjoint 
  weight. This function requires \"DiscreteMath`Combinatorica`\" to access 
  \"KSubsets\".
";

SoberBasisVectorOfModule::usage = "
  SoberBasisVectorOfModule[Algebra, RepName, ket[k]] is true if ket[k] is a 
  meaningful as a basis vector of the Algebra representation of highest \
weight 
  RepName.
";

GradingOfBasisVectorOfModule::usage = "
  GradingOfBasisVectorOfModule[Algebra, RepName, vec[i]] is the 
  mathbb{Z}_2-grading of the vector vec[i], where vec may be either ket or 
  bra, in the Algebra representation RepName.
";

GradingListOfBasisVectorsOfModule::usage = "
  GradingListOfBasisVectorsOfModule[Algebra, RepName] is the list of gradings \
\

  of the basis vectors of the named module.
";

WeightOfBasisVectorOfModule::usage = "
  WeightOfBasisVectorOfModule[Algebra, RepName, vec[i]] is the weight of the 
  representation RepName basis vector vec[i], where vec may be either ket or 
  bra.
";


AlgebraModuleKetActionRules::usage = "
  AlgebraModuleKetActionRules[Algebra, RepName] are the rules used in 
  determining the action of the generators of the Algebra on the 
  highest-weight vector ket[1]. These consist of the following three actions:

    1. The Cartan action on ket[1] is scalar,
    2. ket[1] is annihilated by raising generators, and
    3. For RepName = LGIHighestWeight[Uqgl[m,n]], ket[1] is also annihilated
       by even lowering generators.
";

ApplyAMKARules::usage = "
  ApplyAMKARules[Algebra, RepName, W] repeatedly applies the rules 
  AlgebraModuleKetActionRules[Algebra, RepName] to the expression W until a 
  fixed point is found.
";

AlgebraModuleBraActionRules::usage = "
  AlgebraModuleBraActionRules[Algebra, RepName] are the rules used in 
  determining the action of the generators of the Algebra on the 
  highest-weight vector bra[1]. These consist of the following three actions:

    1. The Cartan action on bra[1] is scalar
    2. bra[1] is annihilated by (the right action of) lowering generators
    3. For RepName = LGIHighestWeight[Uqgl[m,n]], bra[1] is also annihilated \
\
by
       even raising generators.
";

ApplyAMBARules::usage = "
  ApplyAMBARules[Algebra, RepName, W] repeatedly applies the rules
  AlgebraModuleBraActionRules[Algebra, RepName] to the expression W until a
  fixed point is found.
";


NOK::usage = "
  NOK[Algebra, RepName, W] repeatedly applies the normal ordering rules for 
  the Algebra in conjunction with the 
  AlgebraModuleKetActionRules[Algebra, RepName] to the expression W until
  a fixed point is found.
";

NOB::usage = "
  NOB[Algebra, RepName, W] repeatedly applies the normal ordering rules for 
  the Algebra in conjunction with the AlgebraModuleBraActionRules[Algebra, 
  RepName] to the expression W until a fixed point is found.
";

NOBK::usage = "
  NOBK[Algebra, RepName, W] repeatedly applies the normal ordering rules for 
  the Algebra in conjunction with both the 
  AlgebraModuleKetActionRules[Algebra, RepName] and the 
  AlgebraModuleBraActionRules[Algebra, RepName] to the expression W until a 
  fixed point is found.
";

NOBKusingNO::usage = "
  NOBKusingNO[Algebra, RepName, W] repeatedly applies the normal ordering 
  rules of NO for Algebra in conjunction with both the 
  AlgebraModuleKetActionRules[Algebra, RepName] and the 
  AlgebraModuleBraActionRules[Algebra, RepName] to the expression W until a 
  fixed point is found.
";

NOBKC::usage = "
  NOBKC[Algebra, RepName, W] functions as NOBK but first tries to normalise W \
\

  from its centre.
";

NOKL::usage = "
  NOKL[Algebra, RepName, W] functions as NOK but first employs NOL to 
  normalise W from the left.
";


RepresentationPrime::usage = "
  RepresentationPrime[Algebra, RepName, G] is the explicit primed 
  representation of the specified Algebra generator G in the representation 
  RepName. It works in the following way:

    1. Compute pi'(G) via normal orderings and annihilation rules only for
       simple generators G, and Cartan generators of exponent 1.
    2. Compute pi'(K^N)=pi(K)^N via qmapsto q^N
       Build pi'(E^a_b) for nonsimple E^a_b via substitution of pi'(E^c_d) \
etc
       into expansion of E^a_b.

  The function is required by Representation[Algebra, RepName, 
  Generator[Algebra, a, b, N]], which simply calls it and transforms the 
  resulting expression.
";

MetricMatrixofBPrime::usage = "
  MetricMatrixofBPrime[Algebra, RepName] is the metric matrix of the basis 
  Bprime. It is still hopelessly slow for all cases mn > 3. As it is an 
  expensive function, its values for U_q[gl(m|1)] are hard-wired using a 
  conjecture known to be valid for m=1,2,3,4 and are assumed correct for 
  higher m. This requires \"LinearAlgebra`MatrixManipulation`\" to access 
  BlockMatrix.
";

ChangeOfBasisMatrix::usage = "
  ChangeOfBasisMatrix[Algebra, RepName] is the change of basis matrix from 
  Bprime to B for the particular module.
";

Representation::usage = "
  Representation[Algebra, RepName, Generator[Algebra, a, b, N]] is the 
  explicit representation of the specified Algebra Generator in the 
  representation RepName. Works where Algebra is Uqgl[m,n] and RepName is
  LGIHighestWeight[Algebra].
";

EigenvalueOfFirstOrderCasimir::usage = "
  EigenvalueOfFirstOrderCasimir[Uqgl[m,n], RepName] is the eigenvalue of the 
  first order Casimir element of U_q[gl(m|n)] in the representation RepName. 
  The result is returned as a scalar in the representation variables.
";

RepresentationOfqTo2hrho::usage = "
  RepresentationOfqTo2hrho[Uqgl[m,n], RepName] is the representation of the 
  U_q[gl(m|n)] Cartan element q^(2h_{rho}) in the representation RepName. It 
  is presented as a sum of nonzero components in the representation \
variables.
";

StripGradingFromSparseRank2Array::usage = "
  StripGradingFromSparseRank2Array[Algebra, RepName, S] strips the vector 
  space grading from a sparse rank (1,1) object S.
";

(* ------------------------------------------------------------------ *)

(* Begin["`Private`"]; *)

LGIHighestWeight[Uqgl[m_, n_]] :=
  Join[Table[0, {m}], Table[alpha, {n}]] /;
    SoberAlgebraQ[Uqgl[m, n]];

(* ----------------------------------------------------------------- *)

DimensionOfModule[Uqgl[m_, n_], RepName_] :=
  (DimensionOfModule[Uqgl[m, n], RepName] =
  Module[
    {rho0, eprs, D},

    rho0 = HalfSumofPositiveRoots[gl[m, n], 0];
    eprs = RootSystem[gl[m, n], {0}, {+1}];

    D = 2^(m n) *
      Apply[
        Times,
        Map[
          GradedInnerProductOfAlgebraWeights[gl[m,n],RepName+rho0,#] / \
GradedInnerProductOfAlgebraWeights[gl[m,n],rho0,#]&, eprs
        ]
      ];

    D
  ]) /; SoberAlgebraWeightQ[Uqgl[m, n], RepName];

(* ----------------------------------------------------------------- *)

ConvertFromRepresentationToInternalVariables[Uqgl[m_, n_], RepName_, S_] :=
  PowerExpand[
    S /.
      {
        q^((i_:1) alpha + j_) -> p^i q^(+ j - i (m-n)/2),
        q^((i_:1) alpha     ) -> p^i q^(    - i (m-n)/2)
      }
  ] /; (SoberAlgebraWeightQ[Uqgl[m, n], RepName] && (RepName == \
LGIHighestWeight[Uqgl[m,n]]));

(* ----------------------------------------------------------------- *)

ConvertFromInternalToRepresentationVariables[Uqgl[m_, n_], RepName_, S_] :=
  (S /. p -> q^(alpha+(m-n)/2)) /;
    (SoberAlgebraWeightQ[Uqgl[m, n], RepName] && (RepName == \
LGIHighestWeight[Uqgl[m,n]]));

(* ----------------------------------------------------------------- *)

(* This code is hardwired for experiments with simple generators *)

NewGeneratorProductsForBasisVectorsOfModule[Uqgl[2,2], RepName_] :=
  (NewGeneratorProductsForBasisVectorsOfModule[Uqgl[2,2], RepName] =
  Module[
    {Id, E43, E32, E21, gpl},

    Id  = Generator[Uqgl[2,2], 1, 1, 0];
    E43 = Generator[Uqgl[2,2], 4, 3, 1];
    E32 = Generator[Uqgl[2,2], 3, 2, 1];
    E21 = Generator[Uqgl[2,2], 2, 1, 1];
    gpl =
    {
      Id,
      E32,
      E43 ** E32,
      E21 ** E32,
      E21 ** E43 ** E32,
      E32 ** E43 ** E32,
      E32 ** E21 ** E32,
      E21 ** E32 ** E43 ** E32,
      E43 ** E32 ** E21 ** E32,
      E43 ** E21 ** E32 ** E43 ** E32,
      E21 ** E43 ** E32 ** E21 ** E32,
      E32 ** E21 ** E32 ** E43 ** E32,
      E43 ** E32 ** E21 ** E32 ** E43 ** E32,
      E21 ** E32 ** E21 ** E32 ** E43 ** E32,
      E21 ** E43 ** E32 ** E21 ** E32 ** E43 ** E32,
      E21 ** E32 ** E43 ** E32 ** E21 ** E32 ** E43 ** E32
    };

    gpl
  ]) /; (RepName == LGIHighestWeight[Uqgl[2,2]]);

Needs["DiscreteMath`Combinatorica`"];

GeneratorProductsForBasisVectorsOfModule[Uqgl[m_, n_], RepName_] :=
  (GeneratorProductsForBasisVectorsOfModule[Uqgl[m,n], RepName] =
  Module[
    {olgs, gpl},

    olgs = GeneratorSet[Uqgl[m, n], {1}, {-1}];
    olgs = Reverse[olgs];
    gpl = Table[KSubsets[olgs, i], {i, 0, Length[olgs]}];
    gpl[[1]] = {{Generator[Uqgl[m, n], 1, 1, 0]}};
    gpl = Map[Reverse, gpl, {2}];
    gpl = Apply[NonCommutativeMultiply, gpl, {2}];
    gpl = Flatten[gpl];

    gpl
  ]) /; (SoberAlgebraWeightQ[Uqgl[m, n], RepName] && (RepName == \
LGIHighestWeight[Uqgl[m,n]]));

(* ----------------------------------------------------------------- *)

SoberBasisVectorOfModule[Algebra_, RepName_, ket[k_Integer?Positive]] :=
  (SoberAlgebraWeightQ[Algebra, RepName] && (k <= DimensionOfModule[Algebra, \
\
RepName]));


GradingOfBasisVectorOfModule[Algebra_, RepName_, ket[1]] :=
  (GradingOfBasisVectorOfModule[Algebra, RepName, ket[1]] =
    0
  ) /; SoberBasisVectorOfModule[Algebra, RepName, ket[1]];

GradingOfBasisVectorOfModule[Algebra_, RepName_, ket[k_Integer?Positive]] :=
  (GradingOfBasisVectorOfModule[Algebra, RepName, ket[k]] =
  Module[
    {grl, gr},

    grl = Map[GradingOfAlgebraGenerator, Apply[List, \
GeneratorProductsForBasisVectorsOfModule[Algebra, RepName][[k]]]];
    gr = Mod[Apply[Plus, grl], 2];

    gr
  ]) /; (SoberBasisVectorOfModule[Algebra, RepName, ket[k]] && (k =!= 1));

GradingOfBasisVectorOfModule[Algebra_, RepName_, bra[k_]] :=
  GradingOfBasisVectorOfModule[Algebra, RepName, ket[k]] /;
    SoberBasisVectorOfModule[Algebra, RepName, ket[k]];

(* ----------------------------------------------------------------- *)

GradingListOfBasisVectorsOfModule[Algebra_, RepName_] :=
  (GradingListOfBasisVectorsOfModule[Algebra, RepName] =
    Map[
      GradingOfBasisVectorOfModule[Algebra, RepName, {{e[{},{#}],1}}]&,
      Range[DimensionOfModule[Algebra, RepName]]
    ]
  ) /; SoberAlgebraWeightQ[Algebra, RepName];

(* ----------------------------------------------------------------- *)

WeightOfBasisVectorOfModule[Algebra_, RepName_, ket[1]] :=
  (WeightOfBasisVectorOfModule[Algebra, RepName, ket[1]] =
    RepName
  ) /; SoberBasisVectorOfModule[Algebra, RepName, ket[1]];

WeightOfBasisVectorOfModule[Algebra_, RepName_, ket[k_Integer?Positive]] :=
  (WeightOfBasisVectorOfModule[Algebra, RepName, ket[k]] =
  Module[
    {gplk, wt},

    gplk = GeneratorProductsForBasisVectorsOfModule[Algebra, RepName][[k]];
    wt = RepName + AdjointWeightOfAlgebraGenerator[gplk];

    wt
  ]) /; (SoberBasisVectorOfModule[Algebra, RepName, ket[k]] && (k =!= 1));

WeightOfBasisVectorOfModule[Algebra_, RepName_, bra[k_Integer?Positive]] :=
  WeightOfBasisVectorOfModule[Algebra, RepName, ket[k]] /;
    SoberBasisVectorOfModule[Algebra, RepName, ket[k]];

(* ----------------------------------------------------------------- *)

AlgebraModuleKetActionRules[Algebra_, RepName_] :=
  (AlgebraModuleKetActionRules[Algebra, RepName] =
    {
      W_NonCommutativeMultiply :> NonCommutativeMultiply[Generator[Algebra, \
\
1, 1, 0]] /; (Length[W]==0),
      W_NonCommutativeMultiply :> 0 /;
        (
          (LiftingOfAlgebraGenerator[W[[-1]]]==1)
          ||
          ((LiftingOfAlgebraGenerator[W[[-1]]]==-1) && \
(GradingOfAlgebraGenerator[W[[-1]]]==0))
        ),
      W_NonCommutativeMultiply :>
        qSub[Algebra, W[[-1,2]]]^(W[[-1,4]] * RepName[[W[[-1,2]]]]) * Drop[W, \
\
-1] /;
          ((W[[-1,2]]==W[[-1,3]]) && (W[[-1,4]]=!=0))
    }
  ) /; SoberAlgebraWeightQ[Algebra, RepName];

ApplyAMKARules[Algebra_, RepName_, W_] :=
  (ApplyAMKARules[Algebra, RepName, W] =
  Module[
    {Wout},

    Wout = ExpandAll[W //. AlgebraModuleKetActionRules[Algebra, RepName]];
    Wout = Collect[Wout, NonCommutativeMultiply[__], \
SymmetriseqExpressions[Factor[#]]&];

    Wout
  ]) /; SoberAlgebraWeightQ[Algebra, RepName];

(* ----------------------------------------------------------------- *)

AlgebraModuleBraActionRules[Algebra_, RepName_] :=
  (AlgebraModuleBraActionRules[Algebra, RepName] =
    {
      W_NonCommutativeMultiply :> NonCommutativeMultiply[Generator[Algebra, \
\
1, 1, 0]] /; (Length[W]==0),
      W_NonCommutativeMultiply :> 0 /;
        (
          (LiftingOfAlgebraGenerator[W[[+1]]]==-1)
          ||
          ((LiftingOfAlgebraGenerator[W[[+1]]]==+1)&&(\
GradingOfAlgebraGenerator[W[[+1]]]==0))
        ),
      W_NonCommutativeMultiply :>
        qSub[Algebra, W[[+1,2]]]^(W[[+1,4]] * RepName[[W[[+1,2]]]]) * Drop[W, \
\
+1] /;
          ((W[[+1,2]]==W[[+1,3]])&&(W[[+1,4]]=!=0))
    };
  ) /; SoberAlgebraWeightQ[Algebra, RepName];

ApplyAMBARules[Algebra_, RepName_, W_] :=
  (ApplyAMBARules[Algebra, RepName, W] =
  Module[
    {Wout},

    Wout = W //. AlgebraModuleBraActionRules[Algebra, RepName];
    Wout = ExpandAll[Wout];
    Wout = Collect[Wout, NonCommutativeMultiply[__], \
SymmetriseqExpressions[Factor[#]]&];

    Wout
  ]) /; SoberAlgebraWeightQ[Algebra, RepName];

(* ----------------------------------------------------------------- *)

NOK[Algebra_, RepName_, W_Plus] :=
  Module[
    {nokW},

    nokW = Map[NOK[Algebra, RepName, #]&, W];
    nokW = Collect[Expand[nokW], NonCommutativeMultiply[_], \
SymmetriseqExpressions[Factor[#]]&];

    nokW
  ] /; SoberAlgebraWeightQ[Algebra, RepName];

NOK[Algebra_, RepName_, a_ W_] :=
  a * NOK[Algebra, RepName, W] /;
    (SoberAlgebraWeightQ[Algebra, RepName] && !OperatorQ[a]);

NOK[Algebra_, RepName_, W_] :=
  (NOK[Algebra, RepName, W] =
  Module[
    {Wold, Wnew, i},

    Wold = 0; Wnew = W; i = 0;
(*    WriteString["stdout", "NOK[",W,"]:\n"]; *)
    Print["NOK[",W,"]"];
    While[
      ((Wnew =!= Wold) && (Length[Wnew]=!=0)),
      Wold = Wnew;
      Wnew = ApplyNORules[Algebra, Wnew];
      Wnew = ApplyAMKARules[Algebra, RepName, Wnew];
      i++
(* ;
      WriteString[
        "stdout",
        "  ", i, ": Wnew contains ", Length[Wnew], " terms and ", \
ByteCount[Wnew], " bytes.\n"
      ]
*)
    ];

    Wnew
  ]) /; SoberAlgebraWeightQ[Algebra, RepName];

(* ----------------------------------------------------------------- *)

NOB[Algebra_, RepName_, W_Plus] :=
  Module[
    {nobW},

    nobW = Map[NOB[Algebra, RepName, #]&, W];
    nobW = Collect[Expand[nobW], NonCommutativeMultiply[_], \
SymmetriseqExpressions[Factor[#]]&];

    nobW
  ] /; SoberAlgebraWeightQ[Algebra, RepName];

NOB[Algebra_, RepName_, a_ W_] :=
  a * NOB[Algebra, RepName, W] /;
    (SoberAlgebraWeightQ[Algebra, RepName] && !OperatorQ[a]);

NOB[Algebra_, RepName_, W_] :=
  (NOB[Algebra, RepName, W] =
  Module[
    {Wold, Wnew, i},

    Wold = 0; Wnew = W; i = 0;
    WriteString["stdout", "NOB[",W,"]:\n"];
    While[
      ((Wnew =!= Wold) && (Length[Wnew]=!=0)),
      Wold = Wnew;
      Wnew = ApplyNORules[Algebra, Wnew];
      Wnew = ApplyAMBARules[Algebra, RepName, Wnew];
      i++;
      WriteString[
        "stdout",
        "  ", i, ": Wnew contains ", Length[Wnew], " terms and ", \
ByteCount[Wnew], " bytes.\n"
      ]
    ];

    Wnew
  ]) /; SoberAlgebraWeightQ[Algebra, RepName];

(* ----------------------------------------------------------------- *)

NOBK[Algebra_, RepName_, W_Plus] :=
  Module[
    {nobkW},

    nobkW = Map[NOBK[Algebra, RepName, #]&, W];
    nobkW = Collect[Expand[nobkW], NonCommutativeMultiply[_], \
SymmetriseqExpressions[Factor[#]]&];
    nobkW = SymmetriseqExpressions[Factor[nobkW]];

    nobkW
  ] /; SoberAlgebraWeightQ[Algebra, RepName];

NOBK[Algebra_, RepName_, a_ W_] :=
  a * NOBK[Algebra, RepName, W] /;
    (SoberAlgebraWeightQ[Algebra, RepName] && !OperatorQ[a]);

NOBK[Algebra_, RepName_, W_] :=
  (NOBK[Algebra, RepName, W] =
  Module[
    {Wold, Wnew, i},

    Wold = 0; Wnew = W; i = 0;
(*    WriteString["stdout", "NOBK[",W,"]:\n"]; *)
    While[
      ((Wnew =!= Wold) && (Length[Wnew]=!=0)),
      Wold = Wnew;
      Wnew = ApplyNORules[Algebra, Wnew];
      Wnew = ApplyAMKARules[Algebra, RepName, Wnew];
      Wnew = ApplyAMBARules[Algebra, RepName, Wnew];
      i++
(*   ;   WriteString[
        "stdout",
        "  ", i, ": Wnew contains ", Length[Wnew], " terms and ", \
ByteCount[Wnew], " bytes.\n"
      ] *)
    ];
    Wnew = Wnew /. NonCommutativeMultiply[Generator[Algebra, 1, 1, 0]] :> 1;
    Wnew = SymmetriseqExpressions[Factor[Wnew]];

    Wnew
  ]) /; (SoberAlgebraWeightQ[Algebra, RepName] && (RepName == \
LGIHighestWeight[Algebra]));

(* ----------------------------------------------------------------- *)

NOBKusingNO[Algebra_, RepName_, W_Plus] :=
  Module[
    {nobkW},

    nobkW = Map[NOBKusingNO[Algebra, RepName, #]&, W];
    nobkW = Collect[Expand[nobkW], NonCommutativeMultiply[_], \
SymmetriseqExpressions[Factor[#]]&];
    nobkW = SymmetriseqExpressions[Factor[nobkW]];

    nobkW
  ] /; SoberAlgebraWeightQ[Algebra, RepName];

NOBKusingNO[Algebra_, RepName_, a_ W_] :=
  a * NOBKusingNO[Algebra, RepName, W] /;
    (SoberAlgebraWeightQ[Algebra, RepName] && !OperatorQ[a]);

(* input W is presumed to be an unweighted product of generators *)

NOBKusingNO[Algebra_, RepName_, 0] :=
  0 /; SoberAlgebraWeightQ[Algebra, RepName];

NOBKusingNO[Algebra_, RepName_, NonCommutativeMultiply[Generator[Algebra_, 1, \
\
1, 0]]] :=
  (
    NOBKusingNO[Algebra, RepName, NonCommutativeMultiply[Generator[Algebra, \
\
1, 1, 0]]] =
    1
  ) /; (SoberAlgebraWeightQ[Algebra, RepName] && \
SoberAlgebraGeneratorQ[Generator[Algebra, 1, 1, 0]]);

NOBKusingNO[Algebra_, RepName_, W_NonCommutativeMultiply?((Length[#] > 1) &)] \
\
 :=
  (NOBKusingNO[Algebra, RepName, W] =
  Module[
    {i, l, swapped, Wint, disq, Whead, Wtail, Wcurrent, Wcurrentswapped, \
Wout, Wswapped},

    i = 1; swapped = False; l = Length[W];
(*    WriteString["stdout", "NOBKusingNO[", W, "]:\n"]; *)

    Wint = W;
    Wint = ApplyAMKARules[Algebra, RepName, Wint];
    Wint = ApplyAMBARules[Algebra, RepName, Wint];
    If[
      (Wint===W),
      While[
        (!swapped&&(i<l)),
        Wcurrent = Take[Wint, {i, i+1}];
        disq = Apply[DisorderedGeneratorsQ, Wcurrent];
        If[disq,
          Wcurrentswapped = PBWLemmaSwapperFunction[Take[Wint, {i,i+1}]];
          Whead = Take[Wint, i-1]; Wtail = Drop[Wint, i+1];
          Wswapped = Wcurrentswapped;
          If[(Length[Whead] =!= 0), Wswapped = Whead ** Wswapped];
          If[(Length[Wtail] =!= 0), Wswapped = Wswapped ** Wtail];
          swapped = True;
          Wswapped = ExpandAll[Wswapped];
          Wswapped = Collect[Wswapped, NonCommutativeMultiply[__], \
SymmetriseqExpressions[Factor[#]]&];
          Wswapped = MapAll[Distribute[#, Plus, NonCommutativeMultiply]&, \
Wswapped];
          Wswapped = Collect[Wswapped, NonCommutativeMultiply[__], \
SymmetriseqExpressions[Factor[#]]&];
          Wout = NOBKusingNO[Algebra, RepName, Wswapped],
          i++
        ]
      ];
      If[!swapped, Wout = Wint],
      Wout = NOBKusingNO[Algebra, RepName, Wint]
    ];
    Wout = SymmetriseqExpressions[Factor[Wout]];

    Wout
  ]) /; (SoberAlgebraWeightQ[Algebra, RepName] && (RepName == \
LGIHighestWeight[Algebra]));

(* ----------------------------------------------------------------- *)

NOBKC[Algebra_, RepName_, W_Plus] :=
  Module[
    {nobkcW},

    nobkcW = Map[NOBKC[Algebra, RepName, #]&, W];
    nobkcW = Collect[Expand[nobkcW], NonCommutativeMultiply[_], \
SymmetriseqExpressions[Factor[#]]&];
    nobkcW = SymmetriseqExpressions[Factor[nobkcW]];

    nobkcW
  ] /; SoberAlgebraWeightQ[Algebra, RepName];

NOBKC[Algebra_, RepName_, a_ W_] :=
  a * NOBKC[Algebra, RepName, W] /;
    (SoberAlgebraWeightQ[Algebra, RepName] && !OperatorQ[a]);

NOBKC[Algebra_, RepName_, W_] :=
  (NOBKC[Algebra, RepName, W] =
  Module[
    {nobkcout, nocW},

    WriteString["stdout","NOBKC[", W, "]\n"];
    nocW = NOC[Algebra, W];
    nocW = Distribute[nocW];
    nocW = Collect[Expand[nocW], NonCommutativeMultiply[_], \
SymmetriseqExpressions[Factor[#]]&];
    nobkcout = NOBK[Algebra, RepName, nocW];

    nobkcout
  ]) /; (SoberAlgebraWeightQ[Algebra, RepName] && (RepName == \
LGIHighestWeight[Algebra]));

(* ----------------------------------------------------------------- *)

NOKL[Algebra_, RepName_, W_Plus] :=
  Module[
    {noklW},

    noklW = Map[NOKL[Algebra, RepName, #]&, W];
    noklW = Collect[Expand[noklW], NonCommutativeMultiply[_], \
SymmetriseqExpressions[Factor[#]]&];

    noklW
  ] /; SoberAlgebraWeightQ[Algebra, RepName];

NOKL[Algebra_, RepName_, a_ W_] :=
  a * NOKL[Algebra, RepName, W] /;
    (SoberAlgebraWeightQ[Algebra, RepName] && !OperatorQ[W]);

NOKL[Algebra_, RepName_, W_] :=
  (NOKL[Algebra, RepName, W] =
    Module[
      {noklout, nolW},

      WriteString["stdout","NOKL[", W, "]\n"];
      nolW = NOL[Algebra, W];
      nolW = Distribute[nolW];
      nolW = Collect[Expand[nolW], NonCommutativeMultiply[_], \
SymmetriseqExpressions[Factor[#]]&];
      noklout = NOK[Algebra, RepName, nolW];

      noklout
  ]) /; (SoberAlgebraWeightQ[Algebra, RepName] && (RepName == \
LGIHighestWeight[Algebra]));

(* ----------------------------------------------------------------- *)

(* Hard-wired version, as it's expensive *)

MetricMatrixofBPrime[Uqgl[m_, 1], RepName_] :=
  (MetricMatrixofBPrime[Uqgl[m, 1], RepName] =
  Module[
    {gpl, plusgpl, p, inds, FF},

    gpl = GeneratorProductsForBasisVectorsOfModule[Uqgl[m,1], RepName];
    gpl[[1]] = 0;
    plusgpl = Map[Apply[Plus, #]&, gpl];
    inds = plusgpl /. Generator[Uqgl[m, 1], m+1, aj_, 1] :> - 2 * aj;
    p = Map[Length, gpl];
    FF = q^(inds + p(2 m - p + 1)) * Map[Product[qBracket[alpha+j-1], {j, \
#}]&, p];
    FF = DiagonalMatrix[ExpandqBrackets[FF]];

    FF
  ]) /; (SoberAlgebraWeightQ[Uqgl[m,1], RepName] && (RepName == \
LGIHighestWeight[Uqgl[m,1]]));

(*
  For some reason, using NOBKC in MetricMatrixofBPrime is slower than using \
\
NOBK, at least for
  U_q[gl(3|1)], whereas it should theoretically be faster. Compare with the \
\
use of NOKL versus NOK
  in RepresentationPrime.
*)

Needs["LinearAlgebra`MatrixManipulation`"];

MetricMatrixofBPrime[Algebra_, RepName_] :=
  (MetricMatrixofBPrime[Algebra, RepName] =
  Module[
    {gpl, gplconj, MM, metmat, Z, ZT},

    gpl = GeneratorProductsForBasisVectorsOfModule[Algebra, RepName];
    gpl = Split[gpl, (AdjointWeightOfAlgebraGeneratorProduct[#1] == \
AdjointWeightOfAlgebraGeneratorProduct[#2])&];
    gplconj = Map[HermitianConjugate[Algebra, #]&, gpl, {2}];
(*
  Should rewrite ExpandAllAltGeneratorsToStandardGeneratorsUsingZhangsLemma \
\
so that the
  following isn't necessary.
*)
    gplconj = gplconj /. NonCommutativeMultiply[W_AltGenerator] :> W;
    gplconj = \
ExpandAllAltGeneratorsToStandardGeneratorsUsingZhangsLemma[Algebra, gplconj];
    gplconj = gplconj /. W_Generator :> NonCommutativeMultiply[W];
    gplconj = Map[NormalOrder[Algebra, #]&, gplconj, {2}];
    MM = MapThread[Outer[NonCommutativeMultiply, #1, #2]&, {gplconj, gpl}];
    MM = MapAll[Distribute[#, Plus, NonCommutativeMultiply]&, MM];
    MM = Map[NOBK[Algebra, RepName, #]&, MM, {3}];
    MM = SymmetriseqExpressions[Factor[MM]];
    metmat = MM[[1]];
    Do[
      Z = ZeroMatrix[Length[metmat], Length[MM[[i]]]];
      ZT = Transpose[Z];
      metmat = BlockMatrix[{{metmat, Z}, {ZT,  MM[[i]]}}],
      {i, 2, Length[MM]}
    ];

    metmat
  ]) /; (SoberAlgebraWeightQ[Algebra, RepName] && (RepName == \
LGIHighestWeight[Algebra]));

(* ----------------------------------------------------------------- *)

ChangeOfBasisMatrix[Algebra_, RepName_] :=
  (ChangeOfBasisMatrix[Algebra, RepName] =
  Module[
    {T, PP, DD},

    {PP, DD} = JordanDecomposition[MetricMatrixofBPrime[Algebra, RepName]];
    T = Dot[MatrixPower[DD, -1/2], Inverse[PP]];
    T = MapAll[SymmetriseqExpressions, T];

    T
  ]) /; (SoberAlgebraWeightQ[Algebra, RepName] && (RepName == \
LGIHighestWeight[Algebra]));

(* ----------------------------------------------------------------- *)

(* An easy formula! *)

RepresentationPrime[Algebra_, RepName_, Generator[Algebra_, c_, c_, N_]] :=
  (RepresentationPrime[Algebra, RepName, Generator[Algebra, c, c, N]] =
  Module[
    {gpl, gplsums, gplexps, repprime},

    gpl = GeneratorProductsForBasisVectorsOfModule[Algebra, RepName];
    gpl = Map[Apply[List, #]&, gpl, 1];
    gplsums = gpl /. Generator[Algebra, a_, b_, _] :> KroneckerDelta[c, a] - \
\
KroneckerDelta[c, b];
    gplexps = RepName[[c]] + Apply[Plus, gplsums, 2];
    repprime = PowerExpand[qSub[Algebra, c]^(N * gplexps)];
    repprime = DiagonalMatrix[repprime];

    repprime
  ]) /;
  (
    SoberAlgebraWeightQ[Algebra, RepName] && (RepName == \
LGIHighestWeight[Algebra])
    &&
    SoberAlgebraGeneratorQ[Generator[Algebra, c, c, N]]
  );

(* This part of the function shouldn't need to be called. *)

RepresentationPrime[Algebra_, RepName_, Generator[Algebra_, a_, b_, 1]] :=
  (RepresentationPrime[Algebra, RepName, Generator[Algebra, a, b, 1]] =
  Module[
    {expG, reprule, rep},

    expG = ExpandNonSimpleGeneratorOnce[Generator[Algebra, a, b, 1]];
    reprule =
    {
      Generator[Algebra, c_, d_, N_] :> Hold[RepresentationPrime[Algebra, \
RepName, Generator[Algebra, c, d, N]]],
      NonCommutativeMultiply :> Dot
    };
    rep = ReleaseHold[expG /. reprule];

    rep
  ]) /;
  (
    SoberAlgebraWeightQ[Algebra, RepName] && (RepName == \
LGIHighestWeight[Algebra]) &&
    NonSimpleAlgebraGeneratorQ[Generator[Algebra, a, b, 1]]
  );

(*
  Again, as with using NOBKC in MetricMatrixofBPrime, using NOKL instead
  of NOK here is slower, at least for U_q[gl(3|1)], but it should be faster!
*)

RepresentationPrime[Algebra_, RepName_, Generator[Algebra_, a_, b_, 1]] :=
  (RepresentationPrime[Algebra, RepName, Generator[Algebra, a, b, 1]] =
  Module[
    {G, gpl, Ggpl, repprime},

    G = Generator[Algebra, a, b, 1];
    gpl = GeneratorProductsForBasisVectorsOfModule[Algebra, RepName];
    Ggpl = Map[(G ** #)&, gpl];
    Ggpl = Map[NOK[Algebra, RepName, #]&, Ggpl];
    repprime = Outer[Coefficient, Ggpl, gpl];

    repprime
  ]) /;
  (
    SoberAlgebraWeightQ[Algebra, RepName] && (RepName == \
LGIHighestWeight[Algebra])
    &&
    SimpleAlgebraGeneratorQ[Generator[Algebra, a, b, 1]]
  );

(* ----------------------------------------------------------------- *)

Representation[Algebra_, RepName_, Generator[Algebra_, a_, b_, 1]] :=
  (Representation[Algebra, RepName, Generator[Algebra, a, b, 1]] =
  Module[
    {expG, reprule, rep},

    expG = ExpandNonSimpleGeneratorOnce[Generator[Algebra, a, b, 1]];
    reprule = {
      Generator[Algebra, c_, d_, N_] :> Hold[Representation[Algebra, RepName, \
\
Generator[Algebra, c, d, N]]],
      NonCommutativeMultiply :> Dot
    };
    rep = ReleaseHold[expG /. reprule];

    rep
  ]) /;
  (
    SoberAlgebraWeightQ[Algebra, RepName] && (RepName == \
LGIHighestWeight[Algebra])
    &&
    NonSimpleAlgebraGeneratorQ[Generator[Algebra, a, b, 1]]
  );

Representation[Algebra_, RepName_, Generator[Algebra_, a_, b_, 1]] :=
  (Representation[Algebra, RepName, Generator[Algebra, a, b, 1]] =
  Module[
    {repofconj, rep},

    repofconj = Representation[Algebra, RepName, Generator[Algebra, b, a, \
1]];
    rep = Transpose[repofconj] /. q -> q^(-1);
    rep = PowerExpand[rep];

    rep
  ]) /;
  (
    SoberAlgebraWeightQ[Algebra, RepName] && (RepName == \
LGIHighestWeight[Algebra])
    &&
    SimpleAlgebraGeneratorQ[Generator[Algebra, a, b, 1]]
    &&
    (LiftingOfAlgebraGenerator[Generator[Algebra, a, b, 1]] == -1)
  );

Representation[Algebra_, RepName_, Generator[Algebra_, a_, a_, N_]] :=
  (Representation[Algebra, RepName, Generator[Algebra, a, a, N]] =
  Module[
    {rep},

    rep = Representation[Algebra, RepName, Generator[Algebra, a, a, 1]] /. q \
\
-> q^N;
    rep = PowerClean[PowerExpand[rep]];

    rep
  ]) /;
  (
    (N =!= 1)
    &&
    SoberAlgebraWeightQ[Algebra, RepName] && (RepName == \
LGIHighestWeight[Algebra])
    &&
    SoberAlgebraGeneratorQ[Generator[Algebra, a, a, N]]
  );

Representation[Algebra_, RepName_, Generator[Algebra_, a_, b_, 1]] :=
  (Representation[Algebra, RepName, Generator[Algebra, a, b, 1]] =
  Module[
    {repprime, T, rep},

    repprime = RepresentationPrime[Algebra, RepName, Generator[Algebra, a, b, \
\
1]];
    T = ChangeOfBasisMatrix[Algebra, RepName];
    rep = T . repprime . Inverse[T];
    rep = MapAll[SymmetriseqExpressions[Factor[#]]&, rep];

    rep
  ]) /;
  (
    SoberAlgebraWeightQ[Algebra, RepName] && (RepName == \
LGIHighestWeight[Algebra])
    &&
    SimpleAlgebraGeneratorQ[Generator[Algebra, a, b, 1]]
    &&
    (LiftingOfAlgebraGenerator[Algebra, Generator[Algebra, a, b, 1]] =!= -1)
  );

(* ----------------------------------------------------------------- *)

EigenvalueOfFirstOrderCasimir[Uqgl[m_, n_], RepName_] :=
  (EigenvalueOfFirstOrderCasimir[Uqgl[m, n], RepName] =
  Module[
    {rep, eig},

    rep = Apply[Dot,
      Map[
        Representation[Uqgl[m, n], RepName, Generator[Uqgl[m,n], #, #, \
(-1)^(GradingOfAlgebraIndex[gl[m,n], #])]]&,
        Range[m + n]
      ]];
    eig = rep[[1,1]];

    eig
  ]) /; (SoberAlgebraWeightQ[Uqgl[m, n], RepName] && (RepName == \
LGIHighestWeight[Uqgl[m,n]]));

(* ----------------------------------------------------------------- *)

RepresentationOfqTo2hrho[Uqgl[m_, n_], RepName_] :=
  (RepresentationOfqTo2hrho[Uqgl[m, n], RepName] =
  Module[
    {rep, theta, eig},

    rep =
      Apply[Dot,
        Join[
          Map[Representation[Uqgl[m,n], RepName, \
Generator[Uqgl[m,n],#,#,2(m-#)]]&, Range[m]],
          Map[Representation[Uqgl[m,n], RepName, \
Generator[Uqgl[m,n],m+#,m+#,2(1-#)]]&, Range[n]]
        ]
      ];
    eig = EigenvalueOfFirstOrderCasimir[Uqgl[m,n], RepName];
    theta = 0;
    rep = Expand[eig^theta * rep];

    rep
  ]) /; (SoberAlgebraWeightQ[Uqgl[m, n], RepName] && (RepName == \
LGIHighestWeight[Uqgl[m,n]]));

(* ----------------------------------------------------------------- *)

StripGradingFromSparseRank2Array[Algebra_, RepName_, S_] :=
  Module[
    {gr, Sstrip},

    gr = GradingListOfBasisVectorsOfModule[Algebra, RepName];
    Sstrip = S /. {e[{i_}, {j_}], W_} :> {e[{i}, {j}], (-1)^gr[[i]] * W};

    Sstrip
  ] /; (SoberAlgebraWeightQ[Uqgl[m, n], RepName] && (RepName == \
LGIHighestWeight[Uqgl[m,n]]));

(* ----------------------------------------------------------------- *)

(* End[]; *)

EndPackage[];


