(* -----------------------------------------------------------------

  Sparse Tensors.m

  Functions dealing with sparse tensors. Importantly, this code was written 
  for Mathematica 4.0, so does not use any of the native SparseTensor 
  conventions available in Mathematica 5.0. Everything is done internally.

  A basis vector e^{i,j,...}_{k,l,...} is represented by 
  e[{i,j,...},{k,l,...}], so in particular, ket{i} equiv e_i is e[{},{i}], \
its 
  dual bra{i} equiv e^i is e[{i},{}], and the dyad ket{i} bra{j} equiv e^j_i 
  is e[{j},{i}].

  A linear combination sum theta^{i,j,...}_{k,l,...} e^{i,j,...}_{k,l,...} is \
\
\

  then a list {..., {e[{i,j,...},{k,l,...}],theta^{i,j,...}_{k,l,...}}, ... \
\
\
}.

  For example, ket{k} bra{i} otimes ket{l} bra{j} equiv e^{i,j}_{k,l} is 
  e[{i,j},{k,l}].

  Care is taken when multiplying tensor products of graded basis vectors, in 
  particular note that e_{k,l} e^{i,j} = (-)^{[i][l]} e^{i,j}_{k,l} which is 
  e[{},{k,l}] e[{i,j},{}] -> (-1)^{[i][l]} e[{i,j},{k,l}].

  ------------------------------------------------------------------ *)

BeginPackage["LinksGouldExplorer`"];

SoberSparseTensorQ::usage = "
  SoberSparseTensorQ[Z] tests to see whether a claimed sparse tensor Z \
appears 
  to be well-defined. It incorporates the convention that sparse tensors are 
  recorded as components, each of the form {e[{i,j,k,...},{l,m,n,...}], 
  Z^{i,j,k,...}_{l,m,n,...}}.
";

RankOfSparseTensor::usage = "
  RankOfSparseTensor[Z] is the rank {u,l} of a sparese tensor Z; that is the 
  number of upper and lower indices in Z.
";

UpperRankOfSparseTensor::usage = "
  UpperRankOfSparseTensor[Z] is the number of upper indices in Z.
";

LowerRankOfSparseTensor::usage = "
  LowerRankOfSparseTensor[Z] is the number of lower indices in Z.
";

SparseTensorProductOfSparseTensors::usage = "
  SparseTensorProductOfSparseTensors[W,Y] forms the tensor product W otimes Y \
\
\

  (not the multiplication WY) of sparse tensors W and Y.
";

IdentitySparseTensor::usage = "
  IdentitySparseTensor[M, N] is the identity sparse tensor of rank N, where 
  each of the N dimensions is M.
";

ZeroSparseTensor::usage = "
  ZeroSparseTensor[rank] is the zero sparse identity tensor of the given \
rank.
";

TidySparseTensorComponents::usage = "
  TidySparseTensorComponents[Z] takes a sparse tensor Z, sorts components \
into 
  order by index, combines those with common coefficients, simplifies those 
  coefficients, and purges any zero elements. If the result is empty, it 
  replaces Z with the appropriate zero sparse tensor.
";

AccreteSparseTensorXToSparseTensorZatStrings::usage = "
  AccreteSparseTensorXToSparseTensorZatStrings[Z, W, s] is the product Z W', 
  where:

    W' = (I^{otimes (s-1)} otimes W otimes I^{otimes (wZ-s-wW+1)}),

  where wZ (wW) is the width of Z (W). (So, if W is wider than Z, we first 
  widen Z.) If the ranks of Z and W match, and s=1, then the product is just \
\
\
Z 
  W. Its performance is critical to the time requirements of the state model 
  algorithms, so effort has been taken to ensure that it is not woefully 
  inefficient.
";

ContractSparseTensorZOverSparseTensorHOnStrings::usage = "
  ContractSparseTensorZOverSparseTensorHOnStrings[Z, H, s] contracts tensor Z \
\
\

  over handle H along string s. It exploits the fact that such (handles) H \
are 
  diagonal.
";


ConvertRank4ArrayToRank4SparseTensor::usage = "
  ConvertRank4ArrayToRank4SparseTensor[R] builds a rank 4 sparse tensor out \
\
\
of 
  the rank 4 array object R.
";

ConvertRank4SparseTensorToRank4Array::usage = "
  ConvertRank4SparseTensorToRank4Array[R, D] builds a rank 4 array out of the \
\
\

  sparse rank 4 tensor object R, where each index runs from 1 to D.
";

ConvertRank2ArrayToRank2SparseTensor::usage = "
  ConvertRank2ArrayToRank2SparseTensor[r2a].
";

ConvertRank2SparseTensorToRank2Array::usage = "
  ConvertRank2SparseTensorToRank2Array[R, D] builds a rank 2 array out of the \
\
\

  sparse rank 2 tensor object R, where each index runs from 1 to D.
";


DecrementRankOfTensorObject::usage = "
  DecrementRankOfTensorObject[T] takes an array T of rank at least 2, and 
  replaces it with an isomorphic copy such that the forms printed on the page \
\
\

  are identical.
";

ConvertRank4ArrayObjectToRank2ArrayObject::usage = "
  ConvertRank4ArrayObjectToRank2ArrayObject[r4a].
";

ConvertRank2ArrayObjectToRank4ArrayObject::usage = "
  ConvertRank2ArrayObjectToRank4ArrayObject[r2a].
";


BraidRelationTestForRank4ArrayObject::usage = "
  BraidRelationTestForRank4ArrayObject[r4a] is True if r4A satisfies the \
braid 
  relation.
";

BraidRelationTestForRank2ArrayObject::usage = "
  BraidRelationTestForRank2ArrayObject[r2a] is True if r2A satisfies the \
braid 
  relation.
";

(* ------------------------------------------------------------------ *)

(* Begin["`Private`"]; *)

(* Use the first form for debugging purposes *)

SoberSparseTensorQ[Z_List] :=
  Module[
    {indices, widths, indexset, samewidthsok, depthok, posintsok, ans},

    depthok = (Depth[Z] >= 5);
    indices = Apply[List, Z[[All, 1]], 1];
    widths = Map [Length, indices, {2}];
    indexset = Union[Flatten[indices]];
    samewidthsok = (Length[Union[Flatten[widths]]] == 1);
    posintsok = Apply[And, Map[(NumberQ[#] && Positive[#])&, indexset]];
    ans = (depthok && samewidthsok && posintsok);

    ans
  ];

SoberSparseTensorQ[Z_List] := True;

(* ------------------------------------------------------------------ *)

RankOfSparseTensor[Z_List] :=
  Map[Length, Apply[List, Z[[1, 1]]]];

UpperRankOfSparseTensor[Z_List] :=
  RankOfSparseTensor[Z][[1]];

LowerRankOfSparseTensor[Z_List] :=
  RankOfSparseTensor[Z][[2]];

(* ------------------------------------------------------------------ *)

SparseTensorProductOfSparseTensors[W_List?SoberSparseTensorQ, \
Y_List?SoberSparseTensorQ] :=
  Module[
    {Wui, Wli, Wc, Yui, Yli, Yc, WYli, WYui, WYc, WY},

    Wui = W[[All, 1, 1]];
    Wli = W[[All, 1, 2]];
    Wc = W[[All, 2]];
    Yui = Y[[All, 1, 1]];
    Yli = Y[[All, 1, 2]];
    Yc = Y[[All, 2]];
    WYui = Flatten[Outer[Join, Wui, Yui, 1], 1];
    WYli = Flatten[Outer[Join, Wli, Yli, 1], 1];
    WYc = Flatten[Outer[Times, Wc, Yc], 1];
    WY = Thread[{Thread[e[WYui, WYli]], WYc}];

    WY
  ];

(* ----------------------------------------------------------------- *)

IdentitySparseTensor[N_Integer?Positive, 1] :=
  IdentitySparseTensor[N, 1] =
    Map[{e[{#},{#}], 1}&, Range[N]];

IdentitySparseTensor[FF_Integer?Positive, N_Integer?Positive] :=
  IdentitySparseTensor[FF, N] =
  Module[
    {halfN, sit1, sit2, sit},

    halfN = Floor[N/2];
    sit1 = IdentitySparseTensor[FF, halfN];
    sit2 = IdentitySparseTensor[FF, N-halfN];
    sit = SparseTensorProductOfSparseTensors[sit1, sit2];

    sit
  ];

ZeroSparseTensor[rank_List] :=
  {{Apply[e, Map[Table[1, {#}]&, rank]], 0}};

(* ----------------------------------------------------------------- *)

TidySparseTensorComponents[Z_List?SoberSparseTensorQ] :=
  Module[
    {sZ, cZ},

    sZ = Z;
    sZ[[All,2]] = Expand[sZ[[All,2]]];
    sZ = Split[Sort[sZ], (#1[[1]] == #2[[1]])&];
    cZ = Map[{#[[1, 1]], Apply[Plus, #[[All,2]]]}&, sZ];
    cZ = Select[cZ, (#[[2]] =!= 0)&];
    If[cZ=={}, cZ = ZeroSparseTensor[RankOfSparseTensor[Z]]];

    cZ
  ];

(* ----------------------------------------------------------------- *)

AccreteSparseTensorXToSparseTensorZatStrings[Z_List?SoberSparseTensorQ, \
W_List?SoberSparseTensorQ, s_Integer?Positive] :=
  Module[
    {rZ, urW, lrZ, ZW, Wui, Wuip, Wli, Wc, r, sblanks, rblanks, j, ZWj},

    urW = UpperRankOfSparseTensor[W];
    lrZ = LowerRankOfSparseTensor[Z];
    rZ = RankOfSparseTensor[Z];
    r = lrZ - urW - s + 1;
    sblanks = Table[_, {s-1}];
    rblanks = Table[_, {r}];
    ZW = ZeroSparseTensor[rZ];
    Do[
      Wui = W[[j, 1, 1]];
      Wuip = Join[sblanks, Wui, rblanks];
      ZWj = Cases[Z, {e[_List, Wuip], _}];
      Wli = W[[j,1,2]];
      ZWj[[All,1,2]] = Map[Join[Take[#, s-1], Wli, Take[#, -r]]&, \
ZWj[[All,1,2]]];
      Wc = W[[j, 2]];
      ZWj[[All, 2]] = Wc * ZWj[[All, 2]];
      ZW = Join[ZW, ZWj],
      {j, Length[W]}
    ];
    ZW = TidySparseTensorComponents[ZW];

    ZW
  ] /; (s + UpperRankOfSparseTensor[W] - 1 <= LowerRankOfSparseTensor[Z]);

(* ----------------------------------------------------------------- *)

(* Presumes Z has same upper and lower ranks. Could be generalised. *)

ContractSparseTensorZOverSparseTensorHOnStrings[Z_List?SoberSparseTensorQ, \
H_List?SoberSparseTensorQ, s_Integer?Positive] :=
  Module[
    {ZH, Hind, Hc, ZHj, j},

    ZH = {};
    Do[
      Hind = H[[j, 1, All, 1]];
      Hc = H[[j, 2]];
      ZHj = Select[Z, (#[[1, All, s]] == Hind)&];
      ZHj = Map[Delete[#, {{1, 1, s}, {1, 2, s}}]&, ZHj];
      ZHj[[All, 2]] = Hc * ZHj[[All, 2]];
      ZH = Join[ZH, ZHj],
      {j, Length[H]}
    ];
    ZH[[All,2]] = Expand[ZH[[All,2]]];
    ZH = TidySparseTensorComponents[ZH];

    ZH
  ] /; (s <= UpperRankOfSparseTensor[Z]);

(* ----------------------------------------------------------------- *)

(* General purpose utility functions *)

ConvertRank4ArrayToRank4SparseTensor[r4a_List] :=
  Module[
    {D, r4st},

    D = Dimensions[r4a][[1]];
    r4st = Outer[{e[{#1,#3},{#2,#4}], r4a[[##]]}&, Range[D], Range[D], \
Range[D], Range[D]];
    r4st = Flatten[r4st, 3];

    r4st
  ] /; (TensorRank[r4a]==4);

ConvertRank4SparseTensorToRank4Array[r_List, D_Integer?Positive] :=
  Module[
    {rp, r4a},

    rp = Apply[Plus, Apply[Times, r, {1}]];
    r4a = Outer[Coefficient[rp, e[{#1, #3}, {#2, #4}]]&, Range[D], Range[D], \
\
\
Range[D], Range[D]];

    r4a
  ];

ConvertRank2ArrayToRank2SparseTensor[r2a_List] :=
  Module[
    {D, r2st},

    D = Dimensions[r2a][[1]];
    r2st = Outer[{e[{#1},{#2}], r2a[[##]]}&, Range[D], Range[D]];
    r2st = Flatten[r2st, 1];
    r2st = TidySparseTensorComponents[r2st];

    r2st
  ] /; (TensorRank[r2a]==2);

ConvertRank2SparseTensorToRank2Array[r_List, D_Integer?Positive] :=
  Module[
    {rp, r2a},

    rp = Apply[Plus, Apply[Times, r, {1}]];
    r2a = Outer[Coefficient[rp, e[{#1},{#2}]]&, Range[D], Range[D]];

    r2a
  ];


DecrementRankOfTensorObject[T_List] :=
  Module[
    {n, l, tt, nT},

    n = TensorRank[T];
    l = Join[Range[n-3], {n-1,n-2,n}];
    tT = Transpose[T, l];
    tT = Map[Flatten, tT, {n-2}];
    nT = Transpose[tT, Drop[l,-1]];

    nT
  ] /; (TensorRank[T] > 2);

ConvertRank4ArrayObjectToRank2ArrayObject[r4a_List] :=
  Nest[DecrementRankOfTensorObject, r4a, 2];

ConvertRank2ArrayObjectToRank4ArrayObject[r2a_List] :=
  Partition[r2a, {Sqrt[Dimensions[r2a][[1]]], Sqrt[Dimensions[r2a][[1]]]}] /;
    (TensorRank[r2a]==2);

(* ----------------------------------------------------------------- *)

BraidRelationTestForRank4ArrayObject[r4a_] :=
  Module[
    {I, B, BI, IB, ans},

    B = r4a;
    I = IdentityMatrix[Dimensions[B][[1]]];
    BI = Nest[DecrementRankOfTensorObject, Outer[Times, B, I], 4];
    IB = Nest[DecrementRankOfTensorObject, Outer[Times, I, B], 4];
    ans = (Expand[BI.IB.BI] === Expand[IB.BI.IB]);

    ans
  ];

BraidRelationTestForRank2ArrayObject[r2a_] :=
  BraidRelationTestForRank4ArrayObject[
    ConvertRank2ArrayObjectToRank4ArrayObject[r2a]
  ];

(* ------------------------------------------------------------------ *)

(* End[]; *)

EndPackage[];



